(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)

theory vcg

  imports "../../Midend-IVL/Isabelle-UTP-Extended/HoareLogic/TotalCorrectness/utp_hoare_ndes_prog"
          "Eisbach_Tools2"
begin
recall_syntax \<comment> \<open>Fixes notation issue with inclusion of HOL libraries.\<close>

section \<open>simplifier setup\<close>
declare utp_lift.lift_pre_var [simp del]

chapter \<open>VCG\<close> 
  
section \<open>VCG General Purpose Tactics\<close>      
text \<open>Automating premises insertion\<close>  
method_setup insert_assms = \<open>Scan.succeed (fn _ => CONTEXT_METHOD (fn facts => fn (ctxt,st) => let
   val tac = HEADGOAL (Method.insert_tac ctxt facts)
   val ctxt = Method.set_facts [] ctxt
 in Method.CONTEXT ctxt (tac st)
 end))\<close>                      
                       
text \<open>The defer processing and the thin_tac processing in the sequel was inspired by tutorial5.thy in Peter Lammich course
        \url{https://bitbucket.org/plammich/certprog_public/downloads/}\<close>
 
subsection \<open>Deterministic Repeated Elimination Rule\<close>
text \<open>Attention: Slightly different semantics than @{method elim}: repeats the 
  rule as long as possible, but only on the first subgoal.\<close>

method_setup vcg_elim_determ = \<open>
  Attrib.thms >> (fn thms => fn ctxt => 
    SIMPLE_METHOD (REPEAT_DETERM1 (HEADGOAL (ematch_tac ctxt thms))))\<close>
text \<open>The \<open>DETERM\<close> combinator on method level\<close>
  
method_setup determ = \<open>
  Method.text_closure >>
    (fn (text) => fn ctxt => fn using => fn st =>
      Seq.DETERM (Method.evaluate_runtime text ctxt using) st
    )
\<close>        
(*method insert_assms =  tactic \<open>@{context} |>  Assumption.all_prems_of |> (@{context} |>  Method.insert_tac) |> FIRSTGOAL\<close>*)
    

  
text \<open>vcg_can_defer is a tactic that succeed if the conclusion of a goal is not Hoare triple
       or if it has no DEFERRED markup\<close>  
  
definition DEFERRED :: "bool \<Rightarrow> bool" where "DEFERRED P = P"
lemma DEFERREDD: "DEFERRED P \<Longrightarrow> P" by (auto simp: DEFERRED_def)
(*TODO: FINISH THE PROTOTYPE OF THE DEBUGGER VERSION*)    
method vcg_can_defer_debugger =
  (match conclusion 
      in "DEFERRED _" \<Rightarrow> \<open>print_term \<open>''DEFERRED''\<close>,fail\<close>  -- \<open>Refuse to defer already deferred goals\<close>
       \<bar>
         "(\<lbrace>_\<rbrace>a\<lbrace>_\<rbrace>\<^sub>P)" for a \<Rightarrow> \<open>print_term "a",fail\<close>  -- \<open>Refuse to defer Hoare_H1_H3 triples (They are no VCs!)\<close>
       \<bar>
         "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>print_term "''u''",fail\<close>  -- \<open>Refuse to defer Hoare_rel triples (They are no VCs!)\<close>  
       \<bar> 
         "_" \<Rightarrow> \<open>print_term "''succeed''",succeed\<close>)
       
method vcg_defer_debugger = 
       (print_term "''vcg_avant''",vcg_can_defer_debugger, print_term "''vcg_apres''",
        rule DEFERREDD, tactic \<open>FIRSTGOAL defer_tac\<close>)    
  
method vcg_can_defer =
  (match conclusion 
      in "DEFERRED _" \<Rightarrow> \<open>fail\<close>  -- \<open>Refuse to defer already deferred goals\<close>
       \<bar>
         "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>  -- \<open>Refuse to defer Hoare_H1_H3 triples (They are no VCs!)\<close>
       \<bar>
         "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>fail\<close>  -- \<open>Refuse to defer Hoare_rel triples (They are no VCs!)\<close>  
       \<bar> 
         "_" \<Rightarrow> \<open>succeed\<close>)
       
method vcg_defer = (vcg_can_defer,rule DEFERREDD, tactic \<open>FIRSTGOAL defer_tac\<close>)    
  
section \<open>VCG Post Processing Tactics\<close> 
  
text \<open>Tactics and methods in this section are used to do Post-Processing on the generated VCs.
      Namely, The application of symbolic execution laws from the theory usubst
      to the VCs in a very controlled way.\<close>  

subsection \<open>substitution simplifier for debugging mode\<close>  
  
definition "ZERO_SUBST_TAG expr = True" 
definition "ONE_SUBST_TAG expr = True"
definition "LIT_SUBST_TAG expr = True"
definition "VAR_SUBST_TAG expr = True"  
definition "UOP_SUBST_TAG expr = True"    
definition "BOP_SUBST_TAG expr = True"
definition "TROP_SUBST_TAG expr= True"
definition "QTOP_SUBST_TAG expr = True"
  
lemma ZERO_SUBST_DEBUG: 
  "(ZERO_SUBST_TAG 0 \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding ZERO_SUBST_TAG_def 
  by blast 

lemma ONE_SUBST_DEBUG: 
  "(ONE_SUBST_TAG 1 \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding ONE_SUBST_TAG_def 
  by blast 
    
lemma LIT_SUBST_DEBUG: 
  "(LIT_SUBST_TAG (lit v) \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding LIT_SUBST_TAG_def 
  by blast  

lemma VAR_SUBST_DEBUG: 
  "(VAR_SUBST_TAG (utp_expr.var x) \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding VAR_SUBST_TAG_def 
  by blast
    
lemma UOP_SUBST_DEBUG: 
  "(UOP_SUBST_TAG (uop f a) \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding UOP_SUBST_TAG_def 
  by blast
    
lemma BOP_SUBST_DEBUG: 
  "(BOP_SUBST_TAG (bop f a b) \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding BOP_SUBST_TAG_def 
  by blast     
    
lemma EQ_UPRED_SUBST_DEBUG: 
  "(BOP_SUBST_TAG (eq_upred a b) \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding BOP_SUBST_TAG_def 
  by blast
    
lemma TROP_SUBST_DEBUG: 
  "(TROP_SUBST_TAG (trop f a b c) \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding TROP_SUBST_TAG_def 
  by blast     
    
lemma QTOP_SUBST_DEBUG: 
  "(QTOP_SUBST_TAG (qtop f a b c d) \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding QTOP_SUBST_TAG_def 
  by blast 
    
method subst_debugger = 
    (*Zero*)
   (match conclusion in  
    "_ (\<lambda> _. subst _ 0)"  \<Rightarrow>
     \<open>rule ZERO_SUBST_DEBUG , (simp only:subst_zero)\<close>)
    (*Zero*)   
    | (match conclusion in   
    "_ (\<lambda> _. subst _ 1)"  \<Rightarrow>
     \<open>rule ONE_SUBST_DEBUG , (simp only:subst_one)\<close>)
    (*UTP vars*) 
    |(match conclusion in 
    "_ (\<lambda> _. (subst _ (utp_expr.var x)))" for x \<Rightarrow>
     \<open>rule VAR_SUBST_DEBUG[where x= x], (simp only:subst_var)\<close>)
   (*UTP Literals*)
    |(match conclusion in 
    "_ (\<lambda> _ . (subst _ (lit v)))" for v \<Rightarrow>
     \<open>rule LIT_SUBST_DEBUG[where v= v], (simp only:subst_lit)\<close>)
    (*UTP Unary operation*)
    | (match conclusion in 
    "_ (\<lambda> _ .(subst _ (uop f a)))" for f a \<Rightarrow>
     \<open>rule UOP_SUBST_DEBUG[where f= f and a = a], simp only: subst_uop\<close>) 
    (*Derived UOP for UTP Logical Operators*)
    | (match conclusion in 
    "_ (\<lambda> _ .(subst _ (\<not> a)))" for  a \<Rightarrow>
     \<open>rule UOP_SUBST_DEBUG[where f= "Not" and a = a], simp only: subst_uop\<close>)
   | (match conclusion in 
    "_ (\<lambda> _. (subst _ (\<^bold>\<forall>x \<bullet> P x)))" for P \<Rightarrow>
     \<open>rule UOP_SUBST_DEBUG[where f= "All" and a= "(\<lambda>x \<bullet> P x)" ] , simp only: utp_pred.subst_shAll\<close>) 
   | (match conclusion in 
    "_ (\<lambda> _. (subst _ (\<^bold>\<exists>x \<bullet> P x)))" for P \<Rightarrow>
     \<open>rule UOP_SUBST_DEBUG[where f= "Ex" and a= "(\<lambda>x \<bullet> P x)" ] , simp only: utp_pred.subst_shEx\<close>)
   (*UTP Binary operation*)
    | (match conclusion in 
    "_ (\<lambda> _ . (subst _ (bop f a b)))" for f a b \<Rightarrow>
     \<open> rule BOP_SUBST_DEBUG[where f= f and a= a and b = b], simp only:subst_bop\<close>) 
   (*Derived BOP for UTP Arith Operators*)
   | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a =\<^sub>u b)))" for a b \<Rightarrow>
     \<open>rule EQ_UPRED_SUBST_DEBUG[where a= a and b = b], simp only:subst_eq_upred\<close>)
   | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a + b)))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op +)" and a= a and b = b], simp only:subst_plus\<close>)
   | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a - b)))" for a b \<Rightarrow>
     \<open> rule BOP_SUBST_DEBUG[where f= "(op -)" and a= a and b = b], simp only:subst_minus\<close>)
    | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a * b)))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op *)" and a= a and b = b],simp only:subst_times \<close>)
    | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a div b)))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op div)" and a= a and b = b], simp only:subst_div\<close>)
    (*Derived BOP Logical Operators*)
    | (match conclusion in 
    "(_ (\<lambda> _. (subst _ (a \<and> b))))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op \<and>)" and a= a and b = b] , simp only: utp_pred.subst_conj\<close>)
    | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a \<or> b)))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op \<or>)" and a= a and b = b] , simp only: utp_pred.subst_disj\<close>)
    | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a \<Rightarrow> b)))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op \<longrightarrow>)" and a= a and b = b] , simp only: utp_pred.subst_impl\<close>)     
   | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a \<Leftrightarrow> b)))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op \<longleftrightarrow>)" and a= a and b = b] , simp only: utp_pred.subst_iff\<close>)
   | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a \<sqinter> b)))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op \<sqinter>)" and a= a and b = b] , simp only: utp_pred.subst_sup\<close>)
    | (match conclusion in 
    "_ (\<lambda> _. (subst _ (a \<squnion> b)))" for a b \<Rightarrow>
     \<open>rule BOP_SUBST_DEBUG[where f= "(op \<squnion>)" and a= a and b = b] , simp only: utp_pred.subst_inf\<close>)
   (*UTP Ternary operation*)
    | (match conclusion in 
    "_ (\<lambda> _. (subst _ (trop f a b c)))" for f a b c\<Rightarrow>
     \<open>rule TROP_SUBST_DEBUG[where f=f and a=a and b=b and c=c] , simp only:subst_trop\<close>)
    (*UTP Quaternary operation*)
    | (match conclusion in 
    "_ (\<lambda> _. (subst _ (qtop f a b c d)))" for f a b c d\<Rightarrow>
     \<open>rule QTOP_SUBST_DEBUG[where f=f and a=a and b=b and c=c and d=d],  simp only:subst_qtop\<close>)


definition "SUBST_ID_LOOKUP_TAG \<sigma> x = True"  
definition "SUBST_UPD_LOOKUP_TAG \<sigma> x y = True" 
definition "UEX_BOUNDED x = x" 

lemma SUBST_UPD_LOOKUP_DEBUG: 
  "(SUBST_UPD_LOOKUP_TAG \<sigma> (UEX_BOUNDED x) y \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding SUBST_UPD_LOOKUP_TAG_def 
  by blast 

lemma SUBST_ID_LOOKUP_DEBUG: 
  "(SUBST_ID_LOOKUP_TAG id y \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding SUBST_ID_LOOKUP_TAG_def 
  by blast 
    
method subst_lookup_debugger = 
     (match conclusion in  
     "_ (\<lambda> _. (usubst_lookup id y))" for  y  \<Rightarrow>
      \<open>rule SUBST_ID_LOOKUP_DEBUG[where y=y], 
       (simp only: usubst_lookup_id)\<close>)
    |
     (match conclusion in  
      "_ (\<lambda> _. (usubst_lookup (subst_upd \<sigma> x  _) y))" for \<sigma> x y  \<Rightarrow>
       \<open>rule SUBST_UPD_LOOKUP_DEBUG[where \<sigma>=\<sigma> and x=x and y=y], 
        (simp only: usubst_lookup_upd_indep usubst_lookup_ovar_unrest
                    usubst_lookup_ivar_unrest usubst_lookup_upd 
                    vwb_lens_mwb vwb_lens_wb lens_indep_sym)\<close>) 
       
text \<open>very well behaved lens simplifier for debugging mode\<close>  
 
definition "VWB_VAR_TAG x = True"
definition "WB_VAR_TAG x = True"
definition "WEAK_VAR_TAG x = True"
definition "MWB_VAR_TAG x = True"  

lemma VWB_VAR_DEBUG: 
  obtains e where "e = vwb_lens x " "VWB_VAR_TAG x" 
  unfolding VWB_VAR_TAG_def 
  by blast

lemma MWB_VAR_DEBUG: 
  obtains e where "e = mwb_lens x " "MWB_VAR_TAG x" 
  unfolding MWB_VAR_TAG_def 
  by blast   
    
lemma WB_VAR_DEBUG: 
  obtains e where "e = wb_lens x " "WB_VAR_TAG x" 
  unfolding WB_VAR_TAG_def 
  by blast  
    
lemma WEAK_VAR_DEBUG: 
  obtains e where "e = wb_lens x " "WEAK_VAR_TAG x" 
  unfolding WEAK_VAR_TAG_def 
  by blast  

method vwb_lens_debugger =
  (match conclusion in 
   "vwb_lens x" for x \<Rightarrow>
   \<open>(rule VWB_VAR_DEBUG[where x= x],assumption) (*if this fails add a debug message here*)\<close>)
  |(match conclusion in 
   "wb_lens x" for x \<Rightarrow>
   \<open>rule WB_VAR_DEBUG[where x= x],(simp only: vwb_lens_wb)(*if this fails add a debug message here*)\<close>)
  |(match conclusion in 
   "mwb_lens x" for x \<Rightarrow>
   \<open>rule MWB_VAR_DEBUG[where x= x],(simp only: vwb_lens_mwb)(*if this fails add a debug message here*)\<close>)
  |(match conclusion in 
   "mwb_lens x" for x \<Rightarrow>
   \<open>rule WEAK_VAR_DEBUG[where x= x],(simp only: vwb_lens_weak)(*if this fails add a debug message here*)\<close>)

definition "WF_TAG expr = True"
   
lemma WF_DEBUG:
 "(WF_TAG expr \<Longrightarrow> thesis) \<Longrightarrow> thesis" 
  unfolding WF_TAG_def by simp

method wf_debugger =
  (match conclusion in 
   "wf expr" for expr \<Rightarrow>
   \<open>rule WF_DEBUG[where expr = expr], simp\<close>) 

    
section \<open>Custom UTP tactics for VCG\<close>
  
text \<open>For debugging purpose we re-construct utp tactics. It allows  controlled  execution 
       of these tactics.\<close> 
 (*TODO: Debugging version of these tactics*)

text \<open>a generic UTP tactic inspired by @{method gen_rel_tac} and @{method gen_pred_tac}\<close>
  
method utp_tac_control methods utp_defs utp_transfer utp_rewrites utp_solve utp_interp =
  utp_defs? ; utp_transfer, utp_rewrites?, utp_interp?, utp_solve
                                                   
text \<open>For UTP transfer tactics see @{method slow_uexpr_transfer} and @{method fast_uexpr_transfer}\<close>

text \<open>For UTP interp tactics see @{method uexpr_interp_tac}\<close>  

text \<open>For UTP solve tactics see @{method utp_simp_tac}\<close>  
  
subsubsection \<open> UTP predicates tactics\<close>

method upreds_defs = (unfold upred_defs)[1] 
thm lens_defs
method upreds_rewrites =  (simp add: fun_eq_iff lens_defs upred_defs alpha_splits Product_Type.split_beta)
  
method upreds_simp       = utp_tac_control upreds_defs fast_uexpr_transfer upreds_rewrites uexpr_interp_tac utp_simp_tac
method upreds_simp_slow  = utp_tac_control upreds_defs slow_uexpr_transfer upreds_rewrites uexpr_interp_tac utp_simp_tac
method upreds_auto       = utp_tac_control upreds_defs fast_uexpr_transfer upreds_rewrites uexpr_interp_tac utp_auto_tac
method upreds_auto_slow  = utp_tac_control upreds_defs slow_uexpr_transfer upreds_rewrites uexpr_interp_tac utp_auto_tac
method upreds_blast      = utp_tac_control upreds_defs fast_uexpr_transfer upreds_rewrites uexpr_interp_tac utp_blast_tac
method upreds_blast_slow = utp_tac_control upreds_defs slow_uexpr_transfer upreds_rewrites uexpr_interp_tac utp_blast_tac

subsubsection \<open> UTP relations tactics\<close>

method urels_defs       = (unfold upred_defs urel_defs)[1] 
method urels_rewrites   = (simp add: fun_eq_iff relcomp_unfold OO_def lens_defs upred_defs 
                                     alpha_splits Product_Type.split_beta)  
method urels_simp       = utp_tac_control urels_defs fast_uexpr_transfer urels_rewrites uexpr_interp_tac utp_simp_tac
method urels_simp_slow  = utp_tac_control urels_defs slow_uexpr_transfer urels_rewrites uexpr_interp_tac utp_simp_tac
method urels_auto       = utp_tac_control urels_defs fast_uexpr_transfer urels_rewrites uexpr_interp_tac utp_auto_tac
method urels_auto_slow  = utp_tac_control urels_defs slow_uexpr_transfer urels_rewrites uexpr_interp_tac utp_auto_tac
method urels_blast      = utp_tac_control urels_defs fast_uexpr_transfer urels_rewrites uexpr_interp_tac utp_blast_tac
method urels_blast_slow = utp_tac_control urels_defs slow_uexpr_transfer urels_rewrites uexpr_interp_tac utp_blast_tac
 
section \<open>VCG for non-debugging mode\<close> 

text \<open>In this section we define the core tactics for the VCG. Namely, tactics for the computational mode
such as weakest pre-condition and strongest post_condition rules. Also tactics for symbolic execution
on the generated verification conditions are defined.\<close>  

subsection\<open>Auxiliary lemmas for VCG\<close>

lemma FRAME_TRIPLE_E: "FRAME_TRIPLE P \<Longrightarrow> P" by (auto simp: FRAME_TRIPLE_def)

definition DONT_TOUCH :: "bool \<Rightarrow> bool" where "DONT_TOUCH P = P"

lemma DONT_TOUCH_E: "DONT_TOUCH P \<Longrightarrow> P" by (auto simp: DONT_TOUCH_def)

subsection \<open>VCG first goal pre-processing\<close>

method vcg_first_goal_preprocessing =
  (match conclusion in "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>(simp only: seqr_assoc[symmetric])?\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>(simp only: seqr_assoc[symmetric])?\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>G" \<Rightarrow> \<open>(simp only: seqr_assoc[symmetric])?, 
                                     (simp only:Product_Type.split_paired_All Product_Type.prod.case)?, 
                                     ((rule allI)+)?\<close>)
 |(match conclusion in "_" \<Rightarrow> \<open>fail\<close>  -- \<open>fail for any other goal with different syntax\<close>)

method precondition_strengthening =
  (match conclusion in "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>rule pre_weak_prog_hoare\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>rule pre_str_hoare_rel\<close>)
 |(match conclusion in "_"       \<Rightarrow> \<open>fail\<close> -- \<open>fail for any other goal with different syntax\<close>)

method postcondition_weakening =
  (match conclusion in "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>rule post_str_prog_hoare\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>rule post_weak_hoare_rel\<close>)
 |(match conclusion in "_"       \<Rightarrow>  \<open>fail\<close> -- \<open>fail for any other goal with different syntax\<close>)

method hoare_wp_vcg_preprocessing =
       vcg_first_goal_preprocessing, 
       precondition_strengthening

method hoare_sp_vcg_preprocessing =
       vcg_first_goal_preprocessing, 
       postcondition_weakening

subsection\<open>The core WP tactic for VCG\<close>
method hoare_wp_rule_apply_total =
 --\<open>WP Matching rules for total correctness\<close>
  ((match conclusion in "\<lbrace>_\<rbrace>\<langle>_\<rangle>\<^sub>p\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> succeed), rule assigns_prog_hoare_backward)
 |((match conclusion in "\<lbrace>_\<rbrace>SKIP\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> succeed), rule skip_prog_hoare)
 |((match conclusion in "\<lbrace>_\<rbrace>_;_\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> succeed), rule seq_prog_hoare)
 |((match conclusion in "\<lbrace>_\<rbrace>IF _ THEN _ ELSE _ FI\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> succeed), rule cond_prog_hoare_wp)
 (*Scoping programming construct!*)
 |((match conclusion in "\<lbrace>_\<rbrace>MODIFY g DO _ OD\<lbrace>_\<rbrace>\<^sub>P" for g \<Rightarrow> succeed), 
   rule antiframe_prog_hoare_wp_weak[where g= fst_lens and l= snd_lens]; rule FRAME_TRIPLE_E)
 (*Special handling for the VCs generated from scoping construct*)
 |((match conclusion in "FRAME_TRIPLE (g:[_] = _)"for g \<Rightarrow> \<open>succeed\<close>), 
                         (simp only: FRAME_TRIPLE_def), (rule DONT_TOUCH_E))
 |((match conclusion in "FRAME_TRIPLE \<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>succeed\<close>),
                         (simp only: FRAME_TRIPLE_def, rule vcg_proved_specs))
 |((match conclusion in "FRAME_TRIPLE _" \<Rightarrow> \<open>succeed\<close>), (simp only: FRAME_TRIPLE_def))
 (*Annotated loop patterns for total correctness*)
 |((match conclusion in "\<lbrace>_\<rbrace>INVAR _ VRT _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>succeed\<close>), 
                         rule while_invr_vrt_hoare_prog_wp[where R= measure and st' = "&\<Sigma>"]) 
 |((match conclusion in "\<lbrace>_\<rbrace>FOR (_,_,_) INVAR _ VRT _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"    \<Rightarrow> \<open>succeed\<close>), 
                         subst for_lfp_invar_vrt_prog_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>DO _ INVAR _ VRT _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>succeed\<close>), 
                         subst do_while_invar_vrt_lfp_prog_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>FROM _ INVAR _ VRT _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>succeed\<close>), 
                         subst from_until_lfp_invr_vrt_prog_def_alt)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 | (match conclusion in "\<lbrace>_\<rbrace>INVAR _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>) 
 | (match conclusion in "\<lbrace>_\<rbrace>FOR (_,_,_)INVAR _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>DO _ INVAR _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>FROM _ INVAR _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>) 
 | (match conclusion in "\<lbrace>_\<rbrace>FOR (_,_,_)DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>DO _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>FROM _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>)
 (*Annotated SCOPED loop patterns for total correctness*)
 |((match conclusion in "\<lbrace>_\<rbrace>IN _ INVAR _ VRT _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"         \<Rightarrow> \<open>succeed\<close>), 
                         rule while_scp_invr_vrt_prog_hoare_wp[where st' = "&\<Sigma>" and R = measure and 
                                                                     g = fst_lens and l = snd_lens]; 
                         rule FRAME_TRIPLE_E) 
 |((match conclusion in "\<lbrace>_\<rbrace>IN _ FOR (_,_,_) INVAR _ VRT _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>succeed\<close>), 
                         subst for_lfp_scp_invr_vrt_prog_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>IN _ DO _ INVAR _ VRT _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"         \<Rightarrow> \<open>succeed\<close>), 
                         subst do_while_lfp_scp_invr_vrt_prog_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>IN _ FROM _ INVAR _ VRT _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"  \<Rightarrow> \<open>succeed\<close>), 
                         subst from_until_lfp_scp_invr_vrt_prog_def_alt)
 (*Non-annotated SCOPED loop patterns fail the VCG... which requires a proof annotation *)
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ INVAR _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>) 
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ FOR (_,_,_)INVAR _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ DO _ INVAR _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ FROM _ INVAR _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>) 
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ FOR (_,_,_)DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ DO _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ FROM _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>)

method hoare_wp_rule_apply_partial =
 --\<open>WP Matching rules for partial correctness\<close>
 ((match conclusion in "\<lbrace>_\<rbrace>\<langle>_\<rangle>\<^sub>a\<lbrace>_\<rbrace>\<^sub>u"  \<Rightarrow> \<open>succeed\<close>), rule assigns_backward_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>SKIP\<^sub>r\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), rule skip_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>_;;_\<lbrace>_\<rbrace>\<^sub>u"  \<Rightarrow> \<open>succeed\<close>), rule seq_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>if\<^sub>u _ then _ else _\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), rule cond_hoare_rel_wp)
  (*Scoping programming construct for relations!*)
 |((match conclusion in "\<lbrace>_\<rbrace>antiframe _ _\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), rule cond_hoare_rel_wp; rule FRAME_TRIPLE_E)
 (*Special handling for the VCs generated from scoping construct*)
 |((match conclusion in "FRAME_TRIPLE (antiframe _ _ = _)" \<Rightarrow> \<open>succeed\<close>), 
                         (simp only: FRAME_TRIPLE_def), (rule DONT_TOUCH_E))
 |((match conclusion in "FRAME_TRIPLE \<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>),
                         (simp only: FRAME_TRIPLE_def, rule vcg_proved_specs))
 |((match conclusion in "FRAME_TRIPLE _" \<Rightarrow> \<open>succeed\<close>), (simp only: FRAME_TRIPLE_def))
 (*Annotated loop patterns for total correctness*)
 |((match conclusion in "\<lbrace>_\<rbrace>invr _ while\<^sup>\<top> _ do _ od\<lbrace>_\<rbrace>\<^sub>u"  \<Rightarrow> \<open>succeed\<close>), 
                         rule while_gfp_invr_hoare_rel_wp_partial)
 |((match conclusion in "\<lbrace>_\<rbrace>do _ invr _ while\<^sup>\<top> _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst do_while_invar_vrt_gfp_rel_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>for\<^sup>\<top> (_,_,_) invr _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst for_gfp_invar_vrt_rel_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>from\<^sup>\<top> _ invr _ until _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst from_until_gfp_invr_vrt_rel_def_alt)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 |(match conclusion in "\<lbrace>_\<rbrace>while\<^sup>\<top> _ do _ od\<lbrace>_\<rbrace>\<^sub>u"        \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>do _ while\<^sup>\<top> _ od\<lbrace>_\<rbrace>\<^sub>u"        \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>for\<^sup>\<top> (_,_,_) do _ od\<lbrace>_\<rbrace>\<^sub>u"    \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>from\<^sup>\<top> _ until _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>fail\<close>)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 |(match conclusion in "\<lbrace>_\<rbrace>invr _ vrt _ while\<^sub>\<bottom> _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>fail\<close>)
--{*For any goal which is not a Hoare triple*}
 |((match conclusion in "_" \<Rightarrow> \<open>succeed\<close>), vcg_defer)
method hoare_wp_rule_apply =
 hoare_wp_rule_apply_total
| hoare_wp_rule_apply_partial

method wp = hoare_wp_vcg_preprocessing, hoare_wp_rule_apply+, (unfold DEFERRED_def)

subsection\<open>The core SP tactic for VCG\<close>

method hoare_sp_rule_apply =
 --\<open>WP Matching rules for total correctness\<close>
  ((match conclusion in "\<lbrace>_\<rbrace>\<langle>_\<rangle>\<^sub>p\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> succeed), rule assigns_prog_floyd)
 |((match conclusion in "\<lbrace>_\<rbrace>SKIP\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> succeed), rule skip_prog_hoare)
 |((match conclusion in "\<lbrace>_\<rbrace>_;_\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> succeed), rule seq_prog_hoare)
 |((match conclusion in "\<lbrace>_\<rbrace>IF _ THEN _ ELSE _ FI\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> succeed), rule cond_prog_hoare_sp)
 (*Scoping programming construct!*)
 |((match conclusion in "\<lbrace>_\<rbrace>g:[_]\<lbrace>_\<rbrace>\<^sub>P" for g \<Rightarrow> succeed), 
   rule antiframe_prog_hoare_sp_weak[where g= fst_lens and l= snd_lens]; rule FRAME_TRIPLE_E)
 (*Special handling for the VCs generated from scoping construct*)
 |((match conclusion in "FRAME_TRIPLE (g:[_] = _)"for g \<Rightarrow> \<open>succeed\<close>), 
                         (simp only: FRAME_TRIPLE_def), (rule DONT_TOUCH_E))
 |((match conclusion in "FRAME_TRIPLE \<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>succeed\<close>),
                         (simp only: FRAME_TRIPLE_def, rule vcg_proved_specs))
 |((match conclusion in "FRAME_TRIPLE _" \<Rightarrow> \<open>succeed\<close>), (simp only: FRAME_TRIPLE_def))
 (*Annotated loop patterns for total correctness*)
 |((match conclusion in "\<lbrace>_\<rbrace>INVAR _ VRT _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>succeed\<close>), 
                         rule while_invr_vrt_hoare_prog_sp[where R= measure and st' = "&\<Sigma>"]) 
 |((match conclusion in "\<lbrace>_\<rbrace>FOR (_,_,_) INVAR _ VRT _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"    \<Rightarrow> \<open>succeed\<close>), 
                         subst for_lfp_invar_vrt_prog_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>DO _ INVAR _ VRT _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>succeed\<close>), 
                         subst do_while_invar_vrt_lfp_prog_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>FROM _ INVAR _ VRT _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>succeed\<close>), 
                         subst from_until_lfp_invr_vrt_prog_def_alt)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 | (match conclusion in "\<lbrace>_\<rbrace>INVAR _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>) 
 | (match conclusion in "\<lbrace>_\<rbrace>FOR (_,_,_)INVAR _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>DO _ INVAR _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>FROM _ INVAR _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>) 
 | (match conclusion in "\<lbrace>_\<rbrace>FOR (_,_,_)DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>DO _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>)
 | (match conclusion in "\<lbrace>_\<rbrace>FROM _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>)
 --\<open>WP Matching rules for partial correctness\<close>
 |((match conclusion in "\<lbrace>_\<rbrace>\<langle>_\<rangle>\<^sub>a\<lbrace>_\<rbrace>\<^sub>u"  \<Rightarrow> \<open>succeed\<close>), rule assigns_floyd_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>SKIP\<^sub>r\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), rule skip_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>_;;_\<lbrace>_\<rbrace>\<^sub>u"  \<Rightarrow> \<open>succeed\<close>), rule seq_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>if\<^sub>u _ then _ else _\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), rule cond_hoare_rel_sp)
  (*Scoping programming construct for relations!*)
 |((match conclusion in "\<lbrace>_\<rbrace>antiframe _ _\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), rule cond_hoare_rel_sp; 
                                                             rule FRAME_TRIPLE_E)
 (*Special handling for the VCs generated from scoping construct*)
 |((match conclusion in "FRAME_TRIPLE (antiframe _ _ = _)" \<Rightarrow> \<open>succeed\<close>), 
                         (simp only: FRAME_TRIPLE_def), (rule DONT_TOUCH_E))
 |((match conclusion in "FRAME_TRIPLE \<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>),
                         (simp only: FRAME_TRIPLE_def, rule vcg_proved_specs))
 |((match conclusion in "FRAME_TRIPLE _" \<Rightarrow> \<open>succeed\<close>), (simp only: FRAME_TRIPLE_def))
 (*Annotated loop patterns for total correctness*)
 |((match conclusion in "\<lbrace>_\<rbrace>invr _ while\<^sup>\<top> _ do _ od\<lbrace>_\<rbrace>\<^sub>u"  \<Rightarrow> \<open>succeed\<close>), 
                         rule while_gfp_invr_hoare_rel_sp_partial)
 |((match conclusion in "\<lbrace>_\<rbrace>do _ invr _ while\<^sup>\<top> _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst do_while_invar_vrt_gfp_rel_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>for\<^sup>\<top> (_,_,_) invr _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst for_gfp_invar_vrt_rel_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>from\<^sup>\<top> _ invr _ until _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst from_until_gfp_invr_vrt_rel_def_alt)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 |(match conclusion in "\<lbrace>_\<rbrace>while\<^sup>\<top> _ do _ od\<lbrace>_\<rbrace>\<^sub>u"        \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>do _ while\<^sup>\<top> _ od\<lbrace>_\<rbrace>\<^sub>u"        \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>for\<^sup>\<top> (_,_,_) do _ od\<lbrace>_\<rbrace>\<^sub>u"    \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>from\<^sup>\<top> _ until _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>fail\<close>)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 |(match conclusion in "\<lbrace>_\<rbrace>invr _ vrt _ while\<^sub>\<bottom> _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>fail\<close>)
--{*For any goal which is not a Hoare triple*}
 |((match conclusion in "_" \<Rightarrow> \<open>succeed\<close>), vcg_defer)

method sp = hoare_sp_vcg_preprocessing, hoare_sp_rule_apply+, (unfold DEFERRED_def)

subsection \<open>VCG post-processing\<close>
text \<open>After the generation of VCs using @{method wp} a symbolic execution process is required
      to go from UTP expressions to HOL expressions. Symbolic execution consists of two steps
      reasoning:
      \<^item> SE step1: simplification of the substitutions using rules from the meta-theory on substitutions
                  mechanised in @{theory utp_subst}. This requires to carry out lemmas related to
                  lenses.
      \<^item> SE step2: Droping from the simplified UTP expressions to HOL expressions using mainly
                  the lifting and transfer rule generated from the lifting and transfer package of 
                  Isabelle.\<close>

--\<open>In the sequel the setup for SE step1. Namely, the setup for 
   substitution simplifier for non-debugging mode\<close>

named_theorems usubst_simplifier 
declare usubst[usubst_simplifier] 
declare vwb_lens_weak[usubst_simplifier]
declare vwb_lens_mwb[usubst_simplifier] 
declare vwb_lens_wb[usubst_simplifier] 
declare lens_indep_sym[usubst_simplifier] 

--\<open>Since SE step1 and SE step2 require to carry out simplification lemmas for Lens, thus,
   we do the setup for such simplifications in the sequel and this for non-debugging mode\<close>
  
named_theorems vwb_simplifier
declare vwb_lens_wb[vwb_simplifier]
declare vwb_lens_mwb[vwb_simplifier]
declare vwb_lens_weak[vwb_simplifier]
declare bij_lens_vwb[vwb_simplifier]
declare Lens_Algebra.id_bij_lens[vwb_simplifier] 

named_theorems lens_laws_vcg_simps
lemmas [lens_laws_vcg_simps] =
  lens_indep.lens_put_irr1
  lens_indep.lens_put_irr2
  
named_theorems lens_get_lens_put_simplifer
declare lens_laws_vcg_simps[lens_get_lens_put_simplifer]  
declare vwb_simplifier[lens_get_lens_put_simplifer]
declare mwb_lens.put_put[lens_get_lens_put_simplifer] 
declare weak_lens.put_get[lens_get_lens_put_simplifer]
declare wb_lens.get_put[lens_get_lens_put_simplifer]
declare bij_lens.strong_get_put[lens_get_lens_put_simplifer]
  
named_theorems lens_indep_all_simplifier
declare distinct.simps[lens_indep_all_simplifier]
declare HOL.conj_assoc[lens_indep_all_simplifier]
declare HOL.simp_thms[lens_indep_all_simplifier]
declare List.list.set[lens_indep_all_simplifier]
declare Set.ball_simps[lens_indep_all_simplifier]  
declare Set.insert_iff[lens_indep_all_simplifier] 
declare Set.empty_iff [lens_indep_all_simplifier]

--\<open>The method in the sequel implements SE step1.\<close>
method vcg_upreds_post_processing =
        (assumption|simp only: vwb_simplifier)
        |simp
        |(simp only:usubst_simplifier)

--\<open>The method in the sequel implements SE step2.\<close> 

method vcg_hol_post_processing = 
   (upreds_simp)?, (auto simp only: lens_get_lens_put_simplifer)?

--\<open>The method in the sequel puts together SE step1 and SE step2.\<close>

method post_processing = ((unfold pr_var_def in_var_def out_var_def)?, 
                             ((simp only: lens_indep_all_simplifier))?,
                              clarsimp?,
                             (vcg_upreds_post_processing)?,
                              vcg_hol_post_processing(*TODO: ADD SOLVING STEP HERE*))

method vcg_can_do_symbolic_execution =
   (match conclusion in "DONT_TOUCH _" \<Rightarrow> fail
    \<bar>   "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P"\<Rightarrow> fail
    \<bar>   "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>G"\<Rightarrow> fail
    \<bar>   "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u"\<Rightarrow> fail
    \<bar>   "_" \<Rightarrow> succeed) (*TODO: THIS case should be a case match with VC construct*)

method symbolic_execution =
       (vcg_can_do_symbolic_execution,(subst Product_Type.prod.case)?, post_processing)?
  

subsection\<open>VCG Goal Beautify Tactics\<close>    

text \<open>Tactics and methods in the sequel are used to beautify the goals before presenting them to the user.
      Namely, after the application of post-processing a lot of semantic machinery and debugging information
      remains existing in the different proof goals. The methods below clean it up since these 
      assumptions are useless at this point.\<close>
  
definition "LVAR L x = True"  
  
lemma GET_REMOVER: obtains x where "lens_get L s = x" "LVAR L x" unfolding LVAR_def by blast
   
method_setup get_disambiguator = \<open>Scan.succeed (fn ctxt => SIMPLE_METHOD' (fn i => fn st => 
  if i > Thm.nprems_of st then all_tac st
  else
    let 
      fun cnv (Const (@{const_name Trueprop},_)$ (Const (@{const_name LVAR},_) $(Free (name,_)) $ Bound i)) = SOME (name,i)
        | cnv _ = NONE
      val (_, _, Bi, _) = Thm.dest_state (st, i)
      val free_names = Term.fold_aterms (fn Free (x, _) => insert (op =) x | _ => I) Bi [];
      val newnames = Logic.get_goal (Thm.prop_of st) i 
        |> Logic.strip_assums_hyp 
        |> map_filter cnv
        |> sort (apply2 snd #> int_ord #> rev_order)
        |> (fn newnames =>
             fold_map (fn (name, i) => fn free_names =>
                         let fun aux n =
                               if List.exists (fn n0 => n0 = n) free_names then aux (n ^ "'") else n
                             val name = aux name
                         in ((name, i), name :: free_names) end)
                      newnames
                      free_names)
        |> #1
        |> map fst
    in 
      rename_tac newnames i st 
    end))\<close>  
    
(*Frederic's method for removing get functions from the goal.
  Modified by Yakoub.*)    
  
method get_remover =
   (match conclusion in 
          "_ (put\<^bsub>x\<^esub> A (get\<^bsub>x\<^esub> _))" for x A \<Rightarrow> \<open>fail\<close>  --{*In case that a proof engineer forget to specify LENS WELL BEHAVED assumptions*}
         \<bar>"_ (put\<^bsub>x\<^esub> (put\<^bsub>x\<^esub> A _ ) _)" for x A \<Rightarrow> \<open>fail\<close>  --{*In case that a proof engineer forget to specify LENS WELL BEHAVED assumptions*}
         \<bar>"_ (get\<^bsub>_\<^esub> (put\<^bsub>x\<^esub> A _))" for x A \<Rightarrow> \<open>fail\<close>  --{*In case that a proof engineer forget to specify LENS INDEP assumptions*}
         \<bar>"_ (get\<^bsub>x\<^esub> (put\<^bsub>_\<^esub> A _))" for x A \<Rightarrow> \<open>fail\<close> --{*In case that a proof engineer forget to specify LENS INDEP assumptions*}
         \<bar>"_ (get\<^bsub>x\<^esub> A)" for x A \<Rightarrow> \<open>rule GET_REMOVER[where L= x and s= A], simp only:\<close>)+,
  get_disambiguator,
  vcg_elim_determ thin_rl[of "lens_get _ _ = _"] thin_rl[of "LVAR _ _"]
      
named_theorems beautify_thms     
lemma thin_vwb_lens[beautify_thms]: "vwb_lens l \<Longrightarrow> P \<Longrightarrow> P" . 
lemma thin_weak_lens[beautify_thms]: "weak_lens l \<Longrightarrow> P \<Longrightarrow> P" .    
lemma [beautify_thms]: "\<not> ief_lens i \<Longrightarrow> P \<Longrightarrow> P" .  
lemma [beautify_thms]: "i\<bowtie>j \<Longrightarrow> P \<Longrightarrow> P" .  
lemma [beautify_thms]: "i\<noteq>(j::_\<Longrightarrow>_) \<Longrightarrow> P \<Longrightarrow> P" .
lemma [beautify_thms]: "i\<noteq>(j::_\<Longrightarrow>_) \<longrightarrow> i \<bowtie> j \<Longrightarrow> P \<Longrightarrow> P" .    
lemma [beautify_thms]: "get\<^bsub>i\<^esub> A = x \<Longrightarrow> P \<Longrightarrow> P" .


section \<open>Debugging framework for the VCG\<close> 

subsection \<open>VCG UI\<close>

method used_formalism_print_method = formalism_used_is \<open>Verification Condition Generator (VCG) and Symbolic Execution (SE)\<close>
method reasoning_print_method = reasoning \<open>Verification Condition Generator (VCG) and Symbolic Execution (SE)\<close>
method wp_mode_print_method   = computation_mode \<open>Weakest Precondition\<close>
method sp_mode_print_method   = computation_mode \<open>Strongest Postcondition\<close>
method total_correctness_print_method   = correctness   \<open>Total Correctness\<close>
method partial_correctness_print_method = correctness   \<open>Partial Correctness\<close>

subsection \<open>VCG pre-processing\<close>

method vcg_pre_processing_print_method methods vcg_computing_mode_print_method =
   (match conclusion 
         in "\<lbrace>\<lambda>_. case _ of _ \<Rightarrow> REQUIRES\<rbrace>PROGRAM\<lbrace>\<lambda>_. case _ of _ \<Rightarrow> ENSURES\<rbrace>\<^sub>G" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>used_formalism_print_method,
              reasoning_print_method, 
              vcg_computing_mode_print_method,
              total_correctness_print_method,
              print_term \<open>PROGRAM\<close>, 
              print_term \<open>REQUIRES\<close>, 
              print_term \<open>ENSURES\<close>\<close>)
 |(match conclusion
         in "\<lbrace>REQUIRES\<rbrace>PROGRAM\<lbrace>ENSURES\<rbrace>\<^sub>P" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>used_formalism_print_method,
              reasoning_print_method, 
              vcg_computing_mode_print_method,
              total_correctness_print_method,
              print_term \<open>PROGRAM\<close>, 
              print_term \<open>REQUIRES\<close>, 
              print_term \<open>ENSURES\<close>\<close>)
 
 |(match conclusion 
   in "\<lbrace>REQUIRES\<rbrace>PROGRAM\<lbrace>ENSURES\<rbrace>\<^sub>u" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>used_formalism_print_method,
              reasoning_print_method, 
              vcg_computing_mode_print_method,
              partial_correctness_print_method, 
              print_term \<open>PROGRAM\<close>, 
              print_term \<open>REQUIRES\<close>, 
              print_term \<open>ENSURES\<close>\<close>)

 |(match conclusion 
   in "\<lbrace>REQUIRES\<rbrace>PROGRAM\<lbrace>ENSURES\<rbrace>\<^sub>G" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>used_formalism_print_method,
              reasoning_print_method, 
              vcg_computing_mode_print_method,
              total_correctness_print_method,
              print_term \<open>PROGRAM\<close>, 
              print_term \<open>REQUIRES\<close>, 
              print_term \<open>ENSURES\<close>\<close>)
 |(match conclusion 
   in "_" \<Rightarrow> \<open>print_error \<open>Syntax error!\<close>,fail\<close>  -- \<open>fail for any other goal\<close>)

method vcg_pre_first_goal_processing_trace =
  (match conclusion 
         in "\<lbrace>REQUIRES\<rbrace>PROGRAM\<lbrace>ENSURES\<rbrace>\<^sub>P" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>print_string \<open>================================================\<close>,
              print_string \<open>====Starting Preprocessing of the First Goal====\<close>, 
              print_string \<open>================================================\<close>, 
              print_string \<open>++Application of the SIMP rule:\<close>,
              print_fact seqr_assoc[symmetric],
              (simp only: seqr_assoc[symmetric])?\<close>)
 |(match conclusion 
   in "\<lbrace>REQUIRES\<rbrace>PROGRAM\<lbrace>ENSURES\<rbrace>\<^sub>u" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>print_string \<open>================================================\<close>,
              print_string \<open>====Starting Preprocessing of the First Goal====\<close>, 
              print_string \<open>================================================\<close>,
              print_string \<open>++Application of the SIMP rule:\<close>,
              print_fact seqr_assoc[symmetric],
              (simp only: seqr_assoc[symmetric])?\<close>)
 |(match conclusion 
   in "\<lbrace>REQUIRES\<rbrace>PROGRAM\<lbrace>ENSURES\<rbrace>\<^sub>G" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>print_string \<open>================================================\<close>,
              print_string \<open>====Starting Preprocessing of the First Goal====\<close>, 
              print_string \<open>================================================\<close>,
              print_string \<open>++Application of the SIMP rule:\<close>,
              print_fact seqr_assoc[symmetric],
              (simp only: seqr_assoc[symmetric])?, 
              print_string \<open>++Application of the SIMP rule:\<close>,
              print_fact Product_Type.split_paired_All,                 
              print_string \<open>++Application of the SIMP rule:\<close>,
              print_fact Product_Type.prod.case,
              (simp only:Product_Type.split_paired_All Product_Type.prod.case)?, 
              print_string \<open>++Application of the INTRO rule:\<close>,
              print_fact allI,
              ((rule allI)+)?\<close>)
 |(match conclusion 
   in "_" \<Rightarrow> \<open>print_error \<open>Syntax ERROR!\<close>,fail\<close>  -- \<open>fail for any other goal\<close>)


method precondition_strengthening_trace =
  (match conclusion 
         in "\<lbrace>REQUIRES\<rbrace>PROGRAM\<lbrace>ENSURES\<rbrace>\<^sub>P" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>print_string \<open>++Application of the INTRO rule:\<close>,
              print_fact pre_weak_prog_hoare,
              rule pre_weak_prog_hoare,
              print_string \<open>==================================================\<close>,
              print_string \<open>====End of the Preprocessing of the First Goal====\<close>,
              print_string \<open>==================================================\<close>\<close>)
 |(match conclusion 
   in "\<lbrace>REQUIRES\<rbrace>PROGRAM\<lbrace>ENSURES\<rbrace>\<^sub>u" for REQUIRES PROGRAM ENSURES \<Rightarrow> 
             \<open>print_string \<open>++Application of the INTRO rule:\<close>,
              print_fact pre_str_hoare_rel,
              rule pre_str_hoare_rel,
              print_string \<open>==================================================\<close>,
              print_string \<open>====End of the Preprocessing of the First Goal====\<close>,
              print_string \<open>==================================================\<close>\<close>)
 |(match conclusion 
   in "_" \<Rightarrow> \<open>print_error \<open>Syntax ERROR!\<close>,fail,
              print_string \<open>==================================================\<close>,
              print_string \<open>====End of the Preprocessing of the First Goal====\<close>,
              print_string \<open>==================================================\<close>\<close>  
              -- \<open>fail for any other goal\<close>)


method hoare_wp_vcg_preprocessing_trace =
       vcg_pre_first_goal_processing_trace, 
       precondition_strengthening_trace

method vcg_can_defer_trace =
  (match conclusion 
      in "DEFERRED _" \<Rightarrow> \<open>print_string\<open>++++End of VCG DEFFER Goal: VCG Refused to DEFFER an already DEFFERED goal!++++\<close>,fail\<close> 
       \<bar>
         "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>print_string\<open>-End of VCG DEFFER Goal: VCG Refused to DEFFER a goal in form of Hoare Triple!\<close>,fail\<close>
       \<bar>
         "\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>print_string\<open>-End of VCG DEFFER Goal: VCG Refused to DEFFER a goal in form of Hoare Triple!\<close>,fail\<close> 
       \<bar> 
         "_" \<Rightarrow> \<open>succeed\<close>)
       
method vcg_defer_trace = (print_string \<open>++++Start of VCG DEFFER Goal++++\<close>, 
                          vcg_can_defer_trace, 
                          print_fact DEFERREDD, rule DEFERREDD, 
                          tactic \<open>FIRSTGOAL defer_tac\<close>,
                          print_string \<open>++++End of VCG DEFFER Goal: VCG applied DEFFER++++\<close>) 

method hoare_wp_rule_apply_trace =
 --\<open>WP Matching rules for total correctness\<close>
  ((match conclusion in "\<lbrace>_\<rbrace>\<langle>\<sigma>\<rangle>\<^sub>p\<lbrace>Q\<rbrace>\<^sub>P" for \<sigma> Q \<Rightarrow> 
     \<open>print_fact assigns_prog_hoare_backward[where \<sigma> = \<sigma> and Q = Q]\<close>), 
     rule assigns_prog_hoare_backward)
 (*CONTINUE FROM HERE!! ADD the print message for the rest of the rules!! *)
 |((match conclusion in "\<lbrace>_\<rbrace>SKIP\<lbrace>Q\<rbrace>\<^sub>P"for Q \<Rightarrow> 
     \<open>print_fact skip_prog_hoare[where P = Q]\<close>), rule skip_prog_hoare)
 |((match conclusion in "\<lbrace>_\<rbrace>_;_\<lbrace>Q\<rbrace>\<^sub>P"for Q \<Rightarrow> 
     \<open>print_fact seq_prog_hoare[where Q = Q]\<close>), rule seq_prog_hoare)
 |((match conclusion in "\<lbrace>_\<rbrace>IF b THEN _ ELSE _ FI\<lbrace>Q\<rbrace>\<^sub>P"for b Q \<Rightarrow> 
      \<open>print_fact cond_prog_hoare_wp [where b = b and Q = Q]\<close>), rule cond_prog_hoare_wp)
 (*Scoping programming construct!*)
 |((match conclusion in "\<lbrace>_\<rbrace>g:[_]\<lbrace>Q\<rbrace>\<^sub>P" for g Q \<Rightarrow> 
      \<open>print_string \<open><||||>--FRAME RULE: VCG starts applying FRAME RULE--<||||>\<close>,
       print_fact antiframe_prog_hoare_wp_weak[where g = fst_lens and l= snd_lens]\<close>), 
       (rule antiframe_prog_hoare_wp_weak[where g= fst_lens and l= snd_lens]);
       (print_string \<open><||||>--FRAME RULE: VCG ends applying FRAME RULE--<||||>\<close>, 
        print_string \<open><<--FRAME RULE: VCG starts the TAG process for the VCs of the frame rule-->>\<close>, 
        (rule FRAME_TRIPLE_E),
        print_string \<open><<--FRAME RULE: VCG ends the TAG process for the VCs of the frame rule-->>\<close>))
 (*Special handling for the VCs generated from scoping construct*)
 |((match conclusion in "FRAME_TRIPLE (g:[_] = _)"for g \<Rightarrow> \<open>succeed\<close>), 
                         print_string \<open><<--FRAME RULE: VCG starts changing the TAG for the MODIFIES condition-->>\<close>,
                         (simp only: FRAME_TRIPLE_def), (rule DONT_TOUCH_E),                         
                         print_string \<open><<--FRAME RULE: VCG ends changing the TAG for the MODIFIES condition-->>\<close>)
 |((match conclusion in "FRAME_TRIPLE \<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>succeed\<close>),
                         print_string \<open><||||>--FRAME RULE: VCG starts applying already PROVED SPECS--<||||>\<close>,
                         (simp only: FRAME_TRIPLE_def, rule vcg_proved_specs),                         
                         print_string \<open><||||>--FRAME RULE:VCG ends the application of already PROVED SPECS--<||||>\<close>)
 |((match conclusion in "FRAME_TRIPLE _" \<Rightarrow> \<open>succeed\<close>),  
                         print_string \<open><<--FRAME RULE: VCG starts removing the TAG for the normale VCs-->>\<close>,
                         (simp only: FRAME_TRIPLE_def),                         
                         print_string \<open><<--FRAME RULE: VCG ends removing the TAG for the normale VCs-->>\<close>)
 (*Annotated loop patterns for total correctness*)
 |((match conclusion in "\<lbrace>_\<rbrace>INVAR invar VRT vari WHILE b DO _ OD\<lbrace>Q\<rbrace>\<^sub>P" for invar vari b Q \<Rightarrow> \<open>succeed\<close>), 
                        (print_fact while_invr_vrt_hoare_prog_wp[where st' = "&\<Sigma>" and R = measure and 
                                                                 Q = Q and b = b and invar = invar and 
                                                                 expression = vari and body = body],
                         rule while_invr_vrt_hoare_prog_wp[where R= measure and st' = "&\<Sigma>"]))
 |((match conclusion in 
   "\<lbrace>_\<rbrace>FOR (init,b,incr) INVAR invar VRT vari DO body OD\<lbrace>_\<rbrace>\<^sub>P" for init b incr invar vari body \<Rightarrow> \<open>succeed\<close>), 
                         print_fact 
                         for_lfp_invar_vrt_prog_def_alt[where init = init and b = b and 
                                                              incr = incr and body = body and 
                                                              vari= vari and invar = invar],
                         subst for_lfp_invar_vrt_prog_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>DO body INVAR invar VRT vari WHILE b OD\<lbrace>_\<rbrace>\<^sub>P" for body invar vari b  \<Rightarrow> \<open>succeed\<close>), 
                          print_fact 
                          do_while_invar_vrt_lfp_prog_def_alt[where b = b and body = body and 
                                                                    vari = vari and invar = invar],
                         subst do_while_invar_vrt_lfp_prog_def_alt)
 |((match conclusion in 
   "\<lbrace>_\<rbrace>FROM init INVAR invar VRT vari UNTIL b DO body OD\<lbrace>_\<rbrace>\<^sub>P" for init invar vari b body \<Rightarrow> \<open>succeed\<close>), 
                         print_fact from_until_lfp_invr_vrt_prog_def_alt[where b = b and body = body and 
                                                                               vari = vari and invar = invar],
                         subst from_until_lfp_invr_vrt_prog_def_alt)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 | (match conclusion in "\<lbrace>_\<rbrace>INVAR _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant is missing\<close>
 | (match conclusion in "\<lbrace>_\<rbrace>FOR (_,_,_)INVAR _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant is missing\<close>
 | (match conclusion in "\<lbrace>_\<rbrace>DO _ INVAR _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant is missing\<close>
  | (match conclusion in "\<lbrace>_\<rbrace>FROM _ INVAR _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant is missing\<close>
  | (match conclusion in "\<lbrace>_\<rbrace>WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant and invariant are missing\<close>
  | (match conclusion in "\<lbrace>_\<rbrace>FOR (_,_,_)DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant and invariant are missing\<close>
  | (match conclusion in "\<lbrace>_\<rbrace>DO _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant and invariant are missing\<close>
  | (match conclusion in "\<lbrace>_\<rbrace>FROM _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant and invariant are missing\<close>
 (*Annotated SCOPED loop patterns for total correctness*)
  |((match conclusion in 
    "\<lbrace>_\<rbrace>IN G INVAR invar VRT vari WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P" for G invar vari b body Q \<Rightarrow> \<open>succeed\<close>), 
                         rule while_scp_invr_vrt_prog_hoare_wp[where st' = "&\<Sigma>" and R = measure and 
                                                                     Q = Q and invar = invar and
                                                                     expression = vari and
                                                                     b = b and 
                                                                     g = fst_lens and l = snd_lens]; 
                         rule FRAME_TRIPLE_E) 
 |((match conclusion in 
   "\<lbrace>_\<rbrace>IN G FOR (init,b,incr) INVAR invar VRT vari DO body OD\<lbrace>_\<rbrace>\<^sub>P" for G init b incr invar vari body \<Rightarrow> \<open>succeed\<close>), 
                         print_fact 
                         for_lfp_scp_invr_vrt_prog_def_alt [where G = "G;\<^sub>L fst_lens" and 
                                                            init = init and b = b and incr = incr and 
                                                            body = body and vari= vari and invar = invar],
                         subst for_lfp_scp_invr_vrt_prog_def_alt)
 |((match conclusion in 
   "\<lbrace>_\<rbrace>IN G DO body INVAR invar VRT vari WHILE b OD\<lbrace>_\<rbrace>\<^sub>P" for G body invar vari b \<Rightarrow> \<open>succeed\<close>), 
                         print_fact 
                         for_lfp_scp_invr_vrt_prog_def_alt[where G = "G;\<^sub>L fst_lens" and 
                                                           b = b and body = body and 
                                                           vari = vari and invar = invar],
                         subst do_while_lfp_scp_invr_vrt_prog_def_alt)
 |((match conclusion in 
   "\<lbrace>_\<rbrace>IN G FROM init INVAR invar VRT vari UNTIL b DO body OD\<lbrace>_\<rbrace>\<^sub>P" for G init invar vari b body  \<Rightarrow> \<open>succeed\<close>), 
                         print_fact 
                         from_until_lfp_scp_invr_vrt_prog_def_alt[where G = "G;\<^sub>L fst_lens" and 
                                                                  b = b and body = body and 
                                                                  vari = vari and invar = invar and
                                                                  init = init],
                         subst from_until_lfp_scp_invr_vrt_prog_def_alt)
 (*Non-annotated SCOPED loop patterns fail the VCG... which requires a proof annotation *)
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ INVAR _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant is missing\<close> 
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ FOR (_,_,_)INVAR _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant is missing\<close>
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ DO _ INVAR _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant is missing\<close>
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ FROM _ INVAR _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant is missing\<close>
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ WHILE _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant and invariant are missing\<close> 
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ FOR (_,_,_)DO _ OD\<lbrace>_\<rbrace>\<^sub>P"     \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant and invariant are missing\<close>
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ DO _ WHILE _ OD\<lbrace>_\<rbrace>\<^sub>P"        \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant and invariant are missing\<close>
 | (match conclusion in "\<lbrace>_\<rbrace>IN _ FROM _ UNTIL _ DO _ OD\<lbrace>_\<rbrace>\<^sub>P" \<Rightarrow> \<open>fail\<close>),
    print_string \<open>Non-Annotated loop: This loop is not annotated a variant and invariant are missing\<close>
 --\<open>WP Matching rules for partial correctness\<close>
 |((match conclusion in "\<lbrace>_\<rbrace>\<langle>\<sigma>\<rangle>\<^sub>a\<lbrace>Q\<rbrace>\<^sub>u" for \<sigma> Q  \<Rightarrow> \<open>succeed\<close>), 
   print_fact assigns_backward_hoare_rel[where \<sigma> = \<sigma> and p = Q],
   rule assigns_backward_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>SKIP\<^sub>r\<lbrace>Q\<rbrace>\<^sub>u" for Q \<Rightarrow> \<open>succeed\<close>), print_fact skip_hoare_rel[where p = Q], rule skip_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>_;;_\<lbrace>Q\<rbrace>\<^sub>u"for Q  \<Rightarrow> \<open>succeed\<close>), print_fact seq_hoare_rel[where r = Q], rule seq_hoare_rel)
 |((match conclusion in "\<lbrace>_\<rbrace>if\<^sub>u b then _ else _\<lbrace>Q\<rbrace>\<^sub>u" for b Q \<Rightarrow> \<open>succeed\<close>), 
                        print_fact cond_hoare_rel_wp [where b = b and q = Q],rule cond_hoare_rel_wp)
  (*CONTINUE FROM HERE!! ADDING DEBUG MESSAGES TO HOARE CONSTRUCTS IN PARTIAL CORRECTNESS!!*)
  (*Scoping programming construct for relations!*)
 |((match conclusion in "\<lbrace>_\<rbrace>antiframe _ _\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), rule cond_hoare_rel_wp; rule FRAME_TRIPLE_E)
 (*Special handling for the VCs generated from scoping construct*)
 |((match conclusion in "FRAME_TRIPLE (antiframe _ _ = _)" \<Rightarrow> \<open>succeed\<close>), 
                         (simp only: FRAME_TRIPLE_def), (rule DONT_TOUCH_E))
 |((match conclusion in "FRAME_TRIPLE \<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>),
                         (simp only: FRAME_TRIPLE_def, rule vcg_proved_specs))
 |((match conclusion in "FRAME_TRIPLE _" \<Rightarrow> \<open>succeed\<close>), (simp only: FRAME_TRIPLE_def))
 (*Annotated loop patterns for total correctness*)
 |((match conclusion in "\<lbrace>_\<rbrace>invr _ while\<^sup>\<top> _ do _ od\<lbrace>_\<rbrace>\<^sub>u"  \<Rightarrow> \<open>succeed\<close>), 
                         rule while_gfp_invr_hoare_rel_wp_partial)
 |((match conclusion in "\<lbrace>_\<rbrace>do _ invr _ while\<^sup>\<top> _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst do_while_invar_vrt_gfp_rel_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>for\<^sup>\<top> (_,_,_) invr _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst for_gfp_invar_vrt_rel_def_alt)
 |((match conclusion in "\<lbrace>_\<rbrace>from\<^sup>\<top> _ invr _ until _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>succeed\<close>), 
                         subst from_until_gfp_invr_vrt_rel_def_alt)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 |(match conclusion in "\<lbrace>_\<rbrace>while\<^sup>\<top> _ do _ od\<lbrace>_\<rbrace>\<^sub>u"        \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>do _ while\<^sup>\<top> _ od\<lbrace>_\<rbrace>\<^sub>u"        \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>for\<^sup>\<top> (_,_,_) do _ od\<lbrace>_\<rbrace>\<^sub>u"    \<Rightarrow> \<open>fail\<close>)
 |(match conclusion in "\<lbrace>_\<rbrace>from\<^sup>\<top> _ until _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>fail\<close>)
 (*Non-annotated loop patterns fail the VCG... which requires a proof annotation *)
 |(match conclusion in "\<lbrace>_\<rbrace>invr _ vrt _ while\<^sub>\<bottom> _ do _ od\<lbrace>_\<rbrace>\<^sub>u" \<Rightarrow> \<open>fail\<close>)
--{*For any goal which is not a Hoare triple*}
 |((match conclusion in "_" \<Rightarrow> \<open>succeed\<close>), vcg_defer)


method wp_trace =
       vcg_pre_processing_print_method wp_mode_print_method-- \<open>Printing useful information on the VCG\<close>,
       hoare_wp_vcg_preprocessing_trace, 
       print_string \<open>+++++++++++++++++++++++++++++++++++++++++++++++\<close>,
       print_string \<open>++++Begin Application of Hoare Logic Rules+++++\<close>,
       print_string \<open>+++++++++++++++++++++++++++++++++++++++++++++++\<close>,
       hoare_wp_rule_apply_trace+, 
       print_string \<open>+++++++++++++++++++++++++++++++++++++++++++++++\<close>,
       print_string \<open>++++End of Application of Hoare Logic Rules++++\<close>,
       print_string \<open>+++++++++++++++++++++++++++++++++++++++++++++++\<close>,
       (unfold DEFERRED_def)

subsection \<open>Post processing for debugging mode\<close>  
          
method vcg_upreds_post_processing_debugger = 
       (vwb_lens_debugger(*TODO: if this fails add a debug message here*)
        |wf_debugger(*TODO: if this fails add a debug message here*)
        |subst_debugger(*TODO: if this fails add a debug message here*)
        |subst_lookup_debugger(*TODO: if this fails add a debug message here*))



method vcg methods vcg_reasoning_method =
  vcg_reasoning_method ; symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?

(*method vcg_trace methods vcg_reasoning_method =
  (vcg_reasoning_method) ; symbolic_execution; get_remover?; (vcg_elim_determ beautify_thms)?*)

end

