(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)
(******************************************************************************
 * Authors: Josh Bockenek, Yakoub Nemouchi.
 ******************************************************************************)

theory insertion_sort_lists

  imports  "../Backend/VCG/vcg"
          "../Midend-IVL/Isabelle-UTP-Extended/utils/vcg_helpers" 
begin
  
section \<open>Boilerplate for insertion sort\<close>  
named_theorems vcg_simps and vcg_dests
  
definition \<open>outer_invr i array old_array \<equiv>
  mset array = mset old_array
\<and> sorted (take i array) (* everything up to i-1 is sorted *)
\<close>
abbreviation \<open>outer_invr\<^sub>u \<equiv> trop outer_invr\<close>

definition \<open>inner_invr i j array old_array \<equiv>
  i < length array
\<and> i \<ge> j
\<and> mset array = mset old_array
\<and> (let xs\<^sub>1 = take j array; x = array!j; xs\<^sub>2 = drop (Suc j) (take (Suc i) array) in
  sorted (xs\<^sub>1 @ xs\<^sub>2) \<and> (\<forall>y \<in> set xs\<^sub>2. x < y))
\<close>
abbreviation \<open>inner_invr\<^sub>u \<equiv> qtop inner_invr\<close>

lemma outer_invr_init[vcg_simps]:
  assumes \<open>mset array = mset old_array\<close>
  shows \<open>outer_invr (Suc 0) array old_array\<close>
  using assms unfolding outer_invr_def
  by (metis sorted_single sorted_take take_0 take_Suc)

lemma outer_invr_step[vcg_simps]:
  assumes \<open>inner_invr i j array old_array\<close>
      and \<open>j = 0 \<or> \<not> array ! j < array ! (j - Suc 0)\<close>
  shows \<open>outer_invr (Suc i) array old_array\<close>
  using assms unfolding inner_invr_def outer_invr_def Let_def
  apply (erule_tac disjE1)
   apply auto
   apply (metis Cons_nth_drop_Suc Suc_leI drop_0 length_greater_0_conv length_take less_imp_le
      min.absorb2 nth_take sorted.simps)
  apply (drule (1) insert_with_sorted)
    apply auto
   apply (smt One_nat_def diff_Suc_less last_conv_nth le_less_trans length_take list.size(3)
      min.absorb2 not_le_imp_less not_less_iff_gr_or_eq nth_take)
  using take_take[symmetric, where n = j and m = \<open>Suc i\<close> and xs = array]
    id_take_nth_drop[where xs = \<open>take (Suc i) array\<close> and i = j]
  by (auto simp: min_def)

lemma outer_invr_final[vcg_dests]:
  assumes \<open>outer_invr i array old_array\<close>
      and \<open>\<not> i < length array\<close>
  shows \<open>mset array = mset old_array\<close>
    and \<open>sorted array\<close>
  using assms unfolding outer_invr_def
  by auto

lemma inner_invr_init[vcg_simps]:
  assumes \<open>outer_invr i array old_array\<close>
      and \<open>j = i\<close>
      and \<open>i < length array\<close>
    shows \<open>inner_invr i j array old_array\<close>
  using assms unfolding outer_invr_def inner_invr_def
  by auto

lemma inner_invr_step[vcg_simps]:
  assumes \<open>inner_invr i j array old_array\<close>
      and \<open>j > 0\<close>
      and \<open>array!(j - Suc 0) > array!j\<close>
  shows \<open>inner_invr i (j - Suc 0) (swap_at j array) old_array\<close>
  using assms unfolding inner_invr_def Let_def
  apply clarsimp
  apply (safe; (simp add: swap_at_def; fail)?)
proof goal_cases
  case 1
  then show ?case by (simp add: swap_at_def Multiset.mset_swap)
next
  assume 2: \<open>0 < j\<close>
    \<open>array!j < array!(j - Suc 0)\<close>
    \<open>i < length array\<close>
    \<open>j \<le> i\<close>
    \<open>mset array = mset old_array\<close>
    \<open>sorted (take j array @ drop (Suc j) (take (Suc i) array))\<close>
    \<open>\<forall>x \<in> set (drop (Suc j) (take (Suc i) array)). array!j < x\<close>
  define xs\<^sub>1 where \<open>xs\<^sub>1 = take j array\<close>
  define xs\<^sub>2 where \<open>xs\<^sub>2 = drop (Suc j) (take (Suc i) array)\<close>
  define x where \<open>x = array ! j\<close>
  obtain xs\<^sub>1' y where xs_last: \<open>xs\<^sub>1 = xs\<^sub>1' @ [y]\<close>
    unfolding xs\<^sub>1_def using 2
    by (metis Suc_pred diff_le_self le_less_trans take_hd_drop)
  have xs_butlast: \<open>xs\<^sub>1' = take (j - Suc 0) array\<close>
    by (smt "2"(3) "2"(4) Suc_pred append1_eq_conv assms(2) diff_le_self le_less_trans take_hd_drop
        xs\<^sub>1_def xs_last)
  have y: \<open>y = array ! (j - Suc 0)\<close>
    by (metis (no_types, lifting) "2"(3) "2"(4) Cons_nth_drop_Suc One_nat_def Suc_pred assms(2)
        diff_le_self le_less_trans list.sel(1) nth_append_length take_hd_drop xs\<^sub>1_def xs_butlast
        xs_last)
  have xs\<^sub>1'_is_aaker: \<open>xs\<^sub>1' = take (j - Suc 0) (swap_at j array)\<close>
    by (simp add: swap_at_def xs_butlast)
  have y_concat_xs\<^sub>2: \<open>y # xs\<^sub>2 = drop j (take (Suc i) (swap_at j array))\<close>
    using \<open>j > 0\<close>
    apply (auto simp: swap_at_def drop_take list_update_swap)
    by (smt "2"(3) "2"(4) Cons_nth_drop_Suc Suc_diff_Suc drop_take drop_update_cancel le_less_trans
        length_list_update lessI nth_list_update_eq take_Suc_Cons xs\<^sub>2_def y)
  from 2 show \<open>sorted (take (j - Suc 0) (swap_at j array) @ drop j (take (Suc i) (swap_at j array)))\<close>
    by (fold xs\<^sub>1_def xs\<^sub>2_def xs_butlast xs\<^sub>1'_is_aaker y_concat_xs\<^sub>2) (simp add: xs_last)
  {
    fix x
    assume \<open>x \<in> set (drop j (take (Suc i) (swap_at j array)))\<close>
    show \<open>swap_at j array!(j - Suc 0) < x\<close>
      by (smt "2"(2) "2"(3) "2"(4) "2"(7) One_nat_def \<open>x \<in> set (drop j (take (Suc i) (swap_at j
          array)))\<close> diff_le_self le_less_trans length_list_update nth_list_update_eq set_ConsD
          swap_at_def xs\<^sub>2_def y y_concat_xs\<^sub>2)
  }
qed

section \<open>setup and makeup!\<close>    


sledgehammer_params[stop_on_first,parallel_subgoals, join_subgoals]
  
no_adhoc_overloading'_all

section \<open>VCG Helpers\<close>

section \<open>Insertion sort partial correctness\<close>

lemma  insertion_sort_partial:
  assumes \<open>i\<bowtie> j\<close> \<open>i \<bowtie> array\<close> \<open>j \<bowtie> array\<close>
  and \<open>vwb_lens array\<close>  \<open>vwb_lens i\<close> \<open>vwb_lens j\<close>
  shows 
"\<lbrace>&array =\<^sub>u \<guillemotleft>old_array\<guillemotright>\<rbrace>
    assign_r i  \<guillemotleft>1\<guillemotright>;;
    invr  outer_invr\<^sub>u (&i) (&array) \<guillemotleft>old_array\<guillemotright>
    while\<^sup>\<top>  &i <\<^sub>u uop length (&array)
      do
         j :== (&i);;
        (invr inner_invr\<^sub>u (&i) (&j) (&array) \<guillemotleft>old_array\<guillemotright>
         while\<^sup>\<top> &j >\<^sub>u \<guillemotleft>0\<guillemotright> \<and> (bop nth (&array) (&j - \<guillemotleft>1\<guillemotright>)) >\<^sub>u (bop nth (&array) (&j))
         do
          array :== (swap_at\<^sub>u (&j) (&array));;
           j :== (&j -\<guillemotleft>1\<guillemotright>)
        od);;  
         i :== (&i + \<guillemotleft>1\<guillemotright>) 
      od 
 \<lbrace>mset\<^sub>u(&array) =\<^sub>u mset\<^sub>u(\<guillemotleft>old_array\<guillemotright>) \<and> sorted\<^sub>u(&array)\<rbrace>\<^sub>u" 
  apply (insert assms)
  apply (vcg sp)
        apply (simp_all add: vcg_simps)
   apply (auto simp: outer_invr_final)
  done

lemma  insertion_sort_partial_wp:
  assumes \<open>i\<bowtie> j\<close> \<open>i \<bowtie> array\<close> \<open>j \<bowtie> array\<close>
  and \<open>vwb_lens array\<close>  \<open>vwb_lens i\<close> \<open>vwb_lens j\<close>
  shows 
"\<lbrace>&array =\<^sub>u \<guillemotleft>old_array\<guillemotright>\<rbrace>
    assign_r i  \<guillemotleft>1\<guillemotright>;;
    invr  outer_invr\<^sub>u (&i) (&array) \<guillemotleft>old_array\<guillemotright>
    while\<^sup>\<top>  &i <\<^sub>u uop length (&array)
      do
         j :== (&i);;
        (invr inner_invr\<^sub>u (&i) (&j) (&array) \<guillemotleft>old_array\<guillemotright>
         while\<^sup>\<top> &j >\<^sub>u \<guillemotleft>0\<guillemotright> \<and> (bop nth (&array) (&j - \<guillemotleft>1\<guillemotright>)) >\<^sub>u (bop nth (&array) (&j))
         do
          array :== (swap_at\<^sub>u (&j) (&array));;
           j :== (&j -\<guillemotleft>1\<guillemotright>)
        od);;  
         i :== (&i + \<guillemotleft>1\<guillemotright>) 
      od 
 \<lbrace>mset\<^sub>u(&array) =\<^sub>u mset\<^sub>u(\<guillemotleft>old_array\<guillemotright>) \<and> sorted\<^sub>u(&array)\<rbrace>\<^sub>u" 
  apply (insert assms)
  apply (vcg wp_trace)
        apply (simp_all add: vcg_simps)
   apply (auto simp: outer_invr_final)
  done

(* TODO @Frederic; Fix please*)

(*program_spec  insertion_sort_partial_interface:
  assumes_utp"&array =\<^sub>u \<guillemotleft>old_array\<guillemotright>"
  ensures_utp  "mset\<^sub>u(&array) =\<^sub>u mset\<^sub>u(\<guillemotleft>old_array\<guillemotright>) \<and> sorted\<^sub>u(&array)"

  prog_utp 
    "assign_r i  \<guillemotleft>1\<guillemotright>;;
  invr  outer_invr\<^sub>u (&i) (&array) \<guillemotleft>old_array\<guillemotright>
  while\<^sup>\<top>  &i <\<^sub>u uop length (&array)
 do
    assign_r j (&i);;
    ( invr inner_invr\<^sub>u (&i) (&j) (&array) \<guillemotleft>old_array\<guillemotright>
     while\<^sup>\<top> &j >\<^sub>u \<guillemotleft>0\<guillemotright> \<and> (bop nth (&array) (&j - \<guillemotleft>1\<guillemotright>)) >\<^sub>u (bop nth (&array) (&j))
     do
      assign_r array (swap_at\<^sub>u (&j) (&array));;
      assign_r j  (&j -\<guillemotleft>1\<guillemotright>)
    od);;

    assign_r i (&i + \<guillemotleft>1\<guillemotright>) 
  od" 
  apply (insert assms)
  apply (vcg sp)
        apply (simp_all add: vcg_simps)
   apply (auto simp: outer_invr_final)
  done
   
section \<open>Insertion sort total correctness\<close>
 definition ran_sorted :: "(int \<Rightarrow> int) \<Rightarrow> int \<Rightarrow> int \<Rightarrow> bool" where
  "ran_sorted a l h \<equiv> \<forall>i\<in>{l..<h}. \<forall>j\<in>{l..<h}. i\<le>j \<longrightarrow> a i \<le> a j"
 
program_spec insertion_sort_total:
 
  assumes_utp  "&array =\<^sub>u \<guillemotleft>old_array\<guillemotright>"
  ensures_utp "mset\<^sub>u(&array) =\<^sub>u mset\<^sub>u(\<guillemotleft>old_array\<guillemotright>) \<and> sorted\<^sub>u(&array)"
  prog_utp
        "i :== \<guillemotleft>1\<guillemotright>;
         INVAR  outer_invr\<^sub>u (&i) (&array) \<guillemotleft>old_array\<guillemotright>
         VRT  \<guillemotleft>measure (Rep_uexpr (uop length (\<guillemotleft>old_array\<guillemotright>) - (&i)))\<guillemotright>
         WHILE  &i <\<^sub>u uop length (&array)
         DO
           j :== &i;
           INVAR  inner_invr\<^sub>u (&i) (&j) (&array) \<guillemotleft>old_array\<guillemotright> 
           VRT  \<guillemotleft>measure (Rep_uexpr (uop length (\<guillemotleft>old_array\<guillemotright>) - (&i)))\<guillemotright>
           WHILE  &j >\<^sub>u \<guillemotleft>0\<guillemotright> \<and> (bop nth (&array) (&j - \<guillemotleft>1\<guillemotright>)) >\<^sub>u (bop nth (&array) (&j))
           DO array :== swap_at\<^sub>u (&j) (&array);
               j :== (&j - \<guillemotleft>1\<guillemotright>)
           OD;
           i :== (&i + \<guillemotleft>1\<guillemotright>)
         OD"  
      vcg_sp
          apply (simp_all add: vcg_simps)
     apply (auto simp: outer_invr_final)

  unfolding inner_invr_def Ball_def Let_def
    apply auto[]
    
  
    
  oops*) 
    
    
end

