(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)

subsection {* Relational Hoare calculus *}

theory utp_hoare
imports "../../utp/utp_rec_total"
begin

named_theorems hoare_sp_rules and hoare_wp_rules and vcg_proved_specs

subsection {*Hoare triple definition*}


definition hoare_r :: "'\<alpha> cond \<Rightarrow> '\<alpha> hrel \<Rightarrow> '\<alpha> cond \<Rightarrow> bool" ("\<lbrace>_\<rbrace>_\<lbrace>_\<rbrace>\<^sub>u") where
"\<lbrace>p\<rbrace>Q\<lbrace>r\<rbrace>\<^sub>u = ((\<lceil>p\<rceil>\<^sub>< \<Rightarrow> \<lceil>r\<rceil>\<^sub>>) \<sqsubseteq> Q)"

declare hoare_r_def[upred_defs]
lemma assisgns_true_hoare_rel: 
  "\<lbrace>p\<rbrace> \<langle>\<sigma>\<rangle>\<^sub>a \<lbrace>true\<rbrace>\<^sub>u"
  by rel_simp

lemma skip_true_hoare_rel: 
  "\<lbrace>p\<rbrace>SKIP\<^sub>r\<lbrace>true\<rbrace>\<^sub>u"
   by rel_simp

lemma pre_false_hoare_rel: 
  "\<lbrace>false\<rbrace>C\<lbrace>q\<rbrace>\<^sub>u"
  by rel_simp

subsection {*Precondition weakening*}

lemma pre_str_hoare_rel:
  assumes "`p\<^sub>1 \<Rightarrow> p\<^sub>2`" and "\<lbrace>p\<^sub>2\<rbrace>C\<lbrace>q\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<^sub>1\<rbrace>C\<lbrace>q\<rbrace>\<^sub>u" 
  using assms  
  by rel_simp
    
subsection {*Post-condition strengthening*}
  
lemma post_weak_hoare_rel:
  assumes "\<lbrace>p\<rbrace>C\<lbrace>q\<^sub>2\<rbrace>\<^sub>u" and "`q\<^sub>2 \<Rightarrow> q\<^sub>1`"
  shows "\<lbrace>p\<rbrace>C\<lbrace>q\<^sub>1\<rbrace>\<^sub>u" 
  using assms  
  by rel_simp
    
subsection {*Consequence rule*}  
  
lemma consequence_hoare_rel:
  assumes I0': "`p \<Rightarrow> p'`" 
  assumes induct_step:"\<lbrace>p'\<rbrace> C \<lbrace>q'\<rbrace>\<^sub>u" 
  assumes I0 :"`q'\<Rightarrow> q`"  
  shows "\<lbrace>p\<rbrace>C\<lbrace>q\<rbrace>\<^sub>u"
proof(rule post_weak_hoare_rel[OF _ I0], goal_cases)
  case 1
  then show ?case by (rule pre_str_hoare_rel[OF I0' induct_step]) 
qed   
  
subsection {*Hoare and assertion logic*}

lemma conj_hoare_rel: 
  assumes"\<lbrace>p\<rbrace>C\<lbrace>r\<rbrace>\<^sub>u" and "\<lbrace>p\<rbrace>C\<lbrace>s\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<rbrace>C\<lbrace>r \<and> s\<rbrace>\<^sub>u"
  using assms  
  by rel_simp
    
subsection {*Hoare SKIP*}

lemma skip_hoare_rel[hoare_sp_rules, hoare_wp_rules]: 
  "\<lbrace>p\<rbrace>SKIP\<^sub>r\<lbrace>p\<rbrace>\<^sub>u"
  by rel_simp
    
lemma skip_rel_hoare_rel_intro: 
  "`p \<Rightarrow> q`\<Longrightarrow>\<lbrace>p\<rbrace>SKIP\<^sub>r\<lbrace>q\<rbrace>\<^sub>u"
  by rel_simp
  
subsection {*Hoare for assignment*}

lemma assigns_hoare_rel_intro: 
  assumes"`p \<Rightarrow> \<sigma> \<dagger> q`" 
  shows  "\<lbrace>p\<rbrace>\<langle>\<sigma>\<rangle>\<^sub>a\<lbrace>q\<rbrace>\<^sub>u"
  using assms  
  by rel_simp
    
lemma assigns_backward_hoare_rel[hoare_wp_rules]: 
  "\<lbrace>\<sigma> \<dagger> p\<rbrace>\<langle>\<sigma>\<rangle>\<^sub>a\<lbrace>p\<rbrace>\<^sub>u"
  by rel_simp
    
lemma assigns_floyd_hoare_rel[hoare_sp_rules]:
  assumes "vwb_lens x"
  shows \<open>\<lbrace>p\<rbrace>assign_r x e\<lbrace>\<^bold>\<exists>v \<bullet> p\<lbrakk>\<guillemotleft>v\<guillemotright>/x\<rbrakk> \<and> &x =\<^sub>u e\<lbrakk>\<guillemotleft>v\<guillemotright>/x\<rbrakk>\<rbrace>\<^sub>u\<close> (*If I use the syntax sugar the rule  does not match because the same syntax sugar is used for H1 and H3 designs *)
  using assms
  apply rel_simp   
  subgoal for a    
    apply (rule exI[where x = "lens_get x a"])
    apply simp
    done
  done    

subsection {*Hoare for Sequential Composition*}

lemma seq_invariant_hoare_rel: 
 "\<lbrakk>\<lbrace>p\<rbrace>Q\<^sub>1\<lbrace>p\<rbrace>\<^sub>u ; \<lbrace>p\<rbrace>Q\<^sub>2\<lbrace>p\<rbrace>\<^sub>u\<rbrakk> \<Longrightarrow> \<lbrace>p\<rbrace>Q\<^sub>1 ;; Q\<^sub>2\<lbrace>p\<rbrace>\<^sub>u"
  by rel_simp 
    
lemma seq_stronger_pre_1_hoare_rel: 
  "\<lbrakk> \<lbrace>p \<and> q\<rbrace>Q\<^sub>1\<lbrace>p \<and> q\<rbrace>\<^sub>u ; \<lbrace>p \<and> q\<rbrace>Q\<^sub>2\<lbrace>q\<rbrace>\<^sub>u \<rbrakk> \<Longrightarrow> \<lbrace>p \<and> q\<rbrace>Q\<^sub>1 ;; Q\<^sub>2\<lbrace>q\<rbrace>\<^sub>u"
  by rel_simp

lemma seq_stronger_pre_2_hoare_rel: 
  "\<lbrakk> \<lbrace>p \<and> q\<rbrace>Q\<^sub>1\<lbrace>p \<and> q\<rbrace>\<^sub>u ; \<lbrace>p \<and> q\<rbrace>Q\<^sub>2\<lbrace>p\<rbrace>\<^sub>u\<rbrakk> \<Longrightarrow> \<lbrace>p \<and> q\<rbrace>Q\<^sub>1 ;; Q\<^sub>2\<lbrace>p\<rbrace>\<^sub>u"
  by rel_simp
    
lemma seq_hoare_rel[hoare_wp_rules, hoare_sp_rules]: 
  assumes"\<lbrace>p\<rbrace>C\<^sub>1\<lbrace>s\<rbrace>\<^sub>u" and "\<lbrace>s\<rbrace>C\<^sub>2\<lbrace>r\<rbrace>\<^sub>u" 
  shows"\<lbrace>p\<rbrace>C\<^sub>1 ;; C\<^sub>2\<lbrace>r\<rbrace>\<^sub>u"
  using assms
  by rel_simp
    
subsection {*Hoare for Conditional*}

lemma cond_hoare_rel: 
  assumes "\<lbrace>b \<and> p\<rbrace>C\<^sub>1\<lbrace>q\<rbrace>\<^sub>u" and "\<lbrace>\<not>b \<and> p\<rbrace>C\<^sub>2\<lbrace>q\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace> bif b then C\<^sub>1 else C\<^sub>2 eif\<lbrace>q\<rbrace>\<^sub>u"
  using assms
  by pred_simp
    
lemma cond_hoare_rel_wp[hoare_wp_rules]: 
  assumes "\<lbrace>p'\<rbrace>C\<^sub>1\<lbrace>q\<rbrace>\<^sub>u" and "\<lbrace>p''\<rbrace>C\<^sub>2\<lbrace>q\<rbrace>\<^sub>u"
  shows "\<lbrace>(b \<and> p')\<or> (\<not>b \<and> p'')\<rbrace>bif b then C\<^sub>1 else C\<^sub>2 eif\<lbrace>q\<rbrace>\<^sub>u"
  using assms
  by pred_simp
    
lemma cond_hoare_rel_sp[hoare_sp_rules]:
  assumes \<open>\<lbrace>b \<and> p\<rbrace>C\<^sub>1\<lbrace>q\<rbrace>\<^sub>u\<close> and \<open>\<lbrace>\<not>b \<and> p\<rbrace>C\<^sub>2\<lbrace>s\<rbrace>\<^sub>u\<close>
  shows \<open>\<lbrace>p\<rbrace>bif b then C\<^sub>1 else C\<^sub>2 eif\<lbrace>q \<or> s\<rbrace>\<^sub>u\<close>
  using assms
  by pred_simp
    
subsection {*Hoare for recursion*}
  
lemma nu_hoare_rel_partial: 
  assumes induct_step:
    "\<And> P. \<lbrace>p\<rbrace>P\<lbrace>q\<rbrace>\<^sub>u \<Longrightarrow> \<lbrace>p\<rbrace>F P\<lbrace>q\<rbrace>\<^sub>u"   
  shows "\<lbrace>p\<rbrace>\<nu> F \<lbrace>q\<rbrace>\<^sub>u"  
  unfolding hoare_r_def
  apply (rule lfp_lowerbound)
    apply (rule induct_step[unfolded hoare_r_def])
    apply simp
  done    
  
lemma mu_hoare_rel:
  assumes WF: "wf R"
  assumes M:"mono F"  
  assumes induct_step:
    "\<And> st P. \<lbrace>p \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>P\<lbrace>q\<rbrace>\<^sub>u \<Longrightarrow> \<lbrace>p \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>F P\<lbrace>q\<rbrace>\<^sub>u"   
  shows "\<lbrace>p\<rbrace>\<mu> F \<lbrace>q\<rbrace>\<^sub>u"  
  unfolding hoare_r_def
proof (rule mu_rec_total_utp_rule[OF WF M , of _ e ], goal_cases)
  case (1 st)
  then show ?case 
  by (rule induct_step[unfolded hoare_r_def], simp)    
qed
    
lemma mu_hoare_rel':
  assumes WF: "wf R"
  assumes M:"mono F"  
  assumes induct_step:
    "\<And> st P. \<lbrace>p \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace> P \<lbrace>q\<rbrace>\<^sub>u \<Longrightarrow> \<lbrace>p \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace> F P \<lbrace>q\<rbrace>\<^sub>u" 
  assumes I0: "`p' \<Rightarrow> p`"  
  shows "\<lbrace>p'\<rbrace>\<mu> F \<lbrace>q\<rbrace>\<^sub>u"
  by (simp add: pre_str_hoare_rel[OF I0 mu_hoare_rel[OF WF M induct_step]])

subsection \<open>Hoare for frames\<close>

lemma 
  "vwb_lens a \<Longrightarrow> vwb_lens b \<Longrightarrow>a\<bowtie>b \<Longrightarrow> 
   a:[a :== \<guillemotleft>x\<guillemotright>;; b :== \<guillemotleft>y\<guillemotright>] = a :== \<guillemotleft>x\<guillemotright>"
  unfolding lens_indep_def
  by rel_simp


lemma antiframe_hoare_meta_basic:
  assumes "vwb_lens a"
  assumes "a \<sharp> r" 
  assumes "a \<natural> q"    
  assumes "\<lbrace>p\<rbrace>P\<lbrace>q\<rbrace>\<^sub>u"  
  shows "\<lbrace>p \<and> r\<rbrace>a:[P]\<lbrace>q \<and> r\<rbrace>\<^sub>u"
  using assms
  by rel_simp  

lemma antiframe_hoare_rel_basic:
  assumes vwb_lens_g :"vwb_lens g" 
  assumes vwb_lens_g':"vwb_lens g'" 
  assumes vwb_lens_l :"vwb_lens l"
  assumes indep_g_l  :"g \<bowtie> l" 
  assumes region     :"g' \<subseteq>\<^sub>L g"
  assumes modifies   :"{&g', &l}:[PROG] = PROG" 
  assumes hoare      : "\<lbrace>p\<rbrace>PROG\<lbrace>q\<rbrace>\<^sub>u"  
  shows "\<lbrace>p \<and> r\<rbrace>g':[PROG]\<lbrace>(\<exists>l \<bullet> q) \<and>  (\<exists>g' \<bullet> r)\<rbrace>\<^sub>u"
  using vwb_lens_g vwb_lens_g' vwb_lens_l indep_g_l modifies hoare  
  unfolding lens_indep_def
  apply rel_simp
  using indep_g_l 
  apply rel_simp
  apply auto
   apply(insert region)
  subgoal for  a x
    apply (rule exI[where x ="get\<^bsub>l\<^esub>x"])
    using region unfolding lens_indep_def
    apply auto
    apply (drule  sublens_pres_indep[OF _ indep_g_l])
    unfolding lens_indep_def
    apply auto
    done
  subgoal   for  a x
    apply (rule exI[where x ="get\<^bsub>g'\<^esub>a"])
    unfolding lens_indep_def
    apply auto
    done
  done

lemma antiframe_hoare_rel_wp:
  assumes vwb_lens_g :"vwb_lens g" 
  assumes vwb_lens_g':"vwb_lens g'" 
  assumes vwb_lens_l :"vwb_lens l"
  assumes indep_g_l  :"g \<bowtie> l" 
  assumes region     :"g' \<subseteq>\<^sub>L g"
  assumes modifies   :"{&g', &l}:[PROG] = PROG" 
  assumes hoare      : "\<lbrace>p\<rbrace>PROG\<lbrace>q\<rbrace>\<^sub>u" 
  assumes conseq     :"`(\<exists>l \<bullet> q) \<and> (\<exists>g' \<bullet> r) \<Rightarrow> r`" 
  shows "\<lbrace>p \<and>  (\<exists>g' \<bullet> r)\<rbrace>g':[PROG]\<lbrace>r\<rbrace>\<^sub>u"
  apply (rule post_weak_hoare_rel)
   apply (rule antiframe_hoare_rel_basic[OF  vwb_lens_g vwb_lens_g' vwb_lens_l indep_g_l region modifies hoare, simplified])
  apply (simp add: conseq exists_twice vwb_lens_g')
  done
  
lemma antiframe_hoare_rel_stronger:
  assumes "vwb_lens a"
  assumes "a \<sharp> r" 
  assumes "a \<natural> q"    
  assumes "\<lbrace>p \<and> r\<rbrace>P\<lbrace>q\<rbrace>\<^sub>u"  
  shows "\<lbrace>p \<and> r\<rbrace>a:[P]\<lbrace>q \<and> r\<rbrace>\<^sub>u"
  using assms
  by rel_simp

lemma frame_hoare_rel:
  assumes "vwb_lens a"
  assumes "a \<natural> r" 
  assumes "a \<sharp> q"    
  assumes "\<lbrace>p\<rbrace>P\<lbrace>q\<rbrace>\<^sub>u"  
  shows "\<lbrace>p \<and> r\<rbrace> a:\<lbrakk>P\<rbrakk> \<lbrace>q \<and> r\<rbrace>\<^sub>u"
  using assms
  by rel_simp
    
lemma frame_hoare_rel_stronger:
  assumes "vwb_lens a"
  assumes "a \<natural> r" 
  assumes "a \<sharp> q"    
  assumes "\<lbrace>p \<and> r\<rbrace>P\<lbrace>q\<rbrace>\<^sub>u"  
  shows "\<lbrace>p \<and> r\<rbrace> a:\<lbrakk>P\<rbrakk> \<lbrace>q \<and> r\<rbrace>\<^sub>u"
  using assms 
  by rel_simp

lemma frame_hoare_rel_wp_strong:
  assumes 1:"vwb_lens l" 
 assumes 2:"usedBy l r"
  assumes 3:"\<lbrace>p\<rbrace>P\<lbrace> q\<rbrace>\<^sub>u"
  shows "\<lbrace>p \<and> r\<rbrace> l:\<lbrakk>P\<rbrakk> \<lbrace>r\<rbrace>\<^sub>u"
  using 3 2 1 unfolding lens_indep_def
  by rel_simp

lemma frame_hoare_rel_minimal:
  assumes 1:"vwb_lens g" 
  assumes 2:"vwb_lens g'" 
  assumes 3:"vwb_lens l"
  assumes 4:"l \<bowtie> g" 
  assumes 8:"g' \<subseteq>\<^sub>L g"
  assumes 5:"{&g', &l}:[PROG] = PROG" 
  assumes 6: "\<lbrace>p\<rbrace>PROG\<lbrace>q\<rbrace>\<^sub>u"  
  shows "\<lbrace>p \<and> r\<rbrace>l:\<lbrakk>PROG\<rbrakk>\<lbrace>(\<exists>l \<bullet> q) \<and>  (\<exists>g' \<bullet> r)\<rbrace>\<^sub>u"
  using 1 2 3 4 5 6  unfolding lens_indep_def
  apply rel_simp
  apply rel_simp
  using 4 
  apply rel_simp
  apply auto
   apply (metis lens_override_def lens_override_idem)
  using 4 8 unfolding lens_indep_def sublens_def
  apply simp
proof -
  fix a :: 'b and x :: 'b
  assume a1: "\<exists>Z. vwb_lens Z \<and> g' = Z ;\<^sub>L g"
  assume a2: "vwb_lens l"
  assume a3: "\<forall>a b. (\<exists>x. \<lbrakk>PROG\<rbrakk>\<^sub>e (a, x) \<and> b = put\<^bsub>g'\<^esub> (put\<^bsub>l\<^esub> a (get\<^bsub>l\<^esub> x)) (get\<^bsub>g'\<^esub> x)) = \<lbrakk>PROG\<rbrakk>\<^sub>e (a, b)"
  assume a4: "vwb_lens g'"
  assume "\<lbrakk>PROG\<rbrakk>\<^sub>e (a, x)"
  assume a5: "\<lbrakk>r\<rbrakk>\<^sub>e a"
  have "\<forall>c d b l. put\<^bsub>g'\<^esub> (put\<^bsub>l\<^esub> b (d::'d)) c = put\<^bsub>l\<^esub> (put\<^bsub>g'\<^esub> b c) d \<or> \<not> l \<bowtie> g"
    using a1 by (metis (no_types) lens_indep_def lens_indep_right_ext)
  then have "\<forall>c d b. put\<^bsub>g'\<^esub> (put\<^bsub>l\<^esub> b d) c = put\<^bsub>l\<^esub> (put\<^bsub>g'\<^esub> b c) d"
    by (metis "4") (* > 1.0 s, timed out *)
  then show "\<exists>c. \<lbrakk>r\<rbrakk>\<^sub>e (put\<^bsub>g'\<^esub> (put\<^bsub>l\<^esub> x (get\<^bsub>l\<^esub> a)) c)"
    using a5 a4 a3 a2 
    proof -
      have f1: "\<forall>c b. put\<^bsub>l\<^esub> (put\<^bsub>g'\<^esub> b c) (get\<^bsub>l\<^esub> b) = put\<^bsub>g'\<^esub> b c"
        by (metis (no_types) \<open>\<forall>c d b. put\<^bsub>g'\<^esub> (put\<^bsub>l\<^esub> b d) c = put\<^bsub>l\<^esub> (put\<^bsub>g'\<^esub> b c) d\<close> a2 lens_override_def lens_override_idem)
    have f2: "\<forall>c ca b. put\<^bsub>g'\<^esub> (put\<^bsub>g'\<^esub> b c) ca = put\<^bsub>g'\<^esub> b ca"
      by (metis a4 mwb_lens.put_put vwb_lens_mwb)
    have "\<exists>c. \<lbrakk>r\<rbrakk>\<^sub>e (put\<^bsub>g'\<^esub> a c)"
      by (metis (no_types) a4 a5 lens_override_def lens_override_idem)
    then show ?thesis
      using f2 f1 by (metis (no_types) \<open>\<forall>c d b. put\<^bsub>g'\<^esub> (put\<^bsub>l\<^esub> b d) c = put\<^bsub>l\<^esub> (put\<^bsub>g'\<^esub> b c) d\<close> \<open>\<lbrakk>PROG\<rbrakk>\<^sub>e (a, x)\<close> a2 a3 mwb_lens.put_put vwb_lens_mwb)
    qed     
qed
  
lemma frame_hoare_rel_wp:
  assumes 1:"vwb_lens g" 
  assumes 2:"vwb_lens g'" 
  assumes 3:"vwb_lens l"
  assumes 4:"l \<bowtie> g" 
  assumes 8:"g' \<subseteq>\<^sub>L g"
  assumes 5:"{&g', &l}:[PROG] = PROG" 
  assumes 6:"\<lbrace>p\<rbrace>PROG\<lbrace>q\<rbrace>\<^sub>u"
  assumes 7:"`(\<exists>l \<bullet> q) \<Rightarrow> r`"     
  shows "\<lbrace>p \<and>  (\<exists>g' \<bullet> r)\<rbrace> l:\<lbrakk>PROG\<rbrakk> \<lbrace>r\<rbrace>\<^sub>u"
  apply (rule post_weak_hoare_rel)
   apply (rule frame_hoare_rel_minimal[OF 1 2 3 4 8 5 6 ])
  apply (simp add: "2" "7" exists_twice) 
  using "7" refBy_order utp_pred_laws.le_infI1 
  apply blast
  done

lemma frame_hoare_rel_sp:
  assumes 1:"vwb_lens g" 
  assumes 2:"vwb_lens g'" 
  assumes 3:"vwb_lens l"
  assumes 4:"l \<bowtie> g" 
  assumes 8:"g' \<subseteq>\<^sub>L g"
  assumes 5:"{&g', &l}:[C] = C" 
  assumes 6:"\<lbrace>p\<rbrace>C\<lbrace>q\<rbrace>\<^sub>u"
  assumes 7:"`r \<Rightarrow> p`"     
  shows "\<lbrace>r\<rbrace> l:\<lbrakk>C\<rbrakk> \<lbrace>(\<exists> l \<bullet> q) \<and> (\<exists>g' \<bullet> r)\<rbrace>\<^sub>u"
  apply (rule pre_str_hoare_rel)
   defer
   apply (rule frame_hoare_rel_minimal[OF 1 2 3 4 8 5 6])
  apply (simp add: "7" refBy_order utp_pred_laws.inf.absorb2)
  done

subsection {*Hoare for while loop*} 
  
lemma if_rel_mono:
  "mono (\<lambda>X. bif b then P ;; X else Q eif)"
  by (auto intro: monoI seqr_mono cond_mono) 

lemma while_gfp_hoare_rel_minimal_partial:
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes induct_step: "\<lbrace>invar\<and> b\<rbrace> C \<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>while\<^sup>\<top> b do C od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding while_gfp_rel_def_alt
  apply (rule pre_str_hoare_rel[OF seq_step])   
  apply (rule nu_hoare_rel_partial)
  apply (rule cond_hoare_rel)  
   apply (rule seq_hoare_rel[where s="invar"])
  subgoal for st 
    using pre_str_hoare_rel induct_step seq_step
    unfolding hoare_r_def
    apply pred_simp
    done
  subgoal for st 
    apply assumption
    done
  subgoal for st 
    apply pred_simp
    done
  done          

lemma while_gfp_invr_hoare_rel_minimal_partial:
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes induct_step: "\<lbrace>invar\<and> b\<rbrace> C \<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>invr invar while\<^sup>\<top> b do C od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding while_gfp_invr_rel_def 
  by (auto intro!: assms while_gfp_hoare_rel_minimal_partial)
    
lemma while_gfp_hoare_rel_consequence_partial:
  assumes seq_step: "`p\<Rightarrow>invar`"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes I0': "` invar \<and> b  \<Rightarrow> p'`"
  assumes induct_step: "\<lbrace>p'\<rbrace>C \<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>while\<^sup>\<top> b do C od\<lbrace>q\<rbrace>\<^sub>u"
  by (rule consequence_hoare_rel[OF seq_step _ PHI, 
                                 OF while_gfp_hoare_rel_minimal_partial[OF uimp_refl],
                                 OF consequence_hoare_rel[OF I0' induct_step I0]])

lemma while_gfp_invr_hoare_rel_consequence_partial:
  assumes seq_step: "`p\<Rightarrow>invar`"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes I0': "` invar \<and> b  \<Rightarrow> p'`"
  assumes induct_step: "\<lbrace>p'\<rbrace>C \<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>invr invar while\<^sup>\<top> b do C od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding while_gfp_invr_rel_def 
  by (auto intro!: assms while_gfp_hoare_rel_consequence_partial)  

lemma while_gfp_hoare_rel_wp_partial:
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes I0': "`invar \<and> b \<Rightarrow> p'`"
  assumes induct_step: "\<lbrace>p'\<rbrace>C \<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>invar\<rbrace>while\<^sup>\<top> b do C od\<lbrace>q\<rbrace>\<^sub>u"
  by (rule consequence_hoare_rel[OF uimp_refl _ PHI, 
                                 OF while_gfp_hoare_rel_minimal_partial[OF uimp_refl],
                                 OF consequence_hoare_rel[OF I0' induct_step uimp_refl]])    

lemma while_gfp_invr_hoare_rel_wp_partial:
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes I0': "`invar \<and> b \<Rightarrow> p'`"
  assumes induct_step: "\<lbrace>p'\<rbrace>C \<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>invar\<rbrace>invr invar while\<^sup>\<top> b do C od\<lbrace>q\<rbrace>\<^sub>u"                             
  unfolding while_gfp_invr_rel_def 
  by (auto intro!: assms while_gfp_hoare_rel_consequence_partial)  
                           
lemma while_gfp_hoare_rel_sp_partial:
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes induct_step: "\<lbrace>invar \<and> b\<rbrace>C \<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>while\<^sup>\<top> b do C od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  by (rule consequence_hoare_rel[OF seq_step _ uimp_refl, 
                                 OF while_gfp_hoare_rel_minimal_partial[OF uimp_refl],
                                 OF consequence_hoare_rel[OF uimp_refl induct_step I0]])

lemma while_gfp_invr_hoare_rel_sp_partial:
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes induct_step: "\<lbrace> b \<and> invar \<rbrace>C \<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>invr invar while\<^sup>\<top> b do C od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding while_gfp_invr_rel_def 
  by (metis I0 induct_step post_weak_hoare_rel seq_step utp_pred_laws.inf_commute while_gfp_hoare_rel_minimal_partial)

lemma while_lfp_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`" 
  assumes induct_step:"\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"    
  shows "\<lbrace>p\<rbrace>while\<^sub>\<bottom> b do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding while_lfp_rel_def_alt
proof (rule pre_str_hoare_rel[OF seq_step mu_hoare_rel[OF WF if_rel_mono cond_hoare_rel, of _ e]], goal_cases)
  case (1 st X)
  assume *: "\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>X\<lbrace>\<not> b \<and> invar\<rbrace>\<^sub>u"  
  show ?case 
  proof (rule seq_hoare_rel [of _ _ "invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>"], goal_cases)
    case 1
    then show ?case by (rule induct_step)
  next
    case 2
    then show ?case by (rule *) 
  qed 
next
  case (2 st X)
  then show ?case by rel_simp 
qed
    
lemma while_lfp_invr_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`" 
  assumes induct_step:"\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"    
  shows "\<lbrace>p\<rbrace>invr invar while\<^sub>\<bottom> b do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding while_lfp_invr_rel_def
  using assms while_lfp_hoare_rel_minimal
  by blast  

lemma while_lfp_invr_vrt_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`" 
  assumes induct_step:"\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"    
  shows "\<lbrace>p\<rbrace>invr invar vrt \<guillemotleft>R\<guillemotright> while\<^sub>\<bottom> b do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding while_lfp_invr_vrt_rel_def
  using assms while_lfp_hoare_rel_minimal
  by blast  
    
lemma while_lfp_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes  I0: "\<And> st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `"
  shows "\<lbrace>p\<rbrace>while\<^sub>\<bottom> b do body od\<lbrace>q\<rbrace>\<^sub>u" 
  by (rule consequence_hoare_rel[OF seq_step _ PHI, 
                                 OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                 OF consequence_hoare_rel[OF I0' induct_step I0]])

lemma while_lfp_invr_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes  I0: "\<And> st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `"
  shows "\<lbrace>p\<rbrace>invr invar while\<^sub>\<bottom> b do body od\<lbrace>q\<rbrace>\<^sub>u" 
  unfolding while_lfp_invr_rel_def
  by (rule consequence_hoare_rel[OF seq_step _ PHI, 
                                 OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                 OF consequence_hoare_rel[OF I0' induct_step I0]])

lemma while_lfp_invr_vrt_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes  I0: "\<And> st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `"
  shows "\<lbrace>p\<rbrace>invr invar vrt \<guillemotleft>R\<guillemotright> while\<^sub>\<bottom> b do body od\<lbrace>q\<rbrace>\<^sub>u" 
  unfolding while_lfp_invr_vrt_rel_def  
  by (rule consequence_hoare_rel[OF seq_step _ PHI, 
                                 OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                 OF consequence_hoare_rel[OF I0' induct_step I0]])
                              
lemma while_lfp_hoare_rel_wp:
  assumes WF: "wf R"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> \<rbrace>\<^sub>u"
  shows "\<lbrace>invar\<rbrace>while\<^sub>\<bottom> b do body od\<lbrace>q\<rbrace>\<^sub>u" 
  by (rule consequence_hoare_rel[OF uimp_refl _ PHI,
                                 OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                 OF consequence_hoare_rel[OF I0' induct_step uimp_refl]])

lemma while_lfp_invr_hoare_rel_wp:
  assumes WF: "wf R"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> \<rbrace>\<^sub>u"
  shows "\<lbrace>invar\<rbrace>invr invar while\<^sub>\<bottom> b do body od\<lbrace>q\<rbrace>\<^sub>u"                              
  unfolding while_lfp_invr_rel_def
  by (rule consequence_hoare_rel[OF uimp_refl _ PHI,
                                 OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                 OF consequence_hoare_rel[OF I0' induct_step uimp_refl]])

lemma while_lfp_invr_vrt_hoare_rel_wp:
  assumes WF: "wf R"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> \<rbrace>\<^sub>u"
  shows "\<lbrace>invar\<rbrace>invr invar vrt \<guillemotleft>R\<guillemotright> while\<^sub>\<bottom> b do body od\<lbrace>q\<rbrace>\<^sub>u"                              
  unfolding while_lfp_invr_vrt_rel_def
  by (rule consequence_hoare_rel[OF uimp_refl _ PHI,
                                 OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                 OF consequence_hoare_rel[OF I0' induct_step uimp_refl]])
                                
lemma while_lfp_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes  I0: "\<And> st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `"
  shows "\<lbrace>p\<rbrace>while\<^sub>\<bottom> b do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u" 
  by (rule consequence_hoare_rel [OF seq_step _ uimp_refl, 
                                  OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                  OF consequence_hoare_rel[OF uimp_refl induct_step I0]])                              
                              
lemma while_lfp_invr_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes  I0: "\<And> st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `"
  shows "\<lbrace>p\<rbrace>invr invar while\<^sub>\<bottom> b do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u" 
  unfolding while_lfp_invr_rel_def  
  by (rule consequence_hoare_rel[OF seq_step _ uimp_refl, 
                                 OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                 OF consequence_hoare_rel[OF uimp_refl induct_step I0]])

lemma while_lfp_invr_vrt_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "`p \<Rightarrow> invar`"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes  I0: "\<And> st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `"
  shows "\<lbrace>p\<rbrace>invr invar vrt \<guillemotleft>R\<guillemotright> while\<^sub>\<bottom> b do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u" 
  unfolding while_lfp_invr_vrt_rel_def  
  by (rule consequence_hoare_rel[OF seq_step _ uimp_refl, 
                                  OF while_lfp_hoare_rel_minimal[OF WF uimp_refl, of _ _ e],
                                  OF consequence_hoare_rel[OF uimp_refl induct_step I0]])
  
subsection {*Hoare for from_until_loop*}  
  
lemma from_until_lfp_hoare_rel_minimal_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<lbrace>\<not> exit \<and> invar\<rbrace> body\<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>from\<^sup>\<top> init until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
  unfolding from_until_gfp_rel_alt_def
  proof -
    have induct_step': "\<lbrace>invar \<and> \<not> exit\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
      by (metis assms(2) utp_pred_laws.inf.commute)
    have "`\<not> \<not> exit \<and> invar \<Rightarrow> invar \<and> exit`"
      by (simp add: utp_pred_laws.inf.commute)
    then have "\<lbrace>invar\<rbrace>while\<^sup>\<top> \<not> exit do body od\<lbrace>invar \<and> exit\<rbrace>\<^sub>u"
      using induct_step' by (metis (no_types) p_imp_p taut_true while_gfp_hoare_rel_wp_partial)
    then show "\<lbrace>p\<rbrace>init ;; while\<^sup>\<top> \<not> exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
      by (metis (no_types) seq_hoare_rel seq_step utp_pred_laws.inf.commute)
  qed

lemma from_until_lfp_invr_hoare_rel_minimal_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<lbrace>\<not> exit \<and> invar\<rbrace> body\<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>from\<^sup>\<top> init invr invar until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
  unfolding from_until_gfp_invr_rel_def
  by (auto intro!: assms from_until_lfp_hoare_rel_minimal_partial)  

lemma from_until_gfp_hoare_rel_consequence_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q `"
  assumes I0': "`invar \<and> \<not> exit \<Rightarrow> p'`"  
  assumes induct_step: "\<lbrace>p'\<rbrace> body\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>from\<^sup>\<top> init until exit do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding from_until_gfp_rel_alt_def
  proof -
    have PHI': "`\<not> \<not> exit \<and> invar \<Rightarrow> q`"
      by (metis PHI utp_pred_laws.double_compl)
    have "`invar \<Rightarrow> invar`"
      by (metis uimp_refl)
    then have "\<lbrace>invar\<rbrace>while\<^sup>\<top> \<not> exit do body od\<lbrace>q\<rbrace>\<^sub>u"
      using PHI' I0 I0' induct_step while_gfp_hoare_rel_consequence_partial by blast
    then show "\<lbrace>p\<rbrace>init ;; while\<^sup>\<top> \<not> exit do body od\<lbrace>q\<rbrace>\<^sub>u"
      using seq_hoare_rel seq_step by blast
  qed

lemma from_until_gfp_invr_hoare_rel_consequence_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q `"
  assumes I0': "`invar \<and> \<not> exit \<Rightarrow> p'`"  
  assumes induct_step: "\<lbrace>p'\<rbrace> body\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>from\<^sup>\<top> init invr invar until exit do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding from_until_gfp_invr_rel_def
  by (auto intro!: assms from_until_gfp_hoare_rel_consequence_partial)  

lemma from_until_gfp_hoare_rel_wp_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`" 
  assumes I0': "`invar \<and> \<not> exit \<Rightarrow> p'`"  
  assumes induct_step:"\<lbrace>p'\<rbrace> body\<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>from\<^sup>\<top> init until exit do body od\<lbrace>q\<rbrace>\<^sub>u"  
  unfolding from_until_gfp_rel_alt_def
proof -
  have "\<lbrace>invar\<rbrace>while\<^sup>\<top> \<not> exit do body od\<lbrace>q\<rbrace>\<^sub>u"
    by (metis (no_types) I0' PHI induct_step utp_pred_laws.double_compl while_gfp_hoare_rel_wp_partial)
  then show "\<lbrace>p\<rbrace>init ;; while\<^sup>\<top> \<not> exit do body od\<lbrace>q\<rbrace>\<^sub>u"
    using seq_hoare_rel seq_step by blast
qed 

lemma from_until_gfp_invr_hoare_rel_wp_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`" 
  assumes I0': "`invar \<and> \<not> exit \<Rightarrow> p'`"  
  assumes induct_step:"\<lbrace>p'\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>from\<^sup>\<top> init invr invar until exit do body od\<lbrace>q\<rbrace>\<^sub>u"  
  unfolding from_until_gfp_invr_rel_def
  by (auto intro!: assms from_until_gfp_hoare_rel_wp_partial)  

lemma from_until_gfp_hoare_rel_sp_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"
  assumes induct_step: "\<lbrace>invar \<and> \<not> exit\<rbrace> body\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>from\<^sup>\<top> init until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
  unfolding from_until_gfp_rel_alt_def
proof -
  have "\<lbrace>invar\<rbrace>while\<^sup>\<top> \<not> exit do body od\<lbrace>\<not> \<not> exit \<and> invar\<rbrace>\<^sub>u"
    by (metis I0 induct_step uimp_refl while_gfp_hoare_rel_sp_partial)
  then have "\<lbrace>invar\<rbrace>while\<^sup>\<top> \<not> exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
    by (metis double_negation)
  then show "\<lbrace>p\<rbrace>init ;; while\<^sup>\<top> \<not> exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
    by (metis seq_hoare_rel seq_step)
qed

lemma from_until_gfp_invr_hoare_rel_sp_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"
  assumes induct_step: "\<lbrace>invar \<and> \<not> exit\<rbrace> body\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>from\<^sup>\<top> init invr invar until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
  unfolding  from_until_gfp_invr_rel_def
  by (auto intro!: assms from_until_gfp_hoare_rel_sp_partial)
    
lemma from_until_lfp_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
  unfolding from_until_lfp_rel_alt_def while_lfp_rel_def_alt
  by (rule seq_hoare_rel[OF seq_step, 
                         OF mu_hoare_rel[OF WF if_rel_mono cond_hoare_rel, of _ e],
                         OF seq_hoare_rel[OF induct_step], OF _ skip_rel_hoare_rel_intro],                             
      rel_auto+)

lemma from_until_lfp_invr_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init invr invar until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
  unfolding from_until_lfp_invr_rel_def 
  using     from_until_lfp_hoare_rel_minimal [OF WF seq_step induct_step] .
    
lemma from_until_lfp_invr_vrt_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init invr invar vrt \<guillemotleft>R\<guillemotright> until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u"
  unfolding from_until_lfp_invr_vrt_rel_def 
  using     from_until_lfp_hoare_rel_minimal [OF WF seq_step induct_step] .

lemma from_until_lfp_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init until exit do body od\<lbrace>q\<rbrace>\<^sub>u" 
  unfolding from_until_lfp_rel_alt_def
  by (simp add: PHI seq_hoare_rel[OF seq_step] 
                while_lfp_hoare_rel_consequence[OF WF uimp_refl _ I0' induct_step I0])

lemma from_until_lfp_invr_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init invr invar until exit do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding from_until_lfp_invr_rel_def 
  using     from_until_lfp_hoare_rel_consequence [OF WF seq_step PHI I0' induct_step I0].
    
lemma from_until_lfp_invr_vrt_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init invr invar vrt \<guillemotleft>R\<guillemotright> until exit do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding from_until_lfp_invr_vrt_rel_def
  using     from_until_lfp_hoare_rel_consequence [OF WF seq_step PHI I0' induct_step I0].
 
lemma from_until_lfp_hoare_rel_wp:
  assumes WF: "wf R" 
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init until exit do body od\<lbrace>q\<rbrace>\<^sub>u" 
  unfolding from_until_lfp_rel_alt_def
  by (simp add: PHI seq_hoare_rel[OF seq_step]
                while_lfp_hoare_rel_consequence[OF WF uimp_refl _ I0' induct_step uimp_refl])

lemma from_until_lfp_invr_hoare_rel_wp:
  assumes WF: "wf R" 
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init invr invar until exit do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding from_until_lfp_invr_rel_def
  using from_until_lfp_hoare_rel_wp [OF WF seq_step PHI I0' induct_step] .
    
lemma from_until_lfp_invr_vrt_hoare_rel_wp:
  assumes WF: "wf R"   
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init invr invar vrt \<guillemotleft>R\<guillemotright> until exit do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding from_until_lfp_invr_vrt_rel_def
  using from_until_lfp_hoare_rel_wp[OF WF seq_step PHI I0' induct_step] .
    
lemma from_until_lfp_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>u"
  assumes I0': "`q'' \<Rightarrow> invar`"    
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u" 
  unfolding from_until_lfp_rel_alt_def
  by (simp add: uimp_refl seq_hoare_rel[OF seq_step] pre_str_hoare_rel[OF I0']
                while_lfp_hoare_rel_consequence[OF WF uimp_refl _ uimp_refl induct_step I0])

lemma from_until_lfp_invr_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>u"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init invr invar until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u" 
  unfolding from_until_lfp_invr_rel_def
  using from_until_lfp_hoare_rel_sp[OF WF seq_step I0' induct_step I0] .  

lemma from_until_lfp_invr_vrt_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>u"
   assumes I0': "`q'' \<Rightarrow> invar`"    
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>from\<^sub>\<bottom> init invr invar vrt \<guillemotleft>R\<guillemotright> until exit do body od\<lbrace>exit \<and> invar\<rbrace>\<^sub>u" 
  unfolding from_until_lfp_invr_vrt_rel_def
  using from_until_lfp_hoare_rel_sp[OF WF seq_step I0' induct_step I0] .

subsection {*Hoare for do_while_loop*}     

lemma do_while_gfp_hoare_rel_minimal_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"  
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>do body while\<^sup>\<top> b od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding do_while_gfp_rel_alt_def
  proof -
    have "\<lbrace>invar\<rbrace>while\<^sup>\<top> b do body od\<lbrace>invar \<and> \<not> b\<rbrace>\<^sub>u"
      by (metis (no_types) assms(2) conj_comm p_imp_p taut_true while_gfp_hoare_rel_wp_partial)
    then show "\<lbrace>p\<rbrace>body ;; while\<^sup>\<top> b do body od\<lbrace>\<not> b \<and> invar\<rbrace>\<^sub>u"
      by (metis (no_types) conj_comm seq_hoare_rel seq_step)
  qed    

lemma do_while_gfp_invr_hoare_rel_minimal_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"  
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>do body invr invar while\<^sup>\<top> b  od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding do_while_gfp_invr_rel_def
  by (simp add: from_until_lfp_hoare_rel_minimal_partial induct_step seq_step)

lemma do_while_gfp_hoare_rel_consequence_partial:
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes I0': "`b \<and> invar \<Rightarrow> p'`"
  assumes induct_step: "\<lbrace>p'\<rbrace>body\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>do body while\<^sup>\<top> b od\<lbrace>q\<rbrace>\<^sub>u"
proof -
  have "`invar \<and> \<not> \<not> b \<Rightarrow> p'`"
    by (simp add: I0' utp_pred_laws.inf_commute)
  then have "\<lbrace>p\<rbrace>from\<^sup>\<top> body until \<not> b do body od\<lbrace>q\<rbrace>\<^sub>u"
    using I0 PHI from_until_gfp_hoare_rel_consequence_partial induct_step seq_step by blast
  then show ?thesis
    by (metis do_while_gfp_rel_def)
qed  
 
lemma do_while_gfp_invr_hoare_rel_consequence_partial:
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes I0': "`b \<and> invar \<Rightarrow> p'`"
  assumes induct_step: "\<lbrace>p'\<rbrace>body\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>do body invr invar while\<^sup>\<top> b  od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding do_while_gfp_invr_rel_def
  proof -
    have "`invar \<and> \<not> \<not> b \<Rightarrow> p'`"
      by (simp add: I0' utp_pred_laws.inf_commute)
    then show "\<lbrace>p\<rbrace>from\<^sup>\<top> body until \<not> b do body od\<lbrace>q\<rbrace>\<^sub>u"
      using I0 PHI from_until_gfp_hoare_rel_consequence_partial induct_step seq_step by blast
  qed

lemma do_while_gfp_hoare_rel_wp_partial:
  assumes seq_step: "\<lbrace>invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  shows "\<lbrace>invar\<rbrace>do body while\<^sup>\<top> b od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding do_while_gfp_rel_alt_def
  using PHI reverse_impl_refine seq_hoare_rel seq_step utp_pred_laws.inf.cobounded2 
        while_gfp_hoare_rel_wp_partial 
  by fastforce

lemma do_while_gfp_invr_hoare_rel_wp_partial:
  assumes seq_step: "\<lbrace>invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  shows "\<lbrace>invar\<rbrace>do body invr invar while\<^sup>\<top> b  od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding do_while_gfp_invr_rel_def
  using PHI from_until_gfp_hoare_rel_wp_partial reverse_impl_refine seq_step utp_pred_laws.inf_le1 
  by blast
      
lemma do_while_gfp_hoare_rel_sp_partial:
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>do body while\<^sup>\<top> b od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding do_while_gfp_rel_alt_def
  proof -
    have "\<lbrace>invar \<and> b\<rbrace>body\<lbrace>q'\<rbrace>\<^sub>u"
      by (metis conj_comm induct_step)
    then show "\<lbrace>p\<rbrace>body ;; while\<^sup>\<top> b do body od\<lbrace>\<not> b \<and> invar\<rbrace>\<^sub>u"
      by (meson I0 seq_hoare_rel seq_step uimp_refl while_gfp_hoare_rel_sp_partial)
  qed
      
lemma do_while_gfp_invr_hoare_rel_sp_partial:
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q' \<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>do body invr invar while\<^sup>\<top> b  od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding do_while_gfp_rel_alt_def
  using I0 do_while_gfp_invr_hoare_rel_minimal_partial induct_step post_weak_hoare_rel seq_step 
  by blast
    
lemma do_while_lfp_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and>  e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<rbrace>do body while\<^sub>\<bottom> b od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"  
  unfolding do_while_lfp_rel_alt_def
  by (rule seq_hoare_rel[OF seq_step while_lfp_hoare_rel_minimal[OF WF uimp_refl induct_step]])  

lemma do_while_lfp_invr_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and>  e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows" \<lbrace>p\<rbrace>do body while\<^sub>\<bottom> b invr invar od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"  
  unfolding do_while_lfp_invr_rel_def
  by (auto intro!: induct_step from_until_lfp_hoare_rel_minimal[OF WF seq_step ])
 
lemma do_while_lfp_invr_vrt_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and>  e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows" \<lbrace>p\<rbrace>do body while\<^sub>\<bottom> b invr invar vrt \<guillemotleft>R\<guillemotright> od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"  
  unfolding do_while_lfp_invr_vrt_rel_def
  by (auto intro!: induct_step from_until_lfp_hoare_rel_minimal[OF WF seq_step ])
    
lemma do_while_lfp_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>do body while\<^sub>\<bottom> b od\<lbrace>q\<rbrace>\<^sub>u" 
  unfolding do_while_lfp_rel_alt_def
  by (rule seq_hoare_rel[OF seq_step 
                            while_lfp_hoare_rel_consequence[OF WF uimp_refl PHI I0' induct_step I0]])
    
lemma do_while_lfp_invr_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>do body while\<^sub>\<bottom> b invr invar  od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding do_while_lfp_invr_rel_def
  by (auto intro!: assms from_until_lfp_hoare_rel_consequence[OF WF seq_step])

lemma do_while_lfp_invr_vrt_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>do body while\<^sub>\<bottom> b invr invar vrt \<guillemotleft>R\<guillemotright> od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding do_while_lfp_invr_vrt_rel_def
  by (auto intro!: assms from_until_lfp_hoare_rel_consequence[OF WF seq_step])
    
lemma do_while_lfp_hoare_rel_wp:
  assumes WF: "wf R" 
  assumes seq_step: "\<lbrace>invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>invar\<rbrace>do body while\<^sub>\<bottom> b od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding do_while_lfp_rel_alt_def
  by (simp add: uimp_refl seq_hoare_rel[OF seq_step] 
                while_lfp_hoare_rel_wp[OF WF PHI I0' induct_step])

lemma do_while_lfp_invr_hoare_rel_wp:
  assumes WF: "wf R"  
  assumes seq_step: "\<lbrace>invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI : "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>invar\<rbrace>do body while\<^sub>\<bottom> b invr invar od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding do_while_lfp_invr_rel_def do_while_lfp_rel_alt_def  
  by (auto intro!: assms from_until_lfp_hoare_rel_wp)

lemma do_while_lfp_invr_vrt_hoare_rel_wp:
  assumes WF: "wf R"    
  assumes seq_step: "\<lbrace>invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI : "`\<not>b \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>invar\<rbrace>do body while\<^sub>\<bottom> b invr invar vrt \<guillemotleft>R\<guillemotright> od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding do_while_lfp_invr_vrt_rel_def do_while_lfp_invr_rel_def do_while_lfp_rel_alt_def
  by (auto intro!: assms from_until_lfp_hoare_rel_wp)

lemma do_while_lfp_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>q''\<rbrace>\<^sub>u"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>do body while\<^sub>\<bottom> b od\<lbrace>\<not>b  \<and> invar\<rbrace>\<^sub>u"
  unfolding  do_while_lfp_rel_alt_def
  by (simp add: uimp_refl seq_hoare_rel[OF seq_step] pre_str_hoare_rel[OF I0']
                while_lfp_hoare_rel_sp[OF WF uimp_refl induct_step I0])
                 
lemma do_while_lfp_invr_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>q''\<rbrace>\<^sub>u"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>do body  while\<^sub>\<bottom> b invr invar od \<lbrace>\<not>b  \<and> invar\<rbrace>\<^sub>u"
  unfolding do_while_lfp_invr_vrt_rel_def do_while_lfp_invr_rel_def do_while_lfp_rel_alt_def
  by (auto intro!: assms from_until_lfp_hoare_rel_sp)
                 
lemma do_while_lfp_invr_vrt_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>q''\<rbrace>\<^sub>u"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>do body while\<^sub>\<bottom> b invr invar vrt \<guillemotleft>R\<guillemotright> od\<lbrace>\<not>b  \<and> invar\<rbrace>\<^sub>u"
  unfolding do_while_lfp_invr_vrt_rel_def do_while_lfp_rel_alt_def
  by (auto intro!: assms from_until_lfp_hoare_rel_sp)
           
subsection {*Hoare for for_loop*}     

lemma for_hoare_gfp_rel_minimal_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body;;incr\<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>for\<^sup>\<top> (init, b, incr) do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding for_gfp_rel_alt_def
proof -
  have "\<lbrace>invar\<rbrace>while\<^sup>\<top> b do body ;; incr od\<lbrace>invar \<and> \<not> b\<rbrace>\<^sub>u"
    by (metis (no_types) assms(2) p_imp_p taut_true utp_pred_laws.inf_commute while_gfp_hoare_rel_wp_partial)
  then show "\<lbrace>p\<rbrace>init ;; while\<^sup>\<top> b do body ;; incr od\<lbrace>\<not> b \<and> invar\<rbrace>\<^sub>u"
    by (metis (no_types) seq_hoare_rel seq_step utp_pred_laws.inf_commute)
qed
  
lemma for_gfp_invr_hoare_rel_minimal_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body;;incr\<lbrace>invar\<rbrace>\<^sub>u"  
  shows "\<lbrace>p\<rbrace>for\<^sup>\<top> (init, b, incr) invr invar do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  by (simp add: for_gfp_invr_rel_def from_until_lfp_hoare_rel_minimal_partial induct_step seq_step) 
         
lemma for_gfp_hoare_rel_consequence_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "`b \<and> invar \<Rightarrow> p'`"  
  assumes induct_step: "\<lbrace>p'\<rbrace>body;;incr\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q'\<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>for\<^sup>\<top> (init, b, incr) do body od\<lbrace>q\<rbrace>\<^sub>u"
proof -
  have "`invar \<and> \<not> \<not> b \<Rightarrow> p'`"
    by (metis I0' conj_comm double_negation)
  then have "\<lbrace>p\<rbrace>from\<^sup>\<top> init until \<not> b do body ;; incr od\<lbrace>q\<rbrace>\<^sub>u"
    by (meson I0 PHI from_until_gfp_hoare_rel_consequence_partial induct_step seq_step)
  then show ?thesis
    by (metis for_gfp_rel_def)
qed    

lemma for_gfp_invr_hoare_rel_consequence_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "`b \<and> invar \<Rightarrow> p'`"  
  assumes induct_step: "\<lbrace>p'\<rbrace>body;;incr\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q'\<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>for\<^sup>\<top> (init, b, incr) invr invar  do body od\<lbrace>q\<rbrace>\<^sub>u"
  using I0 I0' PHI consequence_hoare_rel for_gfp_invr_hoare_rel_minimal_partial induct_step 
        post_weak_hoare_rel seq_step 
  by blast   
  
lemma for_gfp_hoare_rel_wp_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "`b \<and> invar \<Rightarrow> p'`"  
  assumes induct_step: "\<lbrace>p'\<rbrace>body;;incr\<lbrace>invar\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace>for\<^sup>\<top> (init, b, incr) do body od\<lbrace>q\<rbrace>\<^sub>u"
  using I0' PHI for_gfp_hoare_rel_consequence_partial induct_step seq_step uimp_refl 
  by blast
    
lemma for_gfp_invr_hoare_rel_wp_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow> q`"  
  assumes I0': "`b \<and> invar \<Rightarrow> p'`"  
  assumes induct_step: "\<lbrace>p'\<rbrace>body;;incr\<lbrace>invar\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace>for\<^sup>\<top> (init, b, incr) invr invar do body od\<lbrace>q\<rbrace>\<^sub>u"
  using I0' PHI for_gfp_invr_hoare_rel_consequence_partial induct_step seq_step uimp_refl 
  by blast
    
lemma for_gfp_hoare_rel_sp_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u" 
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body;;incr\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q'\<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>for\<^sup>\<top> (init, b, incr) do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  using I0 for_hoare_gfp_rel_minimal_partial induct_step post_weak_hoare_rel seq_step 
  by blast    
    
lemma for_gfp_invr_hoare_rel_sp_partial:
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u" 
  assumes induct_step: "\<lbrace>b \<and> invar\<rbrace>body;;incr\<lbrace>q'\<rbrace>\<^sub>u"
  assumes I0: "`q'\<Rightarrow> invar`"  
  shows "\<lbrace>p\<rbrace>for\<^sup>\<top> (init, b, incr) invr invar do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  using I0 for_gfp_invr_hoare_rel_minimal_partial induct_step post_weak_hoare_rel seq_step 
  by blast 
    
lemma for_lfp_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body ;; incr\<lbrace>invar \<and> (e,\<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init, b, incr) do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding for_lfp_rel_def
  by (auto intro!: assms from_until_lfp_hoare_rel_minimal)

lemma for_lfp_invr_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body ;; incr\<lbrace>invar \<and> (e,\<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init, b, incr) invr invar do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding for_lfp_invr_rel_def for_lfp_rel_def
  by (simp add: from_until_lfp_hoare_rel_minimal [OF WF seq_step, of _ e] induct_step)  

lemma for_lfp_invr_vrt_hoare_rel_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body ;; incr\<lbrace>invar \<and> (e,\<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u" 
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init, b, incr) invr invar vrt \<guillemotleft>R\<guillemotright> do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding for_lfp_invr_vrt_rel_def for_lfp_rel_def
  by (simp add: from_until_lfp_hoare_rel_minimal [OF WF seq_step, of _ e] induct_step)      

lemma for_lfp_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step:  "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;;incr\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init, b, incr) do body od\<lbrace>q\<rbrace>\<^sub>u"  
  unfolding for_lfp_rel_def
  by(simp add: from_until_lfp_hoare_rel_consequence [OF WF seq_step PHI _ induct_step I0] I0')

lemma for_lfp_invr_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step:  "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;;incr\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init, b, incr) invr invar do body od\<lbrace>q\<rbrace>\<^sub>u"  
  unfolding for_lfp_rel_def for_lfp_invr_rel_def
  by(simp add: from_until_lfp_hoare_rel_consequence [OF WF seq_step PHI _ induct_step I0] I0')

lemma for_lfp_invr_vrt_hoare_rel_consequence:
  assumes WF: "wf R"
  assumes seq_step:  "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;;incr\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init, b, incr) invr invar vrt \<guillemotleft>R\<guillemotright> do body od\<lbrace>q\<rbrace>\<^sub>u"  
  unfolding for_lfp_rel_def for_lfp_invr_vrt_rel_def
  by(simp add: from_until_lfp_hoare_rel_consequence [OF WF seq_step PHI _ induct_step I0] I0')
    
lemma for_lfp_hoare_rel_wp:
  assumes WF : "wf R"  
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e=\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;;incr\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init,b,incr) do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding for_lfp_rel_alt_def
  by (simp add: uimp_refl seq_hoare_rel[OF seq_step] 
                while_lfp_hoare_rel_wp[OF WF PHI I0' induct_step])

lemma for_lfp_invr_hoare_rel_wp:
  assumes WF : "wf R"  
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e=\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;;incr\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init,b,incr) invr invar do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding for_lfp_invr_rel_def for_lfp_rel_alt_def from_until_lfp_rel_alt_def
  by (simp add: seq_hoare_rel[OF seq_step] 
                while_lfp_hoare_rel_wp[OF WF PHI I0' induct_step])

lemma for_lfp_invr_vrt_hoare_rel_wp:
  assumes WF : "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>u"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e=\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;;incr\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>u"
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init,b,incr) invr invar vrt \<guillemotleft>R\<guillemotright> do body od\<lbrace>q\<rbrace>\<^sub>u"
  unfolding for_lfp_invr_vrt_rel_def for_lfp_rel_alt_def from_until_lfp_rel_alt_def
  by (simp add: uimp_refl seq_hoare_rel[OF seq_step] 
                while_lfp_hoare_rel_wp[OF WF  PHI I0' induct_step])

lemma for_lfp_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>u"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body;;incr\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init,b,incr) do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding for_lfp_rel_alt_def  
  by (simp add: seq_hoare_rel[OF seq_step] pre_str_hoare_rel[OF I0']
                while_lfp_hoare_rel_sp[OF WF uimp_refl induct_step I0])
 
lemma for_lfp_invr_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>u"
  assumes I0':  "`q'' \<Rightarrow> invar`"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body;;incr\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init,b,incr) invr invar do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding for_lfp_invr_rel_def for_lfp_rel_alt_def from_until_lfp_rel_alt_def
  by (simp add: seq_hoare_rel[OF seq_step] pre_str_hoare_rel[OF I0']
                while_lfp_hoare_rel_sp[OF WF uimp_refl induct_step I0])
         
lemma for_lfp_invr_vrt_hoare_rel_sp:
  assumes WF: "wf R"
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>u"
  assumes I0':"`q''\<Rightarrow> invar`" 
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body;;incr\<lbrace>q' st\<rbrace>\<^sub>u"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>for\<^sub>\<bottom> (init,b,incr) invr invar vrt \<guillemotleft>R\<guillemotright> do body od\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>u"
  unfolding for_lfp_invr_vrt_rel_def for_lfp_rel_alt_def  from_until_lfp_rel_alt_def
  by (simp add: seq_hoare_rel[OF seq_step] pre_str_hoare_rel[OF I0']
                while_lfp_hoare_rel_sp[OF WF uimp_refl induct_step I0])
      
lemmas loop_invr_vrt_hoare_rel_sp_instantiated [hoare_sp_rules] = 
       while_lfp_invr_vrt_hoare_rel_sp [where e = "&\<Sigma>"]
       for_lfp_invr_vrt_hoare_rel_sp [where e = "&\<Sigma>"]
       do_while_lfp_invr_vrt_hoare_rel_sp [where e = "&\<Sigma>"]
       from_until_lfp_invr_vrt_hoare_rel_sp[where e = "&\<Sigma>"]
       while_gfp_invr_hoare_rel_sp_partial
       do_while_gfp_invr_hoare_rel_sp_partial
       for_gfp_invr_hoare_rel_sp_partial
       from_until_gfp_invr_hoare_rel_sp_partial

lemmas loop_invr_vrt_hoare_rel_wp_instantiated [hoare_wp_rules] = 
       while_lfp_invr_vrt_hoare_rel_wp [where e = "&\<Sigma>"] 
       for_lfp_invr_vrt_hoare_rel_wp [where e = "&\<Sigma>"]
       do_while_lfp_invr_vrt_hoare_rel_wp [where e = "&\<Sigma>"]
       from_until_lfp_invr_vrt_hoare_rel_wp[where e = "&\<Sigma>"]
       while_gfp_invr_hoare_rel_wp_partial
       do_while_gfp_invr_hoare_rel_wp_partial
       for_gfp_invr_hoare_rel_wp_partial
       from_until_gfp_invr_hoare_rel_wp_partial

end

