(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)

theory utp_hoare_ndes_prog

  imports "../../AlgebraicLaws/algebraic_laws_prog"

begin
chapter \<open>Axiomatic semantics for the theory of total correctness\<close>

text \<open>This theory introduces a mechanization of the axiomatic semantics known as Hoare logic in UTP.
      A Hoare triple is composed of:
      \<^item> Precondition: an assumption on the initial state of the program
      \<^item> Postcondition: an assertion on the final state of the program
      \<^item> Program: a relation between the initial and the final state of the program
      Hoare logic allows for forward and backward VCG-based reasoning:
      \<^item> Weakest Precondition(WP): where we fix the postcondition and the program and we compute
        the weakest precondition. It is also known as backward reasoning.
      \<^item> Strongest Postcondition(SP): where we fix the precondition and the program and we compute
        the strongest postcondition. It is also known as forward reasoning.
      In the sequel we derive from the defined denotational semantics the basic rules for Hoare 
      logic as well as the backward and forward rules. The novelty are the rules for scoping. The latter
      are derived from frame and antiframe programming constructs. These rules are used for providing
      modular VCG-based reasoning. The combination of the semantics of frame and antiframe together
      with the denotational semantics of iterations leads to an extension of hoare logic with 
      new derived rules to reason about iterations. \<close>

section \<open>Hoare triple definition\<close>

lift_definition hoare_prog_t :: \<open>'\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> bool\<close> ("\<lbrace>(1_)\<rbrace>/(_)/\<lbrace>(1_)\<rbrace>\<^sub>P") 
 is  hoare_d.

declare hoare_prog_t.rep_eq [prog_rep_eq]

lemma hoare_prog_alt_def:  
  \<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P = (Abs_prog ((P \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>)) \<sqsubseteq> PROG)\<close>
  apply (simp only: prog_rep_eq hoare_d_def)
  using utp_designs.ndesign_H1_H3[of P "\<lceil>Q\<rceil>\<^sub>>"] Abs_prog_inverse[of "(P \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>)"]
  apply simp
  done  

section \<open>Consequence rules for Hoare Logic\<close>

lemma pre_weak_prog_hoare:
  assumes \<open>`P \<Rightarrow> R`\<close> 
  assumes \<open>\<lbrace>R\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>
  shows   \<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  using assms  
  by (simp add: prog_rep_eq hoare_des)

lemma post_str_prog_hoare:
  assumes \<open>\<lbrace>P\<rbrace>PROG\<lbrace>R\<rbrace>\<^sub>P\<close> 
  assumes \<open>`R \<Rightarrow> Q`\<close>
  shows   \<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  using assms  
  by (simp add: prog_rep_eq hoare_des)
 
lemma consequence_prog_hoare:
  assumes I0': "`P \<Rightarrow> R`" 
  assumes induct_step:"\<lbrace>R\<rbrace>PROG\<lbrace>S\<rbrace>\<^sub>P" 
  assumes I0 :"`S \<Rightarrow> Q`"  
  shows "\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P"
proof(rule post_str_prog_hoare[OF _ I0], goal_cases)
  case 1
  then show ?case by (rule pre_weak_prog_hoare[OF I0' induct_step ]) 
qed   
  
section \<open>Hoare Logic and Assertion Logic\<close>

lemma assisgns_true_prog_hoare: 
  \<open>\<lbrace>P\<rbrace>\<langle>\<sigma>\<rangle>\<^sub>p\<lbrace>true\<rbrace>\<^sub>P\<close>
  by (simp add: prog_rep_eq hoare_des)

lemma skip_true_prog_hoare: 
  \<open>\<lbrace>P\<rbrace>SKIP\<lbrace>true\<rbrace>\<^sub>P\<close>
  by (simp add: prog_rep_eq hoare_des)

lemma false_prog_hoare: 
  \<open>\<lbrace>false\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>
  by (simp add: prog_rep_eq hoare_des)


lemma pre_conj_split_prog_hoare:
  assumes \<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>
  assumes \<open>\<lbrace>R\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>
  shows   \<open>\<lbrace>P \<and> R\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>
  using assms
  apply (simp add: prog_rep_eq)
  apply rel_simp
  done

lemma post_conj_split_prog_hoare:
  assumes \<open>\<lbrace>P\<rbrace>PROG\<lbrace>R\<rbrace>\<^sub>P\<close>
  assumes \<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>
  shows   \<open>\<lbrace>P \<rbrace>PROG\<lbrace>Q \<and> R\<rbrace>\<^sub>P\<close>
  using assms
  apply (simp add: prog_rep_eq)
  apply rel_simp
  done

section \<open>Hoare Logic for SKIP\<close>

lemma skip_prog_hoare[hoare_sp_rules, hoare_wp_rules]: 
  \<open>\<lbrace>P\<rbrace>SKIP\<lbrace>P\<rbrace>\<^sub>P\<close>
  by (simp add: prog_rep_eq hoare_des)
    
lemma skip_prog_hoare_intro: 
  \<open>`P \<Rightarrow> Q`\<Longrightarrow>\<lbrace>P\<rbrace>SKIP\<lbrace>Q\<rbrace>\<^sub>P\<close>
  by (simp add: prog_rep_eq hoare_des)
    
section \<open>Hoare Logic for assignment\<close>

lemma assigns_prog_hoare_backward_intro: 
  assumes \<open>`P \<Rightarrow> \<sigma> \<dagger> Q`\<close> 
  shows   \<open>\<lbrace>P\<rbrace>\<langle>\<sigma>\<rangle>\<^sub>p\<lbrace>Q\<rbrace>\<^sub>P\<close>
  using assms  
  by (simp add: prog_rep_eq hoare_des)
 
lemma assigns_prog_hoare_backward[hoare_wp_rules]: 
  \<open>\<lbrace>\<sigma> \<dagger> Q\<rbrace>\<langle>\<sigma>\<rangle>\<^sub>p\<lbrace>Q\<rbrace>\<^sub>P\<close>
  by (simp add: prog_rep_eq hoare_des)

lemma assigns_prog_hoare_backward':
  \<open>\<lbrace>P\<lbrakk>e/x\<rbrakk> \<rbrace>x :== e\<lbrace>P\<rbrace>\<^sub>P\<close>
  apply (rule pre_weak_prog_hoare)
   defer
   apply (rule assigns_prog_hoare_backward)
  apply rel_simp
  done

lemma assigns_prog_floyd':
  \<open>\<lbrace>P\<rbrace>\<langle>\<sigma>\<rangle>\<^sub>p
   \<lbrace>\<^bold>\<exists>v \<bullet> ((subst_upd id (1\<^sub>L) \<guillemotleft>v\<guillemotright>) \<dagger> P) \<and> 
         &\<Sigma> =\<^sub>u ((subst_upd id (1\<^sub>L) \<guillemotleft>v\<guillemotright>) \<dagger> (usubst_lookup \<sigma> 1\<^sub>L))\<rbrace>\<^sub>P\<close>
 apply (simp add: prog_rep_eq )
  apply rel_simp
  apply auto
  done

lemma assigns_prog_floyd[hoare_sp_rules]:
  assumes \<open>vwb_lens x\<close>
  shows   \<open>\<lbrace>P\<rbrace>x :== e
           \<lbrace>\<^bold>\<exists>v \<bullet> P\<lbrakk>\<guillemotleft>v\<guillemotright>/x\<rbrakk> \<and> &x =\<^sub>u e\<lbrakk>\<guillemotleft>v\<guillemotright>/x\<rbrakk>\<rbrace>\<^sub>P\<close>
  apply (rule post_str_prog_hoare)
   apply (rule assigns_prog_floyd')
  apply rel_simp
  subgoal for \<sigma>
    apply (rule exI[where x = \<open>get\<^bsub>x\<^esub> \<sigma>\<close>])
    using assms 
    apply simp
    done
  done

section \<open>Hoare Logic for Sequential Composition\<close>

lemma seq_prog_hoare[hoare_wp_rules, hoare_sp_rules]:
  assumes \<open>\<lbrace>P\<rbrace>PROG\<^sub>1\<lbrace>R\<rbrace>\<^sub>P\<close> 
  assumes \<open>\<lbrace>R\<rbrace>PROG\<^sub>2\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  shows   \<open>\<lbrace>P\<rbrace>PROG\<^sub>1;PROG\<^sub>2\<lbrace>Q\<rbrace>\<^sub>P\<close>
  using assms
  by (simp add: prog_rep_eq hoare_des)  

section \<open>Hoare for Conditional\<close>

lemma cond_prog_hoare_basic:
  assumes \<open>`P \<Rightarrow> (b \<and> S) \<or> (\<not>b \<and> R)`\<close>
  assumes \<open>\<lbrace>S\<rbrace>PROG\<^sub>1\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  assumes \<open>\<lbrace>R\<rbrace>PROG\<^sub>2\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  shows   \<open>\<lbrace>P\<rbrace>IF b THEN PROG\<^sub>1 ELSE PROG\<^sub>2 FI\<lbrace>Q\<rbrace>\<^sub>P\<close>
  using assms
  apply (simp add: prog_rep_eq ) 
  apply rel_blast
  done

lemma cond_prog_hoare_wp[hoare_wp_rules]: 
  assumes \<open>\<lbrace>P'\<rbrace>PROG\<^sub>1\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  assumes \<open>\<lbrace>P''\<rbrace>PROG\<^sub>2\<lbrace>Q\<rbrace>\<^sub>P\<close>
  shows \<open>\<lbrace>(b \<and> P') \<or> (\<not>b \<and> P'')\<rbrace>
         IF b THEN PROG\<^sub>1 ELSE PROG\<^sub>2 FI
         \<lbrace>Q\<rbrace>\<^sub>P\<close>
  apply (rule cond_prog_hoare_basic)
    defer
    apply (rule assms(1))
  apply (rule assms(2))  
  apply pred_simp
  done
    
lemma cond_prog_hoare_sp[hoare_sp_rules]:
  assumes \<open>\<lbrace>b \<and> P\<rbrace>PROG\<^sub>1\<lbrace>Q\<rbrace>\<^sub>P\<close> and \<open>\<lbrace>\<not>b \<and> P\<rbrace>PROG\<^sub>2\<lbrace>R\<rbrace>\<^sub>P\<close>
  shows \<open>\<lbrace>P\<rbrace>IF b THEN PROG\<^sub>1 ELSE PROG\<^sub>2 FI\<lbrace>Q \<or> R\<rbrace>\<^sub>P\<close>
  using assms
  by (simp add: prog_rep_eq hoare_des) 

section \<open>Hoare Logic for frame and antiframe\<close>

subsection \<open>antiframe\<close>
    

lemma antiframe_prog_hoare_basic_weak:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_g':\<open>vwb_lens g'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>g \<bowtie> l\<close> 
  assumes region     :\<open>g' \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>{&g', &l}:[PROG] = PROG\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>  
  shows \<open>\<lbrace>P \<and> R\<rbrace>MODIFY g' DO PROG OD\<lbrace>(\<exists>l \<bullet> Q) \<and>  (\<exists>g' \<bullet> R)\<rbrace>\<^sub>P\<close>
  using vwb_lens_g vwb_lens_g' vwb_lens_l indep_g_l modifies hoare  
  unfolding lens_indep_def
  apply (simp add:prog_rep_eq)
  apply rel_simp
  using indep_g_l 
  apply rel_simp
  apply auto
   apply(insert region)
  subgoal for  _ more _ x
    apply (rule exI[where x = \<open>get\<^bsub>l\<^esub>x\<close>])
    using region unfolding lens_indep_def
    apply auto
    apply (drule  sublens_pres_indep[OF _ indep_g_l])
    unfolding lens_indep_def
    apply simp
    apply blast
    done
  subgoal for  _ more _ _
    apply (rule exI[where x = \<open>get\<^bsub>g'\<^esub>more\<close>])
    unfolding lens_indep_def
    apply (smt lens_put_of_quotient mwb_lens.put_put vwb_lens_mwb vwb_lens_wb wb_lens.get_put)
    done
  done

lemma antiframe_prog_hoare_basic_strong:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_g':\<open>vwb_lens g'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>g \<bowtie> l\<close> 
  assumes region     :\<open>g' \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>PROG \<sqsubseteq> MODIFY g'+\<^sub>L l DO PROG OD\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>  
  shows   \<open>\<lbrace>P \<and> R\<rbrace>MODIFY g' DO PROG OD\<lbrace>(\<exists>l \<bullet> Q) \<and>  (\<exists>g' \<bullet> R)\<rbrace>\<^sub>P\<close>
  using vwb_lens_g vwb_lens_g' vwb_lens_l indep_g_l  hoare  
  unfolding lens_indep_def
  apply (simp add:prog_rep_eq)
  apply rel_simp
  using indep_g_l 
  apply rel_simp
  apply auto
   apply(insert region)
  subgoal for  _ more _ x
    apply (rule exI[where x ="get\<^bsub>l\<^esub>x"])
    using region modifies unfolding lens_indep_def
    apply auto
    apply (drule sublens_pres_indep[OF _ indep_g_l])
    unfolding lens_indep_def
    apply (simp add:prog_rep_eq)
    apply rel_simp
    apply blast
    done
  subgoal for  _ more _ _
    apply (rule exI[where x ="get\<^bsub>g'\<^esub>more"])
    unfolding lens_indep_def
    apply (smt lens_put_of_quotient mwb_lens.put_put vwb_lens_mwb vwb_lens_wb wb_lens.get_put)
    done
  done

lemma antiframe_prog_hoare_wp_weak:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_g':\<open>vwb_lens g'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>g \<bowtie> l\<close> 
  assumes region     :\<open>g' \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY g' +\<^sub>L l DO PROG OD = PROG\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  assumes conseq     :\<open>`(\<exists>l \<bullet> Q) \<and> (\<exists>g' \<bullet> R) \<Rightarrow> R`\<close> 
  shows   \<open>\<lbrace>P \<and>  (\<exists>g' \<bullet> R)\<rbrace>MODIFY g' DO PROG OD\<lbrace>R\<rbrace>\<^sub>P\<close>
  apply (rule post_str_prog_hoare)
   apply (rule antiframe_prog_hoare_basic_weak[unfolded pr_var_def, 
                                               OF vwb_lens_g vwb_lens_g' vwb_lens_l indep_g_l 
                                                  region modifies hoare, 
                                               simplified])
  apply (simp add: conseq exists_twice vwb_lens_g')
  done

lemma antiframe_prog_hoare_wp_strong:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_g':\<open>vwb_lens g'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>g \<bowtie> l\<close> 
  assumes region     :\<open>g' \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>PROG \<sqsubseteq> MODIFY g' +\<^sub>L l DO PROG OD\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  assumes conseq     :\<open>`(\<exists>l \<bullet> Q) \<and> (\<exists>g' \<bullet> R) \<Rightarrow> R`\<close> 
  shows   \<open>\<lbrace>P \<and>  (\<exists>g' \<bullet> R)\<rbrace>MODIFY g' DO PROG OD\<lbrace>R\<rbrace>\<^sub>P\<close>
  apply (rule post_str_prog_hoare)
   apply (rule antiframe_prog_hoare_basic_strong[unfolded pr_var_def,
                                                 OF vwb_lens_g vwb_lens_g' vwb_lens_l indep_g_l 
                                                    region modifies hoare, 
                                                 simplified])
  apply (simp add: conseq exists_twice vwb_lens_g')
  done

lemma antiframe_prog_hoare_sp_weak:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_l':\<open>vwb_lens g'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>g \<bowtie> l\<close> 
  assumes region     :\<open>g' \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY g' +\<^sub>L l DO PROG OD = PROG\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>
  assumes conseq     :\<open>`R \<Rightarrow> P`\<close>     
  shows   \<open>\<lbrace>R\<rbrace>MODIFY g' DO PROG OD\<lbrace>(\<exists> l \<bullet> Q) \<and> (\<exists>g' \<bullet> R)\<rbrace>\<^sub>P\<close>
  apply (rule pre_weak_prog_hoare)
  defer
   apply (rule antiframe_prog_hoare_basic_weak[unfolded pr_var_def,
                                               OF vwb_lens_g vwb_lens_l' vwb_lens_l indep_g_l 
                                                  region modifies hoare])
  apply (simp add: conseq refBy_order utp_pred_laws.inf.absorb2)
  done

lemma antiframe_prog_hoare_sp_strong:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_g':\<open>vwb_lens g'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>g \<bowtie> l\<close> 
  assumes region     :\<open>g' \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>PROG \<sqsubseteq> MODIFY g' +\<^sub>L l DO PROG OD\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>  
  assumes conseq     :\<open>`R \<Rightarrow> P`\<close>     
  shows   \<open>\<lbrace>R\<rbrace> MODIFY g' DO PROG OD \<lbrace>(\<exists> l \<bullet> Q) \<and> (\<exists>g' \<bullet> R)\<rbrace>\<^sub>P\<close>
  apply (rule pre_weak_prog_hoare)
   defer
   apply (rule antiframe_prog_hoare_basic_strong[unfolded pr_var_def,
                                                 OF vwb_lens_g vwb_lens_g' vwb_lens_l 
                                                    indep_g_l region modifies hoare])
  apply (simp add: conseq refBy_order utp_pred_laws.inf.absorb2)
  done

subsection \<open>frame\<close>
(*TODO: MOVE THIS DEFINITION TO THE RIGHT PLACE*)
definition FRAME_TRIPLE :: \<open>bool \<Rightarrow> bool\<close> where \<open>FRAME_TRIPLE P = P\<close>

lemma frame_prog_hoare_basic_weak:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_l':\<open>vwb_lens l'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>l  \<bowtie> g\<close> 
  assumes region     :\<open>l' \<subseteq>\<^sub>L l\<close>
  assumes modifies   :\<open>MODIFY g +\<^sub>L l' DO PROG OD = PROG\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>  
  shows   \<open>\<lbrace>P \<and> R\<rbrace> RESTOR l' DO PROG OD\<lbrace>(\<exists>l' \<bullet> Q) \<and>  (\<exists>g \<bullet> R)\<rbrace>\<^sub>P\<close>
  using vwb_lens_g vwb_lens_l' vwb_lens_l indep_g_l modifies hoare  
  unfolding lens_indep_def
  apply (simp add:prog_rep_eq)
  apply rel_simp
  using indep_g_l 
  apply rel_simp
  apply auto
   apply (metis lens_override_def lens_override_idem)
  apply(insert region)
  apply (drule sublens_pres_indep)
   apply assumption
  unfolding lens_indep_def
  apply auto
  apply (fold lens_override_def)
    apply (metis "vwb_lens_l'" lens_override_def mwb_lens.put_put vwb_lens.put_eq vwb_lens_mwb)
  done

lemma frame_prog_hoare_wp_weak:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_l':\<open>vwb_lens l'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>l \<bowtie> g\<close> 
  assumes region     :\<open>l' \<subseteq>\<^sub>L l\<close>
  assumes modifies   :\<open>MODIFY g +\<^sub>L l' DO PROG OD = PROG\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>  
  assumes conseq     :\<open>`(\<exists>l' \<bullet> Q) \<Rightarrow> R`\<close>     
  shows   \<open>\<lbrace>P \<and> (\<exists>g \<bullet> R)\<rbrace>RESTOR l' DO PROG OD\<lbrace>R\<rbrace>\<^sub>P\<close>
  apply (rule post_str_prog_hoare)
   apply (rule frame_prog_hoare_basic_weak[OF vwb_lens_g vwb_lens_l' vwb_lens_l   
                                              indep_g_l region modifies hoare])
  using conseq refBy_order utp_pred_laws.le_infI1 
  apply blast
  done

lemma frame_prog_hoare_sp_weak:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_l':\<open>vwb_lens l'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>l \<bowtie> g\<close> 
  assumes region     :\<open>l' \<subseteq>\<^sub>L l\<close>
  assumes modifies   :\<open>MODIFY g +\<^sub>L l' DO PROG OD = PROG\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  assumes conseq     :\<open>`R \<Rightarrow> P`\<close>     
  shows   \<open>\<lbrace>R\<rbrace> RESTOR l' DO PROG OD \<lbrace>(\<exists>l' \<bullet> Q) \<and> (\<exists>g \<bullet> R)\<rbrace>\<^sub>P\<close>
  apply (rule pre_weak_prog_hoare)
   defer
   apply (rule frame_prog_hoare_basic_weak[OF vwb_lens_g vwb_lens_l' vwb_lens_l   
                                              indep_g_l region modifies hoare])  
  apply (simp add: conseq refBy_order utp_pred_laws.inf.absorb2)
  done

section \<open>Hoare Logic for recursion\<close>

lemma lfp_prog_wf_refine_intro: 
  assumes WF         : \<open>wf R\<close>
  assumes M          : \<open>mono_prog F\<close>
  assumes induct_step:
    \<open>\<And>st. Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> (F (Abs_prog ((P \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q)))\<close>
  shows \<open>Abs_prog ((P \<turnstile>\<^sub>n Q)) \<sqsubseteq> \<mu>\<^sub>p F\<close>            
proof -          
  {
    fix st
    have "Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq>  \<mu>\<^sub>p F" 
      using WF proof (induction rule: wf_induct_rule)
      case (less st)
      hence 0: "Abs_prog ((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq>  \<mu>\<^sub>p F"
        apply (simp add: prog_rep_eq)
        apply (simp add: utp_designs.ndesign_H1_H3 Abs_prog_inverse)
        apply rel_simp
        apply (metis (no_types, lifting))
        done
      from M 
      have 1: "\<mu>\<^sub>p F \<sqsubseteq>  F (\<mu>\<^sub>p F)"
        by (simp add: lfp_prog_fixed_point)
      from 0 1 have 2:"Abs_prog((P \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> F (\<mu>\<^sub>p F)"
        by simp
      have 3: "F (Abs_prog ((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q)) \<sqsubseteq> F (\<mu>\<^sub>p F)"
        by (auto intro: Mono_progD M 0)
      have 4:"Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> \<dots>" 
        by (rule induct_step)
      show ?case
        using order_trans[OF 3 4]  M 
      proof -
        show ?thesis
          using \<open>Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> F (\<mu>\<^sub>p F)\<close>[unfolded lfp_prog_fixed_point[OF M]]
          by simp
      qed
    qed
  }
  thus ?thesis
    apply (simp add: prog_rep_eq)
    apply (simp add: utp_designs.ndesign_H1_H3 Abs_prog_inverse)
    unfolding comp_def
    apply rel_simp
    done
qed 

lemma lfp_prog_wf_refine_intro': 
  assumes WF          :\<open>wf R\<close>
  assumes M           :\<open>mono_prog F\<close>
  assumes  induct_step:
    \<open>\<And>PROG st. (Abs_prog ((P \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q)) \<sqsubseteq> PROG \<Longrightarrow> 
             \<mu>\<^sub>p F = PROG \<Longrightarrow>
             Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> F PROG\<close>
  shows \<open>Abs_prog ((P \<turnstile>\<^sub>n Q)) \<sqsubseteq> \<mu>\<^sub>p F\<close>      
proof -          
  {
    fix st
    have "Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq>  \<mu>\<^sub>p F" 
      using WF       
      apply (induction rule: wf_induct_rule) 
      apply (subst lfp_prog_unfold[OF M])
      apply (rule induct_step)
       apply (simp add: prog_rep_eq)
        apply (simp add: utp_designs.ndesign_H1_H3 Abs_prog_inverse)
       apply rel_blast
      apply simp
      done
  }
  thus ?thesis
    apply (simp add: prog_rep_eq)
    apply (simp add: utp_designs.ndesign_H1_H3 Abs_prog_inverse)
    apply rel_simp
    done
qed 

lemma lfp_prog_hoare_prog:
  assumes WF         :\<open>wf R\<close>
  assumes M          :\<open>mono_prog F\<close>  
  assumes induct_step:
    \<open>\<And>st PROG. \<lbrace>P \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace> PROG \<lbrace>Q\<rbrace>\<^sub>P \<Longrightarrow> \<lbrace>P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace> F PROG \<lbrace>Q\<rbrace>\<^sub>P\<close>   
  shows \<open>\<lbrace>P\<rbrace>\<mu>\<^sub>p F \<lbrace>Q\<rbrace>\<^sub>P\<close>   
 unfolding hoare_prog_alt_def
  apply (rule lfp_prog_wf_refine_intro[OF  WF M, of _ st'])
  apply (rule induct_step[unfolded hoare_prog_alt_def])
  apply (rule order.refl)
  done

lemma lfp_prog_hoare_prog_wp:
  assumes WF         :\<open>wf R\<close>
  assumes M          :\<open>mono_prog F\<close>  
  assumes I0'        :\<open>\<And>st. `invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:
    \<open>\<And> st PROG. \<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace> PROG \<lbrace>Q\<rbrace>\<^sub>P \<Longrightarrow>  \<lbrace>P' st\<rbrace> F PROG \<lbrace>Q\<rbrace>\<^sub>P\<close>    
  shows \<open>\<lbrace>invar\<rbrace>\<mu>\<^sub>p F \<lbrace>Q\<rbrace>\<^sub>P\<close>  
   apply (rule lfp_prog_hoare_prog[OF WF M])
  apply (rule pre_weak_prog_hoare[OF I0'])
  apply (rule induct_step)
   apply assumption
  done

lemma lfp_prog_hoare_prog':
  assumes WF         :\<open>wf R\<close>
  assumes M          :\<open>mono_prog F\<close>  
  assumes induct_step:
    \<open>\<And> st PROG. \<lbrace>P \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace> PROG \<lbrace>Q\<rbrace>\<^sub>P \<Longrightarrow> \<lbrace>P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace> F PROG \<lbrace>Q\<rbrace>\<^sub>P\<close> 
  assumes I0: \<open>`P' \<Rightarrow> P`\<close>  
  shows   \<open>\<lbrace>P'\<rbrace>\<mu>\<^sub>p F \<lbrace>Q\<rbrace>\<^sub>P\<close>
  by (simp add: pre_weak_prog_hoare[OF I0 lfp_prog_hoare_prog[OF WF M induct_step]])

section \<open>Hoare Logic for scoped recursion\<close>

lemma lfp_scp_prog_wf_refine_intro_strong: 
  assumes vwb_lens:\<open>vwb_lens G\<close>
  assumes WF      :\<open>wf R\<close>
  assumes M       :\<open>mono_prog F\<close>
  assumes H       :\<open>(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. MODIFY G DO P OD = P}\<close>
  assumes  induct_step:
    \<open>\<And>st. Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> (F (Abs_prog ((P \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q)))\<close>
  shows \<open>Abs_prog (P \<turnstile>\<^sub>n Q) \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F\<close>            
proof -          
  {
  fix st
  have "Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F" 
    using WF proof (induction rule: wf_induct_rule)
    case (less st)
    hence 0: "Abs_prog ((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F"
       apply (simp add: prog_rep_eq)
      apply (simp add: utp_designs.ndesign_H1_H3 Abs_prog_inverse)
      unfolding comp_def
      apply rel_simp
       apply (metis (no_types, lifting))
      done
    from M H 
    have 1: "\<mu>\<^bsub>G\<^esub> F \<sqsubseteq>  F (\<mu>\<^bsub>G\<^esub> F)"
      by (simp add: lfp_scp_prog_fixed_point_strong)
    from 0 1 have 2:"Abs_prog((P \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> F (\<mu>\<^bsub>G\<^esub> F)"
      by simp
    have 3: "F (Abs_prog ((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q)) \<sqsubseteq> F (\<mu>\<^bsub>G\<^esub> F)"
      by (auto intro: Mono_progD M 0)
    have 4:"Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> \<dots>" 
      by (rule induct_step)
    show ?case
      using order_trans[OF 3 4] H M 
    proof -
      show ?thesis
        by (metis (no_types) H M \<open>Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> F (\<mu>\<^bsub>G\<^esub> F)\<close> lfp_scp_prog_unfold_strong)
    qed
  qed
  }
  thus ?thesis
    apply (simp add: prog_rep_eq)
      apply (simp add: utp_designs.ndesign_H1_H3 Abs_prog_inverse)
      unfolding comp_def
      apply rel_simp
      done
  qed 

lemma lfp_scp_prog_wf_refine_intro_weak: 
  assumes vwb_lens   :\<open>vwb_lens G\<close>
  assumes WF         :\<open>wf R\<close>
  assumes M          :\<open>mono_prog F\<close>
  assumes H          :\<open>(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. MODIFY G DO P OD = P}\<close>
  assumes induct_step:
    \<open>\<And>st. Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> 
          ((F o pantiframe_prog G) (Abs_prog ((P \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q)))\<close>
  shows \<open>Abs_prog (P \<turnstile>\<^sub>n Q) \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F\<close>            
proof -          
  {
    fix st
    have "Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F" 
      using WF proof (induction rule: wf_induct_rule)
      case (less st)
      hence 0: "Abs_prog ((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F"
        apply (simp add: prog_rep_eq)
        apply (simp add: utp_designs.ndesign_H1_H3 Abs_prog_inverse)
        unfolding comp_def
        apply rel_simp
        apply (metis (no_types, lifting))
        done
      from M H 
      have 1: "\<mu>\<^bsub>G\<^esub> F \<sqsubseteq> (F o pantiframe_prog G) (\<mu>\<^bsub>G\<^esub> F)"
        using lfp_scp_prog_unfold_strong mu_scp_antiframe_modifies by fastforce
      from 0 1 have 2:"Abs_prog((P \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> (F o pantiframe_prog G) (\<mu>\<^bsub>G\<^esub> F)"
        by simp
      have 3: "(F o pantiframe_prog G) (Abs_prog ((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n Q)) \<sqsubseteq> 
               (F o pantiframe_prog G) (\<mu>\<^bsub>G\<^esub> F)"
        unfolding comp_def
        using M
        apply (rule Mono_progD)
        apply (rule antiframe_prog_monoI)
        apply (rule 0[unfolded comp_def])
        done
      have 4:"Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n Q) \<sqsubseteq> \<dots>" 
        by (rule induct_step)
      show ?case
        apply (rule order_trans[OF  _ 4])
        apply (subst lfp_scp_prog_unfold_weak[OF M H])
        apply (rule 3)
        done   
    qed
  }
  thus ?thesis
    apply (simp add: prog_rep_eq)
    apply (simp add: utp_designs.ndesign_H1_H3 Abs_prog_inverse)
    unfolding comp_def
    apply rel_simp
    done
qed  

lemma lfp_scp_prog_wf_region_refine_intro_strong:
  assumes vwb_lens1     :\<open>vwb_lens g\<close>
  assumes vwb_lens2     :\<open>vwb_lens G\<close>
  assumes vwb_lens3     :\<open>vwb_lens l\<close>
  assumes lens_indep_g_l:\<open>g \<bowtie> l\<close>
  assumes region        :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes WF            :\<open>wf R\<close>
  assumes M             :\<open>mono_prog F\<close>
  assumes H             :\<open>(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. MODIFY G DO P OD = P}\<close>
  assumes induct_step   :
    \<open>\<And>st. Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>) \<sqsubseteq> F (Abs_prog ((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>))\<close>
  shows \<open>Abs_prog ((P \<and> S) \<turnstile>\<^sub>n \<lceil>(\<exists>l \<bullet> Q) \<and> (\<exists>G \<bullet> S)\<rceil>\<^sub>>) \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F\<close>
  apply (subst mu_scp_antiframe_modifies[THEN sym, OF M H])
      apply (rule antiframe_prog_hoare_basic_strong[unfolded hoare_prog_alt_def, where g'= G and g=g,
                                                    OF vwb_lens1 vwb_lens2 vwb_lens3 lens_indep_g_l region])
   apply (subst mu_scp_antiframe_modifies[OF M])
    apply (insert H)
    apply auto[]
    apply (subst (asm) Pi_iff)
    apply auto[]
    apply (rule modifies_indep[OF vwb_lens2 vwb_lens3])
  using lens_indep_g_l region sublens_pres_indep apply auto[1]    
    apply auto[]
   apply (rule order.refl)
  apply (rule lfp_scp_prog_wf_refine_intro_strong[OF vwb_lens2 WF M H])
  apply (rule induct_step)
  done

lemma lfp_scp_prog_wf_region_refine_intro_weak:
  assumes vwb_lens1     :\<open>vwb_lens g\<close>
  assumes vwb_lens2     :\<open>vwb_lens G\<close>
  assumes vwb_lens3     :\<open>vwb_lens l\<close>
  assumes lens_indep_g_l:\<open>g \<bowtie> l\<close>
  assumes region        :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes WF            :\<open>wf R\<close>
  assumes M             :\<open>mono_prog F\<close>
  assumes H             :\<open>(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. MODIFY G DO P OD = P}\<close>
  assumes  induct_step  :
    \<open>\<And>st. Abs_prog ((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>) \<sqsubseteq> 
          (F o pantiframe_prog G) (Abs_prog ((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>))\<close>
  shows \<open>Abs_prog ((P \<and> S) \<turnstile>\<^sub>n \<lceil>(\<exists>l \<bullet> Q) \<and> (\<exists>G \<bullet> S)\<rceil>\<^sub>>) \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F\<close>
  apply (subst mu_scp_antiframe_modifies[THEN sym, OF M H])
      apply (rule antiframe_prog_hoare_basic_strong[unfolded hoare_prog_alt_def, where g'= G and g=g,
                                                      OF vwb_lens1 vwb_lens2 vwb_lens3 lens_indep_g_l region])
   apply (subst mu_scp_antiframe_modifies[OF M])
    apply (insert H)
    apply auto[]
    apply (subst (asm) Pi_iff)
    apply auto[]
    apply (rule modifies_indep[OF vwb_lens2 vwb_lens3])
  using lens_indep_g_l region sublens_pres_indep apply auto[1]    
    apply auto[]
   apply (rule order.refl)
  apply (rule lfp_scp_prog_wf_refine_intro_weak[OF vwb_lens2 WF M H])
  apply (rule induct_step)
  done

lemma lfp_scp_prog_hoare_strong:
  assumes vwb_lens   :\<open>vwb_lens G\<close>
  assumes WF         :\<open>wf R\<close>
  assumes M          :\<open>mono_prog F\<close>  
  assumes H          :\<open>(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. MODIFY G DO P OD = P}\<close>
  assumes induct_step:
    \<open>\<And> st PROG. \<lbrace>P \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace> PROG \<lbrace>Q\<rbrace>\<^sub>P \<Longrightarrow> \<lbrace>P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace> F PROG \<lbrace>Q\<rbrace>\<^sub>P\<close>   
  shows \<open>\<lbrace>P\<rbrace>\<mu>\<^bsub>G\<^esub> F \<lbrace>Q\<rbrace>\<^sub>P\<close> 
proof -
  { fix st
    have *:"(Abs_prog((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>)) \<sqsubseteq> F (Abs_prog((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>))"
      apply (rule induct_step [unfolded hoare_prog_alt_def])
      apply (rule order.refl)
      done
  }
  then show ?thesis  
  unfolding hoare_prog_alt_def 
  apply (subst lfp_scp_prog_wf_refine_intro_strong[OF vwb_lens WF M H, of _ st'])
  unfolding comp_def
  apply auto
  done
qed

lemma lfp_scp_prog_hoare_weak:
  assumes vwb_lens   :\<open>vwb_lens G\<close>
  assumes WF         :\<open>wf R\<close>
  assumes M          :\<open>mono_prog F\<close>  
  assumes H          :\<open>(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. MODIFY G DO P OD = P}\<close>
  assumes induct_step:
    \<open>\<And> st PROG. \<lbrace>P \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace> PROG \<lbrace>Q\<rbrace>\<^sub>P \<Longrightarrow> 
                \<lbrace>P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace> (F o pantiframe_prog G) PROG \<lbrace>Q\<rbrace>\<^sub>P\<close>   
  shows   \<open>\<lbrace>P\<rbrace>\<mu>\<^bsub>G\<^esub> F \<lbrace>Q\<rbrace>\<^sub>P\<close> 
proof -
  { fix st
    have *:"(Abs_prog((P \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>) \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>)) \<sqsubseteq> 
            (F o pantiframe_prog G) (Abs_prog((P \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>) \<turnstile>\<^sub>n \<lceil>Q\<rceil>\<^sub>>))"
      apply (rule induct_step [unfolded hoare_prog_alt_def])
      apply (rule order.refl)
      done
  }
  then show ?thesis  
  unfolding hoare_prog_alt_def 
  apply (subst lfp_scp_prog_wf_refine_intro_weak[OF vwb_lens WF M H, of _ st'])
  unfolding comp_def
  apply auto
  done
qed

subsection \<open>Hoare for scoped while loop\<close>  

lemma while_scp_prog_hoare_basic:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf R\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P\<close>    
  shows   \<open>\<lbrace>invar \<and> S\<rbrace>IN G WHILE b DO body OD\<lbrace>(\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> S)\<rbrace>\<^sub>P\<close>
  unfolding while_lfp_scp_prog_mu_prog hoare_prog_alt_def
proof (rule lfp_scp_prog_wf_region_refine_intro_strong[where g= g and R= R and st'=st'], goal_cases)
  case 1
  then show ?case using vwb_lens1 by assumption
next
  case 2
  then show ?case using vwb_lens2 by assumption 
next
  case 3
  then show ?case using vwb_lens3 by assumption  
next
  case 4
  then show ?case using lens_indep by assumption 
next
  case 5
  then show ?case using region by assumption 
next
  case 6
  then show ?case using WF by assumption 
next
  case 7
  then show ?case
    by (simp add: if_prog_mono mono_Monotone_prog) 
next
  case 8
  then show ?case  apply auto
    apply (subst If_prog_anti_frame_prog_distr[THEN sym])
    apply (subst seq_prog_anti_frame_prog_distr[THEN sym, OF _ modifies])
    apply (simp add: assms)
    apply (subst antiframe_prog_idemp[OF vwb_lens2])
    apply (subst (2) modifies[THEN sym])
    apply (subst SKIP_prog_anti_frame_prog_modifies[OF vwb_lens2])
    apply (rule refl)
    done 
next
  case (9 st)
  then show ?case 
    apply (rule pre_weak_prog_hoare[unfolded hoare_prog_alt_def])
    defer
    apply (rule cond_prog_hoare_wp[unfolded hoare_prog_alt_def, where P'' = "\<not> b \<and> invar"])
     apply  (rule seq_prog_hoare[unfolded hoare_prog_alt_def])
      apply (rule induct_step[unfolded  hoare_prog_alt_def])
      apply (simp add: prog_rep_eq)
    apply rel_blast
    apply (rule skip_prog_hoare_intro[unfolded hoare_prog_alt_def])
    apply pred_simp+
    done
qed
    
lemma while_scp_prog_hoare_consequence:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> (invar \<and> S)`\<close>
  assumes PHI        :\<open>`((\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> S)) \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `\<close>
  shows   \<open>\<lbrace>P\<rbrace>IN G WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  by (rule consequence_prog_hoare[OF seq_step _ PHI,
                                  OF while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
                                                                   lens_indep region modifies WF],
                                  OF  consequence_prog_hoare[OF I0' induct_step I0]])

lemma while_scp_invr_prog_hoare_consequence:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> (invar \<and> r)`\<close>
  assumes PHI        :\<open>`((\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> r)) \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `\<close>
  shows   \<open>\<lbrace>P\<rbrace>IN G INVAR invar WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  unfolding while_lfp_scp_invr_prog_def
  by (rule consequence_prog_hoare[OF seq_step _ PHI,
                                  OF while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
                                                                   lens_indep region modifies WF],
                                  OF  consequence_prog_hoare[OF I0' induct_step I0]])

lemma while_scp_invr_vrt_prog_hoare_consequence:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf (R (Rep_uexpr (expression)))\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> (invar \<and> S)`\<close>
  assumes PHI        :\<open>`((\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> S)) \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright> `\<close>
  shows   \<open>\<lbrace>P\<rbrace>IN G INVAR invar VRT expression WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  unfolding while_lfp_scp_invr_vrt_prog_def
  by (rule consequence_prog_hoare[OF seq_step _ PHI,
                                 OF while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
                                                                  lens_indep region modifies WF],
                                 OF  consequence_prog_hoare[OF I0' induct_step I0]])

lemma while_scp_prog_hoare_wp:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep : \<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf R\<close>
  assumes PHI        :\<open>`((\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> Q)) \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> \<rbrace>\<^sub>P\<close>
  shows   \<open>\<lbrace>invar \<and> (\<exists>G \<bullet> Q)\<rbrace>IN G WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  apply (rule post_str_prog_hoare)
  apply (rule while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
                                            lens_indep region modifies WF, of _ _ st'])
   apply (rule consequence_prog_hoare[OF I0' induct_step uimp_refl])
  apply (simp add: PHI exists_twice vwb_lens2)
  done

lemma while_scp_invr_prog_hoare_wp:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf R\<close>
  assumes PHI        :\<open>`((\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> Q)) \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> \<rbrace>\<^sub>P\<close>
  shows \<open>\<lbrace>invar \<and> (\<exists>G \<bullet> Q)\<rbrace>IN G INVAR invar WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close>                              
  unfolding while_lfp_scp_invr_prog_def
  apply (rule post_str_prog_hoare)
   apply (rule while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
                                             lens_indep region modifies WF, of _ _ st'])
   apply (rule consequence_prog_hoare[OF I0' induct_step uimp_refl])
  apply (simp add: PHI exists_twice vwb_lens2)
  done

lemma while_scp_invr_vrt_prog_hoare_wp:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf (R (Rep_uexpr (expression)))\<close>
  assumes PHI        :\<open>`((\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> Q)) \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R (Rep_uexpr (expression))\<guillemotright> \<rbrace>\<^sub>P\<close>
  shows   \<open>\<lbrace>invar \<and> (\<exists>G \<bullet> Q)\<rbrace>IN G INVAR invar VRT expression WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close>                              
  unfolding while_lfp_scp_invr_vrt_prog_def
  apply (rule post_str_prog_hoare)
   apply (rule while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
        lens_indep region modifies WF, of _ _ st'])
   apply (rule consequence_prog_hoare[OF I0' induct_step uimp_refl])
  apply (simp add: PHI exists_twice vwb_lens2)
  done
                                 
lemma while_scp_prog_hoare_sp:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> invar`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`\<close>
  shows   \<open>\<lbrace>P\<rbrace>IN G WHILE b DO body OD\<lbrace>(\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> P)\<rbrace>\<^sub>P\<close> 
  apply (rule pre_weak_prog_hoare)
  defer
   apply (rule while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
                                             lens_indep region modifies WF, of _ _ st'])
   apply (rule consequence_prog_hoare[OF uimp_refl induct_step I0])
  apply (simp add: refBy_order seq_step utp_pred_laws.inf.absorb2)
  done

lemma while_scp_invr_prog_hoare_sp:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>G:[body] = body\<close>
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> invar`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `\<close>
  shows   \<open>\<lbrace>P\<rbrace>IN G INVAR invar WHILE b DO body OD\<lbrace>(\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> P)\<rbrace>\<^sub>P\<close> 
  unfolding while_lfp_scp_invr_prog_def  
  apply (rule pre_weak_prog_hoare)
  defer
   apply (rule while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
                                             lens_indep region modifies WF, of _ _ st'])
   apply (rule consequence_prog_hoare[OF uimp_refl induct_step I0])
  apply (simp add: refBy_order seq_step utp_pred_laws.inf.absorb2)
  done

lemma while_scp_invr_vrt_hoare_prog_sp:
  assumes vwb_lens1  :\<open>vwb_lens g\<close>
  assumes vwb_lens2  :\<open>vwb_lens G\<close>
  assumes vwb_lens3  :\<open>vwb_lens l\<close>
  assumes lens_indep :\<open>g \<bowtie> l\<close>
  assumes region     :\<open>G \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open> MODIFY G DO body OD = body\<close>
  assumes WF         :\<open>wf (R (Rep_uexpr (expression)))\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> invar`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright> `\<close>
  shows   \<open>\<lbrace>P\<rbrace>IN G INVAR invar VRT expression WHILE b DO body OD\<lbrace>(\<exists>l \<bullet> (\<not>b \<and> invar)) \<and> (\<exists>G \<bullet> P)\<rbrace>\<^sub>P\<close> 
  unfolding while_lfp_scp_invr_vrt_prog_def  
  apply (rule pre_weak_prog_hoare)
  defer
   apply (rule while_scp_prog_hoare_basic[OF vwb_lens1 vwb_lens2 vwb_lens3
                                             lens_indep region modifies WF, of _ _ st'])
   apply (rule consequence_prog_hoare[OF uimp_refl induct_step I0])
  apply (simp add: refBy_order seq_step utp_pred_laws.inf.absorb2)
  done

subsection \<open>Hoare for while loop\<close>  

lemma while_hoare_prog_basic:
  assumes WF         :\<open>wf R\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>  
                              body                      
                             \<lbrace>invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P\<close>    
  shows   \<open>\<lbrace>invar\<rbrace>WHILE b DO body OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P\<close>
  unfolding while_lfp_prog_mu_prog
  apply (rule pre_weak_prog_hoare)
   defer
   apply (rule lfp_prog_hoare_prog_wp[OF WF if_prog_mono_prog])
    defer
    apply (rule cond_prog_hoare_wp)
     apply (rule seq_prog_hoare)
      apply (rule induct_step)
     apply assumption
    apply (rule skip_prog_hoare)
   apply simp
  apply pred_simp
  done
    
lemma while_hoare_prog_consequence:
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`p \<Rightarrow> invar`\<close>
  assumes PHI        :\<open>`\<not>b \<and> invar \<Rightarrow> q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `\<close>
  shows   \<open>\<lbrace>p\<rbrace>WHILE b DO body OD\<lbrace>q\<rbrace>\<^sub>P\<close> 
  by (rule  consequence_prog_hoare[OF seq_step _ PHI, 
                                   OF while_hoare_prog_basic[OF WF,of _ _ e],  
                                   OF consequence_prog_hoare[OF I0' induct_step I0]])

lemma while_invr_hoare_prog_consequence:
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> invar`\<close>
  assumes PHI        :\<open>`\<not>b \<and> invar \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`\<close>
  shows   \<open>\<lbrace>P\<rbrace>INVAR invar WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  unfolding while_lfp_invr_prog_def
  by (rule consequence_prog_hoare[OF seq_step _ PHI, 
                                  OF while_hoare_prog_basic[OF WF,of _ _ st'],
                                  OF consequence_prog_hoare[OF I0' induct_step I0]])

lemma while_invr_vrt_hoare_prog_consequence:
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> invar`\<close>
  assumes PHI        :\<open>`\<not>b \<and> invar \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`\<close>
  shows   \<open>\<lbrace>P\<rbrace>INVAR invar VRT \<guillemotleft>R\<guillemotright> WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close> 
  unfolding while_lfp_invr_vrt_prog_def  
  by (rule consequence_prog_hoare[OF seq_step _ PHI, 
                                  OF while_hoare_prog_basic[OF WF,of _ _ st'],
                                  OF consequence_prog_hoare[OF I0' induct_step I0]])

lemma while_hoare_prog_wp:
  assumes WF         :\<open>wf (R (Rep_uexpr (expression)))\<close>
  assumes PHI        :\<open>`\<not>b \<and> invar \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright> \<rbrace>\<^sub>P\<close>
  shows \<open>\<lbrace>invar\<rbrace>WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close> 
 by (rule consequence_prog_hoare[OF uimp_refl _ PHI,
                                 OF while_hoare_prog_basic[OF WF,of _ _ st'],
                                 OF consequence_prog_hoare[OF I0' induct_step uimp_refl]])
   
lemma while_invr_hoare_prog_wp:
  assumes WF         :\<open>wf R\<close>
  assumes PHI        :\<open>`\<not>b \<and> invar \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> \<rbrace>\<^sub>P\<close>
  shows   \<open>\<lbrace>invar\<rbrace>INVAR invar WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close>                              
  unfolding while_lfp_invr_prog_def
  by (rule consequence_prog_hoare[OF uimp_refl _ PHI,
                                  OF while_hoare_prog_basic[OF WF,of _ _ st'],
                                  OF consequence_prog_hoare[OF I0' induct_step uimp_refl]])

lemma while_invr_vrt_hoare_prog_wp:
  assumes WF         :\<open>wf (R (Rep_uexpr (expression)))\<close>
  assumes PHI        :\<open>`\<not>b \<and> invar \<Rightarrow> Q`\<close>  
  assumes I0'        :\<open>\<And>st. `b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> P' st`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>P' st\<rbrace>body\<lbrace>invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright>\<rbrace>\<^sub>P\<close>
  shows   \<open>\<lbrace>invar\<rbrace>INVAR invar VRT expression WHILE b DO body OD\<lbrace>Q\<rbrace>\<^sub>P\<close>                              
  unfolding while_lfp_invr_vrt_prog_def
  by (rule consequence_prog_hoare[OF uimp_refl _ PHI,
                                     OF while_hoare_prog_basic[OF WF,of _ _ st'],
                                     OF consequence_prog_hoare[OF I0' induct_step uimp_refl]])
                   
lemma while_hoare_prog_sp:
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> invar`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `\<close>
  shows   \<open>\<lbrace>P\<rbrace>WHILE b DO body OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P\<close> 
  by (rule consequence_prog_hoare[OF seq_step _ uimp_refl, 
                                  OF while_hoare_prog_basic[OF WF,of _ _ st'],
                                  OF consequence_prog_hoare[OF uimp_refl induct_step I0]])                              
                              
lemma while_invr_hoare_prog_sp:
  assumes WF         :\<open>wf R\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> invar`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and> (st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright> `\<close>
  shows   \<open>\<lbrace>P\<rbrace>INVAR invar WHILE b DO body OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P\<close> 
  unfolding while_lfp_invr_prog_def  
  by (rule consequence_prog_hoare[OF seq_step _ uimp_refl, 
                                  OF while_hoare_prog_basic[OF WF,of _ _ st'],
                                  OF consequence_prog_hoare[OF uimp_refl induct_step I0]])

lemma while_invr_vrt_hoare_prog_sp:
  assumes WF         :\<open>wf (R (Rep_uexpr (expression)))\<close>
  assumes seq_step   :\<open>`P \<Rightarrow> invar`\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>b \<and> invar \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>Q' st\<rbrace>\<^sub>P\<close>
  assumes I0         :\<open>\<And> st. `Q' st \<Rightarrow> invar \<and>(st',\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright>`\<close>
  shows   \<open>\<lbrace>P\<rbrace>INVAR invar VRT expression WHILE b DO body OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P\<close> 
  unfolding while_lfp_invr_vrt_prog_def  
  by (rule consequence_prog_hoare[OF seq_step _ uimp_refl, 
                                  OF while_hoare_prog_basic[OF WF,of _ _ st'],
                                  OF consequence_prog_hoare[OF uimp_refl induct_step I0]])

text \<open>To void a blow up in our Hoare Logic we do not develop reasoning laws
      for the different loop patterns such as: for, do_while, from_until.
      Instead we use the algebraic laws to convert any loop pattern
      to a pattern based on while iterations.\<close>

subsection \<open>Hoare for from_until_loop\<close>   
(*lemma from_until_hoare_prog_basic:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>invar\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>invar\<rbrace>FROM init UNTIL exit DO body OD\<lbrace>exit \<and> invar\<rbrace>\<^sub>P"
  unfolding from_until_lfp_prog_def_alt while_lfp_prog_mu_prog
  by (rule seq_prog_hoare[OF seq_step, 
                          OF lfp_prog_hoare_prog[OF WF if_prog_mono_prog cond_prog_hoare_basic, of _ e],
                          OF seq_prog_hoare[OF induct_step], OF _ skip_prog_hoare_intro],                             
      pauto)

lemma from_until_hoare_prog_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>p\<rbrace>FROM init UNTIL exit DO body OD\<lbrace>exit \<and> invar\<rbrace>\<^sub>P"
  unfolding from_until_lfp_prog_def_alt while_lfp_prog_mu_prog
  by (rule seq_prog_hoare[OF seq_step, 
                          OF lfp_prog_hoare_prog[OF WF if_prog_mono_prog cond_prog_hoare_basic, of _ e],
                          OF seq_prog_hoare[OF induct_step], OF _ skip_prog_hoare_intro],                             
      pauto)

lemma from_until_invr_hoare_prog_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>p\<rbrace>FROM init INVAR invar UNTIL exit DO body OD\<lbrace>exit \<and> invar\<rbrace>\<^sub>P"
  unfolding from_until_lfp_invr_prog_def 
  using     from_until_hoare_prog_minimal [OF WF seq_step induct_step] .
    
lemma from_until_invr_vrt_hoare_prog_minimal:
  assumes WF: "wf (R (Rep_uexpr (expression)))"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>p\<rbrace>FROM init INVAR invar VRT expression UNTIL exit DO body OD\<lbrace>exit \<and> invar\<rbrace>\<^sub>P"
  unfolding from_until_lfp_invr_vrt_prog_def 
  using     from_until_hoare_prog_minimal [OF WF seq_step induct_step] .

lemma from_until_hoare_prog_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>FROM init UNTIL exit DO body OD\<lbrace>q\<rbrace>\<^sub>P" 
  unfolding from_until_lfp_prog_def_alt
  by (simp add: uintro PHI seq_prog_hoare[OF seq_step] 
                       while_hoare_prog_consequence[OF WF uimp_refl _ I0' induct_step I0])

lemma from_until_invr_hoare_prog_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>FROM init INVAR invar UNTIL exit DO body OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding from_until_lfp_invr_prog_def 
  using     from_until_hoare_prog_consequence [OF WF seq_step PHI I0' induct_step I0].
    
lemma from_until_invr_vrt_hoare_prog_consequence:
  assumes WF: "wf (R (Rep_uexpr (expression)))"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>FROM init INVAR invar VRT expression UNTIL exit DO body OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding from_until_lfp_invr_vrt_prog_def
  using     from_until_hoare_prog_consequence [OF WF seq_step PHI I0' induct_step I0].
 
lemma from_until_hoare_prog_wp:
  assumes WF: "wf R" 
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P" 
  shows "\<lbrace>p\<rbrace>FROM init UNTIL exit DO body OD\<lbrace>q\<rbrace>\<^sub>P" 
  unfolding from_until_lfp_prog_def_alt
  by (simp add: uintro PHI seq_prog_hoare[OF seq_step] pre_str_prog_hoare[OF uimp_refl]
                       while_hoare_prog_consequence[OF WF uimp_refl _ I0' induct_step uimp_refl])

lemma from_until_invr_hoare_prog_wp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P" 
  shows "\<lbrace>p\<rbrace>FROM init INVAR invar UNTIL exit DO body OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding from_until_lfp_invr_prog_def
  using from_until_hoare_prog_wp [OF WF seq_step PHI I0' induct_step] .
    
lemma from_until_invr_vrt_hoare_prog_wp:
  assumes WF: "wf (R (Rep_uexpr (expression)))"   
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`exit \<and> invar \<Rightarrow> q`"  
  assumes I0': "\<And>st. `\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright>\<rbrace>\<^sub>P" 
  shows "\<lbrace>p\<rbrace>FROM init INVAR invar VRT expression UNTIL exit DO body OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding from_until_lfp_invr_vrt_prog_def
  using from_until_hoare_prog_wp [OF WF  seq_step PHI I0' induct_step] .
    
lemma from_until_hoare_prog_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>P"
  assumes I0': "`q'' \<Rightarrow> invar`"    
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>FROM init UNTIL exit DO body OD\<lbrace>exit \<and> invar\<rbrace>\<^sub>P" 
  unfolding from_until_lfp_prog_def_alt
  by (simp add: uintro uimp_refl seq_prog_hoare[OF seq_step] pre_str_prog_hoare[OF I0']
                       while_hoare_prog_consequence[OF WF uimp_refl _ uimp_refl induct_step I0])

lemma from_until_invr_hoare_prog_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>P"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>FROM init INVAR invar UNTIL exit DO body OD\<lbrace>exit \<and> invar\<rbrace>\<^sub>P" 
  unfolding from_until_lfp_invr_prog_def
  using from_until_hoare_prog_sp[OF WF seq_step I0' induct_step I0] .  

lemma from_until_invr_vrt_hoare_prog_sp:
  assumes WF: "wf (R (Rep_uexpr (expression)))"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>P"
   assumes I0': "`q'' \<Rightarrow> invar`"    
  assumes induct_step: "\<And>st. \<lbrace>\<not> exit \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P" 
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>(R (Rep_uexpr (expression)))\<guillemotright>`"     
  shows "\<lbrace>p\<rbrace>FROM init INVAR invar VRT expression UNTIL exit DO body OD\<lbrace>exit \<and> invar\<rbrace>\<^sub>P" 
  unfolding from_until_lfp_invr_vrt_prog_def
  using from_until_hoare_prog_sp[OF WF seq_step I0' induct_step I0] .*)

subsection \<open>Hoare for do_while_loop\<close>     

(*lemma do_while_hoare_prog_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and>  e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows" \<lbrace>p\<rbrace>DO body WHILE b OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"  
  unfolding do_while_lfp_prog_def_alt
  by (rule seq_prog_hoare[OF seq_step while_hoare_prog_basic[OF WF induct_step]])  

lemma do_while_invr_hoare_prog_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and>  e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows" \<lbrace>p\<rbrace>DO body INVAR invar WHILE b OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"  
  unfolding do_while_lfp_invr_prog_def
  by (rule do_while_hoare_prog_minimal[OF WF seq_step induct_step])  

lemma do_while_invr_vrt_hoare_prog_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and>  e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows" \<lbrace>p\<rbrace>DO body INVAR invar VRT \<guillemotleft>R\<guillemotright> WHILE b OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"  
  unfolding do_while_lfp_invr_vrt_prog_def
  by (rule do_while_hoare_prog_minimal[OF WF seq_step induct_step])  
    
lemma do_while_hoare_prog_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>DO body WHILE b OD\<lbrace>q\<rbrace>\<^sub>P" 
  unfolding do_while_lfp_prog_def_alt
  by (rule seq_prog_hoare[OF seq_step 
                             while_hoare_prog_consequence[OF WF uimp_refl PHI I0' induct_step I0]])
    
lemma do_while_invr_hoare_prog_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>DO body INVAR invar WHILE b OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding do_while_lfp_invr_prog_def
  by (rule do_while_hoare_prog_consequence[OF WF seq_step PHI I0' induct_step I0])  

lemma do_while_invr_vrt_hoare_prog_consequence:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>DO body INVAR invar VRT \<guillemotleft>R\<guillemotright> WHILE b OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding do_while_lfp_invr_vrt_prog_def
  by (rule do_while_hoare_prog_consequence[OF WF seq_step PHI I0' induct_step I0])  
    
lemma do_while_hoare_prog_wp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>invar\<rbrace>DO body WHILE b OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding do_while_lfp_prog_def_alt 
  by (simp add: uimp_refl seq_prog_hoare[OF seq_step] 
                   while_hoare_prog_wp[OF WF PHI I0' induct_step])

lemma do_while_invr_hoare_prog_wp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>invar\<rbrace>DO body INVAR invar WHILE b OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding do_while_lfp_invr_prog_def do_while_lfp_prog_def_alt
  by (simp add: uimp_refl seq_prog_hoare[OF seq_step]
                while_hoare_prog_wp[OF WF PHI I0' induct_step])

lemma do_while_invr_vrt_hoare_prog_wp:
  assumes WF: "wf R"  
  assumes seq_step: "\<lbrace>invar\<rbrace>body\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI : "`\<not>b  \<and> invar \<Rightarrow> q`"
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And> st. \<lbrace>p' st\<rbrace>body\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>invar\<rbrace>DO body INVAR invar VRT \<guillemotleft>R\<guillemotright> WHILE b OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding do_while_lfp_invr_vrt_prog_def do_while_lfp_invr_prog_def do_while_lfp_prog_def_alt
  by (simp add: seq_prog_hoare[OF seq_step] 
                while_hoare_prog_wp[OF WF PHI I0' induct_step])

lemma do_while_hoare_prog_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>q''\<rbrace>\<^sub>P"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>DO body WHILE b OD\<lbrace>\<not>b  \<and> invar\<rbrace>\<^sub>P"
  unfolding  do_while_lfp_prog_def_alt
  by (simp add: uintro uimp_refl seq_prog_hoare[OF seq_step] pre_str_prog_hoare[OF I0']
                   while_hoare_prog_sp[OF WF uimp_refl induct_step I0])
                 
lemma do_while_invr_hoare_prog_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>q''\<rbrace>\<^sub>P"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>DO body INVAR invar WHILE b OD\<lbrace>\<not>b  \<and> invar\<rbrace>\<^sub>P"
  unfolding do_while_lfp_invr_prog_def do_while_lfp_prog_def_alt
  by (simp add: uintro uimp_refl seq_prog_hoare[OF seq_step] pre_str_prog_hoare[OF I0']
                   while_hoare_prog_sp[OF WF uimp_refl induct_step I0])
                 
lemma do_while_invr_vrt_hoare_prog_sp:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>body\<lbrace>q''\<rbrace>\<^sub>P"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And> st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>DO body INVAR invar VRT \<guillemotleft>R\<guillemotright> WHILE b OD\<lbrace>\<not>b  \<and> invar\<rbrace>\<^sub>P"
  unfolding do_while_lfp_invr_vrt_prog_def do_while_lfp_prog_def_alt
  by (simp add: uintro uimp_refl seq_prog_hoare[OF seq_step] pre_str_prog_hoare[OF I0']
                   while_hoare_prog_sp[OF WF uimp_refl induct_step I0])*)
                 
subsection \<open>Hoare for for_loop\<close>     
    
(*lemma for_hoare_prog_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body ; incr\<lbrace>invar \<and> (e,\<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P" 
  shows "\<lbrace>p\<rbrace>FOR (init, b, incr) DO body OD \<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"
  unfolding for_lfp_prog_def
  by (simp add: from_until_hoare_prog_minimal [OF WF seq_step, of _ e] induct_step)  

lemma for_invr_hoare_prog_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body ; incr\<lbrace>invar \<and> (e,\<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P" 
  shows "\<lbrace>p\<rbrace>FOR (init, b, incr) INVAR invar DO body OD \<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"
  unfolding for_lfp_invr_prog_def for_lfp_prog_def
  by (simp add: from_until_hoare_prog_minimal [OF WF seq_step, of _ e] induct_step)  

lemma for_invr_vrt_hoare_prog_minimal:
  assumes WF: "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body ; incr\<lbrace>invar \<and> (e,\<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P" 
  shows "\<lbrace>p\<rbrace>FOR (init, b, incr) INVAR invar VRT \<guillemotleft>R\<guillemotright> DO body OD \<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"
  unfolding for_lfp_invr_vrt_prog_def for_lfp_prog_def
  by (simp add: from_until_hoare_prog_minimal [OF WF seq_step, of _ e] induct_step)      

lemma for_hoare_prog_consequence:
  assumes WF: "wf R"
  assumes seq_step:  "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;incr\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"
  shows "\<lbrace>p\<rbrace>FOR (init, b, incr) DO body OD\<lbrace>q\<rbrace>\<^sub>P"  
  unfolding for_lfp_prog_def
  by(simp add: from_until_hoare_prog_consequence [OF WF seq_step PHI _ induct_step I0] I0')

lemma for_invr_hoare_prog_consequence:
  assumes WF: "wf R"
  assumes seq_step:  "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;incr\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"
  shows "\<lbrace>p\<rbrace>FOR (init, b, incr) INVAR invar DO body OD\<lbrace>q\<rbrace>\<^sub>P"  
  unfolding for_lfp_prog_def for_lfp_invr_prog_def
  by(simp add: from_until_hoare_prog_consequence [OF WF seq_step PHI _ induct_step I0] I0')

lemma for_invr_vrt_hoare_prog_consequence:
  assumes WF: "wf R"
  assumes seq_step:  "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"  
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;incr\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"
  shows "\<lbrace>p\<rbrace>FOR (init, b, incr) INVAR invar VRT \<guillemotleft>R\<guillemotright> DO body OD\<lbrace>q\<rbrace>\<^sub>P"  
  unfolding for_lfp_prog_def for_lfp_invr_vrt_prog_def
  by(simp add: from_until_hoare_prog_consequence [OF WF seq_step PHI _ induct_step I0] I0')
    
lemma for_hoare_prog_wp:
  assumes WF : "wf R" 
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e=\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;incr\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>p\<rbrace>FOR (init,b,incr) DO body OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding for_lfp_prog_def_alt
  by (simp add: uimp_refl seq_prog_hoare[OF seq_step]
                while_hoare_prog_wp[OF WF PHI I0' induct_step])

lemma for_invr_hoare_prog_wp:
  assumes WF : "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e=\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;incr\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>p\<rbrace>FOR (init,b,incr) INVAR invar DO body OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding for_lfp_invr_prog_def for_lfp_prog_def_alt
  by (simp add: uimp_refl seq_prog_hoare[OF seq_step] 
                while_hoare_prog_wp[OF WF PHI I0' induct_step])

lemma for_invr_vrt_hoare_prog_wp:
  assumes WF : "wf R"
  assumes seq_step: "\<lbrace>p\<rbrace>init\<lbrace>invar\<rbrace>\<^sub>P"
  assumes PHI: "`\<not>b \<and> invar \<Rightarrow>q`"  
  assumes I0': "\<And>st. `b \<and> invar \<and> e=\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' st`"
  assumes induct_step: "\<And>st. \<lbrace>p' st\<rbrace>body;incr\<lbrace>invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<lbrace>p\<rbrace>FOR (init,b,incr) INVAR invar VRT \<guillemotleft>R\<guillemotright>DO body OD\<lbrace>q\<rbrace>\<^sub>P"
  unfolding for_lfp_invr_vrt_prog_def for_lfp_prog_def_alt
  by (simp add: seq_prog_hoare[OF seq_step]
                while_hoare_prog_wp[OF WF  PHI I0' induct_step])

lemma for_hoare_prog_sp:
  assumes WF: "wf R"
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>P"
  assumes I0': "`q'' \<Rightarrow> invar`"  
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body;incr\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>FOR (init,b,incr) DO body OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"
  unfolding for_lfp_prog_def_alt  
  by (simp add: uintro uimp_refl seq_prog_hoare[OF seq_step] pre_str_prog_hoare[OF I0']
                while_hoare_prog_sp[OF WF uimp_refl induct_step I0])
 
lemma for_invr_hoare_prog_sp:
  assumes WF: "wf R"
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>P"
  assumes I0':  "`q'' \<Rightarrow> invar`"
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body;incr\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>FOR (init,b,incr) INVAR invar DO body OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"
  unfolding for_lfp_invr_prog_def for_lfp_prog_def_alt  
  by (simp add: uintro uimp_refl seq_prog_hoare[OF seq_step] pre_str_prog_hoare[OF I0']
                while_hoare_prog_sp[OF WF uimp_refl induct_step I0])
     
    
lemma for_invr_vrt_hoare_prog_sp:
  assumes WF: "wf R"
  assumes seq_step:"\<lbrace>p\<rbrace>init\<lbrace>q''\<rbrace>\<^sub>P"
  assumes I0':"`q''\<Rightarrow> invar`" 
  assumes induct_step: "\<And>st. \<lbrace>b \<and> invar \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>body;incr\<lbrace>q' st\<rbrace>\<^sub>P"
  assumes I0: "\<And>st. `q' st \<Rightarrow> invar \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>`"  
  shows "\<lbrace>p\<rbrace>FOR (init,b,incr) INVAR invar VRT \<guillemotleft>R\<guillemotright>DO body OD\<lbrace>\<not>b \<and> invar\<rbrace>\<^sub>P"
  unfolding for_lfp_invr_vrt_prog_def for_lfp_prog_def_alt  
  by (simp add: uintro uimp_refl seq_prog_hoare[OF seq_step] pre_str_prog_hoare[OF I0']
                while_hoare_prog_sp[OF WF uimp_refl induct_step I0])*)


section \<open>Hoare for modular reasoning\<close> 
subsection \<open>assertion logic\<close>
text\<open>For technical reasons we eta-expand our assertion logic. It will be useful for modular reasoning
     where the eta expanded part will contain Isabelle variables. Isabelle variables
     are used as logical variables in Hoare specifications.The eta-expansion of the assertion
     logic allows for binding
     Isabelle variables which avoid unification problems when reusing already proved specifications.
     Namely, without biding Isabelle variables  the specification will contain Isabelle 
     free variables. Reusing Isabelle Free variables imposes manual instantiations of the latter
     each time the proved specification is used to prove another specification. Therefore, 
     we will not get automation. In Isabelle, free variables in theorems are converted to schematic
     variables, and schematic variables can not be unified automatically. Hence, we use eta-expansion
     to have automation.
     \<close>
type_synonym ('\<alpha>,'d) assn    = \<open>'d \<Rightarrow> '\<alpha> cond\<close>
type_synonym ('t,'\<alpha>,'d) vrnt = \<open>'d \<Rightarrow> ('t,'\<alpha>) uexpr\<close>
(*definition hoare_prog_modify :: 
  "('a \<Rightarrow> 'b upred) \<Rightarrow> 'd prog \<Rightarrow>  
   ('a \<Rightarrow> 'b upred) \<Rightarrow> ('b \<Longrightarrow> 'd) \<Rightarrow> ('c \<Longrightarrow> 'd) \<Rightarrow> bool"   ("\<lbrace>(1_)\<rbrace>/(_)/\<lbrace>(1_)\<rbrace> modf _ not _" [0,0,0,500,500] 500)
where
  "hoare_prog_modify p  PROG q  G L \<longleftrightarrow> (
   (vwb_lens G \<and> vwb_lens L \<and> G \<bowtie> L \<and>(\<forall>a. \<lbrace>(p a)\<oplus>\<^sub>pG \<rbrace>PROG\<lbrace>(q a)\<oplus>\<^sub>pG\<rbrace>\<^sub>P)))"*)

abbreviation hoare_prog':: 
  \<open>('a \<Rightarrow> 'b upred) \<Rightarrow> ('a \<Rightarrow>'b prog) \<Rightarrow>  ('a \<Rightarrow> 'b upred) \<Rightarrow> bool\<close> ("\<lbrace>(1_)\<rbrace>/(_)/\<lbrace>(1_)\<rbrace>\<^sub>G" [0,0,0] 500)
  where \<open>hoare_prog' p PROG q \<equiv> (\<forall>a. \<lbrace>(p a)\<rbrace>PROG a\<lbrace>(q a)\<rbrace>\<^sub>P)\<close>

abbreviation hoare_prog_modularity:: 
  \<open>('b \<Rightarrow> 'b) \<Rightarrow> 'b upred \<Rightarrow> 'b prog \<Rightarrow>  'b upred \<Rightarrow> bool\<close> ("\<langle>_\<rangle>\<lbrace>(1_)\<rbrace>/(_)/\<lbrace>(1_)\<rbrace>\<^sub>G" [0,0,0] 500)
  where \<open>hoare_prog_modularity \<sigma> p PROG q \<equiv> (\<lbrace>p \<rbrace> \<langle>\<sigma>\<rangle>\<^sub>p; PROG\<lbrace>q \<rbrace>\<^sub>P)\<close>

(*
syntax
  --\<open> Syntactic sugar of Hoare triples for modular reasoning with structured state space\<close>
  "_hoare_prog_mod":: "logic \<Rightarrow> logic \<Rightarrow> logic\<Rightarrow> logic"("\<lbrace>(1_)\<rbrace>/(_)/\<lbrace>(1_)\<rbrace>\<^sub>G"[0,0,0] 500)
  "_Gvar":: "salpha\<Rightarrow>logic"("Gvar _ "[500] 0)
  "_Lvar":: "salpha\<Rightarrow>salpha"("Lvar _ "[500] 0)

translations
  "\<lbrace>(\<lambda>_. p)\<rbrace>PROG\<lbrace>(\<lambda>_. q)\<rbrace>\<^sub>G" => "CONST All (\<lambda>_. (CONST hoare_prog_t ((CONST aext) ((\<lambda>_. p) _) (CONST fst_lens)) PROG ((CONST aext) ((\<lambda>_. q) _) (CONST fst_lens))))"
  "_Gvar gvar" \<rightleftharpoons> "(CONST lens_comp gvar (CONST fst_lens))"
  "_Lvar lvar" \<rightleftharpoons> "(CONST lens_comp lvar (CONST snd_lens))"
term "\<lbrace>p\<rbrace>PROG\<lbrace>q\<rbrace>\<^sub>G"
term "CONST All (\<lambda>s. (CONST hoare_prog_t ((CONST aext) ((\<lambda>_. p) _) (CONST fst_lens)) PROG ((CONST aext) ((\<lambda>_. q) _) (CONST fst_lens))))"
term "(\<forall>a. \<lbrace>((\<lambda>s. p) a)\<oplus>\<^sub>pfst\<^sub>L\<rbrace>PROG\<lbrace>((\<lambda>t. q) a)\<oplus>\<^sub>pfst\<^sub>L\<rbrace>\<^sub>P)"

lemma hoare_prog_mod_syntax_test:
 "\<lbrace>p\<rbrace>PROG\<lbrace>q\<rbrace> modf (fst\<^sub>L) not snd\<^sub>L = \<lbrace>p\<rbrace>PROG\<lbrace>q\<rbrace>\<^sub>G"
  by simp*)

lemma hoare_prog_mod_hoare_prog_eq:
  \<open>\<lbrace>(\<lambda>s. p) \<rbrace>(\<lambda>s. PROG)\<lbrace>(\<lambda>t. q)\<rbrace>\<^sub>G = (\<forall>a. \<lbrace>((\<lambda>s. p) a)\<rbrace>(\<lambda>s. PROG) a\<lbrace>((\<lambda>t. q) a)\<rbrace>\<^sub>P)\<close>
  by auto

definition from_until_lfp_invr_vrt_prog_G :: 
  \<open>'\<alpha> prog \<Rightarrow>  ('t,'\<alpha>) uexpr \<Rightarrow> ('\<alpha>,'d) assn \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog\<close> ("FROM (_)/ VRT (_)/ INVAR (_)/  UNTIL (_)/ DO (_) OD")
  where \<open>FROM init VRT vari INVAR invar UNTIL exit DO body OD = FROM init UNTIL exit DO body OD\<close>
declare from_until_lfp_invr_vrt_prog_G_def [prog_rep_eq]

definition while_lfp_invr_vrt_prog_G :: 
  \<open>('t,'\<alpha>) uexpr \<Rightarrow> ('\<alpha>,'d) assn \<Rightarrow>  '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog\<close> ("VRT (_)/ INVAR (_)/  WHILE (_)/ DO (_) OD")
  where \<open>while_lfp_invr_vrt_prog_G vari invar b body = WHILE b DO body OD\<close>
declare while_lfp_invr_vrt_prog_G_def [prog_rep_eq]

definition for_lfp_invr_vrt_prog_G :: 
  \<open>'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow>  ('t,'\<alpha>) uexpr \<Rightarrow> ('\<alpha>,'d) assn \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog\<close> 
  ("FOR ('(_,_,_'))/ VRT (_)/ INVAR (_)/  DO (_) OD")
where \<open>FOR (init, b, incr) VRT vari INVAR invar DO body OD = FOR (init, b, incr) DO body OD\<close>  
declare for_lfp_invr_vrt_prog_G_def [prog_rep_eq]

definition do_while_lfp_invr_vrt_prog_G :: 
  \<open>'\<alpha> prog \<Rightarrow>  ('t,'\<alpha>) uexpr \<Rightarrow> ('\<alpha>,'d) assn \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog\<close> ("DO (_)/ VRT (_)/ INVAR (_)/ WHILE (_) OD")
  where \<open>DO body  VRT vari INVAR invar WHILE b OD = DO body WHILE b OD\<close> 
declare do_while_lfp_invr_vrt_prog_G_def [prog_rep_eq]

lemma while_hoare_prog_G_basic:
  assumes WF         :\<open>wf R\<close>
  assumes induct_step:\<open>\<And>st. \<lbrace>\<lambda>z. b \<and> invar z \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright>\<rbrace>(\<lambda>z. body)\<lbrace>\<lambda>z. invar z \<and> (st', \<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>G\<close>    
  shows   \<open>\<lbrace>invar\<rbrace>(\<lambda>z. WHILE b DO body OD)\<lbrace>\<lambda>z. \<not>b \<and> invar z\<rbrace>\<^sub>G\<close>
  apply (rule allI)
  apply (rule while_hoare_prog_basic)
    apply (rule WF) 
  using induct_step
      apply (auto simp  add: prog_rep_eq)
  done

lemma while_vrt_invr_hoare_prog_wp:
  assumes WF         :\<open>wf R\<close>
  assumes PHI        :\<open>\<And> z.` \<not>b \<and> (\<lambda>x. invar x) z \<Rightarrow> (\<lambda>x. q x) z`\<close>  
  assumes I0'        :\<open>\<And>st z. `b \<and> (\<lambda>x. invar x) z \<and> st' =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' z st`\<close>
  assumes induct_step:\<open>\<And>st z. \<lbrace>(p' z) st\<rbrace>body\<lbrace> (\<lambda>x. invar x) z \<and>(st', \<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P\<close>
  shows   \<open>\<lbrace>(\<lambda>x. invar x) z\<rbrace>VRT \<guillemotleft>R\<guillemotright> INVAR invar WHILE b DO body OD\<lbrace> q z\<rbrace>\<^sub>P\<close>  
  unfolding while_lfp_invr_vrt_prog_G_def
  apply (rule while_hoare_prog_wp)
     apply (rule WF)
    apply (rule PHI)
   apply (rule I0')
  apply (rule induct_step)
  done

(*lemma for_vrt_invr_hoare_prog_wp:
  assumes WF : "wf R"
  assumes seq_step: "\<And>z. \<lbrace>p z\<rbrace>init\<lbrace>invar z\<rbrace>\<^sub>P"
  assumes PHI: "\<And>z.` \<not>b \<and> invar z \<Rightarrow> q z`"  
  assumes I0': "\<And>st z . `b \<and> invar z \<and> e=\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' z st`"
  assumes induct_step: "\<And>st z . \<lbrace>(p' z) st\<rbrace>body;incr\<lbrace>invar z \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<And>z. \<lbrace>p z\<rbrace>FOR (init,b,incr) VRT \<guillemotleft>R\<guillemotright> INVAR invar DO body OD\<lbrace>q z\<rbrace>\<^sub>P"
  unfolding for_lfp_invr_vrt_prog_G_def
  apply (rule for_hoare_prog_wp)
      apply (rule WF)
     apply (rule seq_step)
    apply (rule PHI)
   apply (rule I0')
  apply (rule induct_step)
  done

lemma do_while_vrt_invr_hoare_prog_wp:
  assumes WF: "wf R"  
  assumes seq_step: "\<And>z. \<lbrace>invar z\<rbrace>body\<lbrace>invar z\<rbrace>\<^sub>P"
  assumes PHI : "\<And>z. `\<not>b  \<and> invar z \<Rightarrow> q z`"
  assumes I0': "\<And>st z. `b \<and> invar z \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' z st`"  
  assumes induct_step: "\<And>st z. \<lbrace>(p' z) st\<rbrace>body\<lbrace>invar z \<and> (e, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P"
  shows "\<And>z. \<lbrace>invar z\<rbrace>DO body VRT \<guillemotleft>R\<guillemotright> INVAR invar WHILE b OD\<lbrace>q z\<rbrace>\<^sub>P"
 unfolding do_while_lfp_invr_vrt_prog_G_def
  apply (rule do_while_hoare_prog_wp)
      apply (rule WF)
     apply (rule seq_step)
    apply (rule PHI)
   apply (rule I0')
  apply (rule induct_step)
  done

lemma from_until_vrt_invr_hoare_prog_wp:
  assumes WF: "wf R"   
  assumes seq_step: "\<And>z. \<lbrace>p  z\<rbrace>init\<lbrace>invar z\<rbrace>\<^sub>P"
  assumes PHI: "\<And>z. `exit \<and> invar z \<Rightarrow> q z`"  
  assumes I0': "\<And>st z. `\<not> exit \<and> invar z \<and> e =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow> p' z st`"  
  assumes induct_step: "\<And>st z. \<lbrace>(p' z) st\<rbrace>body\<lbrace>invar z \<and>(e,\<guillemotleft>st\<guillemotright>)\<^sub>u\<in>\<^sub>u\<guillemotleft>R\<guillemotright>\<rbrace>\<^sub>P" 
  shows "\<And>z. \<lbrace>p z\<rbrace>FROM init VRT \<guillemotleft>R\<guillemotright> INVAR (\<lambda>_. invar z)  UNTIL exit DO body OD\<lbrace>q z\<rbrace>\<^sub>P"
  unfolding from_until_lfp_invr_vrt_prog_G_def
  apply (rule from_until_hoare_prog_wp)
      apply (rule WF)
     apply (rule seq_step)
    apply (rule PHI)
   apply (rule I0')
  apply (rule induct_step)
  done*)

lemmas loop_invr_vrt_hoare_prog_sp_instantiated [hoare_sp_rules] = 
       while_invr_vrt_hoare_prog_sp [where st' = "&\<Sigma>"]
       (*for_invr_vrt_hoare_prog_sp [where e = "&\<Sigma>"]
       do_while_invr_vrt_hoare_prog_sp [where e = "&\<Sigma>"]
       from_until_invr_vrt_hoare_prog_sp[where e = "&\<Sigma>"]*)

lemmas loop_invr_vrt_hoare_prog_wp_instantiated [hoare_wp_rules] = 
       while_invr_vrt_hoare_prog_wp [where st' = "&\<Sigma>"]
       while_vrt_invr_hoare_prog_wp [where st' = "&\<Sigma>"]
       antiframe_prog_hoare_wp_weak[where g=fst_lens and l=snd_lens] 

       (*for_invr_vrt_hoare_prog_wp [where e = "&\<Sigma>"]
       for_vrt_invr_hoare_prog_wp [where e = "&\<Sigma>"]
       do_while_invr_vrt_hoare_prog_wp [where e = "&\<Sigma>"]
       do_while_vrt_invr_hoare_prog_wp [where e = "&\<Sigma>"]
       from_until_invr_vrt_hoare_prog_wp[where e = "&\<Sigma>"]
       from_until_vrt_invr_hoare_prog_wp[where e = "&\<Sigma>"]*)

end

