(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)

chapter \<open>Abrupt termination and exceptions\<close>
  
theory utp_prog_abrupt_lift
  imports 
    "../theories/Abrupt/utp_prog_abrupt"
begin
  
  
section \<open>Program Type\<close>

typedef ('s, 'e) aprog = "{P :: ('s, 'e) abr_prog. abrh P = P}"
  by (rule exI[where x= "abrh MAGIC"], simp)
    
named_theorems aprog_rep_eq
    
notation Rep_aprog ("\<lbrakk>_\<rbrakk>\<^sub>a")

lemma Rep_prog_abrh_closed: "abrh \<lbrakk>P\<rbrakk>\<^sub>a = \<lbrakk>P\<rbrakk>\<^sub>a  "
  using Rep_aprog by auto
    
setup_lifting type_definition_aprog

section \<open>The Order\<close>
    
instantiation aprog :: (type,type) refine
begin
  lift_definition less_eq_aprog :: "('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog \<Rightarrow> bool" is
  "op \<le>" .
  lift_definition less_aprog :: "('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog \<Rightarrow> bool" is
  "op <" .
instance 
  apply (intro_classes)
     apply (simp add: less_le_not_le utp_prog_abrupt_lift.less_aprog.rep_eq utp_prog_abrupt_lift.less_eq_aprog.rep_eq)
    apply (simp add: utp_prog_abrupt_lift.less_eq_aprog.rep_eq)
   apply (meson dual_order.trans utp_prog_abrupt_lift.less_eq_aprog.rep_eq)
  apply (simp add: Rep_aprog_inject utp_prog_abrupt_lift.less_eq_aprog.rep_eq)
  done
end
  
lemma Rep_prog_refine [aprog_rep_eq]:
  "P \<sqsubseteq> Q \<longleftrightarrow> \<lbrakk>P\<rbrakk>\<^sub>a \<sqsubseteq> \<lbrakk>Q\<rbrakk>\<^sub>a"
  by (simp add: less_eq_aprog.rep_eq)

lemma Rep_prog_eq [aprog_rep_eq]:
  "P = Q \<longleftrightarrow> \<lbrakk>P\<rbrakk>\<^sub>a = \<lbrakk>Q\<rbrakk>\<^sub>a"
  by (metis Rep_aprog_inverse)

section \<open>Transfer methods\<close>

method atransfer = (simp add: aprog_rep_eq)
  
subsection \<open>Proof Tactics\<close>
  
method a_auto   = (atransfer, ptransfer, rel_auto)
method prefine = (atransfer, ptransfer, ndes_refine)
method peq     = (atransfer, ptransfer, ndes_eq)
section \<open>Programming constructs \<close>
        
subsection \<open>Abort\<close>  

no_notation abort ("ABORT")   
  
lift_definition abort_a    :: "('s, 'e) aprog" ("ABORT") is "abrh utp_prog.abort" by simp

subsection \<open>Magic\<close>  
    
no_notation magic ("MAGIC")       
  
lift_definition magic_a    :: "('s, 'e) aprog" ("MAGIC") is "abrh utp_prog.magic" by simp

subsection \<open>Skip\<close>  
    
no_notation skip ("SKIP") 
  
lift_definition skip_a :: "('s, 'e) aprog" ("SKIP")is "utp_prog.skip" by simp

subsection \<open>Throw\<close>  
no_notation throw_abr ("THROW '(_')")     
lift_definition throw_a :: "'e \<Rightarrow> ('s, 'e) aprog" ("THROW '(_')")is "throw_abr" unfolding throw_abr_def by simp  

subsection \<open>Sequence\<close>  
    
purge_notation pseq (infixr ";" 71)   
  
lift_definition seq_a    :: "('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" (infixr ";" 71) is 
                "pseq" by (simp add: closure)

subsection \<open>Assignment\<close>  
     
no_notation assigns_abr ("\<langle>_\<rangle>\<^sub>a")   
no_notation passigns ("\<langle>_\<rangle>\<^sub>a")   
no_notation assigns_r ("\<langle>_\<rangle>\<^sub>a")     
no_notation uassigns("\<langle>_\<rangle>\<^sub>a") 
  
lift_definition assigns_a :: "'s usubst \<Rightarrow> ('s, 'e) aprog" ("\<langle>_\<rangle>\<^sub>a") is "assigns_abr" unfolding assigns_abr_def by simp
subsubsection \<open>Syntax Translations\<close>
  
adhoc_overloading
  uassigns assigns_a

translations
  "_assignment xs vs" => "CONST assigns_a (_mk_usubst (CONST id) xs vs)"
  "x :== v" <= "CONST assigns_a (CONST subst_upd (CONST id) (CONST svar x) v)"

subsection \<open>Substitution\<close> 
  
lift_definition subst_a :: "'s usubst \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" is 
  "\<lambda> \<sigma> P. abrh (psubst (\<sigma> \<oplus>\<^sub>s \<Sigma>\<^sub>A\<^sub>B\<^sub>R) ( P))" 
  by simp
  
adhoc_overloading
   usubst subst_a      
subsection \<open>Conditional\<close>  
    
no_notation pcond_prog ("IF (_)/ THEN (_) ELSE (_) FI") 
  
lift_definition cond_a :: "'s cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" ("IF (_)/ THEN (_) ELSE (_) FI") 
  is "if_abr" 
  unfolding if_abr_def 
  by (simp add: prog_rep_eq abrh_def) (metis Algebraic_Laws.cond_distr)
    
subsection \<open>assert and assume\<close>

no_notation passume_prog ("_\<^sup>\<top>\<^sup>P")
  
abbreviation assume_a :: "'s upred \<Rightarrow> ('s, 'e) aprog" ("_\<^sup>\<top>\<^sup>P")  
  where "assume_a c \<equiv> (IF c THEN SKIP ELSE MAGIC FI)"

no_notation passert_prog ("_\<^sub>\<bottom>\<^sub>P")  
  
abbreviation assert_a :: "'s upred \<Rightarrow> ('s, 'e) aprog" ("_\<^sub>\<bottom>\<^sub>P")
  where "assert_a c \<equiv> (IF c THEN SKIP ELSE ABORT FI)"
    
subsection \<open>try catch\<close>  
    
lift_definition try_catch_a :: "('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" ("TRY (_)/ CATCH (_) END") 
  is "try_catch_abr" 
  unfolding try_catch_abr_def by simp
 
    
subsection \<open>Lattices\<close>    

lift_definition Lower_a:: "('s, 'e) aprog set \<Rightarrow> ('s, 'e) aprog set"
  is "Lower_abr" 
  by (simp add: Lower_abr_def)
    
lift_definition Upper_a:: "('s, 'e) aprog set \<Rightarrow> ('s, 'e) aprog set"
  is "Upper_abr" 
  by (simp add: Upper_abr_def)
    
lift_definition least_a:: "('s, 'e) aprog \<Rightarrow> (('s, 'e) aprog) set \<Rightarrow> bool"
  is "least_abr".

lift_definition greatest_a:: " ('s, 'e) aprog \<Rightarrow> (('s, 'e) aprog) set \<Rightarrow> bool"
  is "greatest_abr". 
    
lift_definition inf_a :: "(('s, 'e) aprog) set \<Rightarrow> ('s, 'e) aprog" ("\<Sqinter>\<^sub>a_" [900] 900)
  is "inf_abr"    
  by (simp add: inf_abr_prog_is_abrh subsetI)

lift_definition sup_a  :: "(('s, 'e) aprog) set \<Rightarrow> ('s, 'e) aprog" ("\<Squnion>\<^sub>a_" [900] 900)
  is "sup_abr"
  by (simp add: subsetI sup_abr_prog_is_abrh)
   
lift_definition meet_a:: "('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" (infixl "\<sqinter>\<^sub>p" 70)
  is "meet_abr_prog" 
  by (simp add: closure)

lift_definition join_a:: "('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" (infixl "\<squnion>\<^sub>p" 65)
  is "join_abr_prog"
  by (simp add: closure)

subsection \<open>fixed points\<close>    

lift_definition lfp_a :: "(('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog) \<Rightarrow> ('s, 'e) aprog" 
  is "LFP_abr" 
  apply (subst (3) lfp_abr_healthy_comp)
  apply (subst (4) lfp_abr_healthy_comp)
  apply auto
   apply (simp_all add: comp_def lfp_abr_prog_is_abrh)
  done

lift_definition gfp_a :: "(('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog) \<Rightarrow> ('s, 'e) aprog" 
  is "GFP_abr" 
   apply (subst (3) gfp_abr_healthy_comp)
  apply (subst (4) gfp_abr_healthy_comp)
  apply auto
   apply (simp_all add: comp_def gfp_abr_prog_is_abrh)
  done
    
subsubsection \<open>Syntax translation\<close>

syntax
  "_amu" :: "pttrn \<Rightarrow> logic \<Rightarrow> logic" ("\<mu>\<^sub>a _ \<bullet> _" [0, 10] 10)
  "_anu" :: "pttrn \<Rightarrow> logic \<Rightarrow> logic" ("\<nu>\<^sub>a _ \<bullet> _" [0, 10] 10)

notation lfp_a ("\<mu>\<^sub>a")
notation gfp_a ("\<nu>\<^sub>a")

translations
  "\<mu>\<^sub>a X \<bullet> P" == "CONST lfp_a (\<lambda> X. P)"
  "\<nu>\<^sub>a X \<bullet> P" == "CONST gfp_a (\<lambda> X. P)"
  
  
subsection \<open>Iterations\<close>
  
no_notation from_until_lfp_prog  ("FROM (_)/ UNTIL (_)/ DO (_) OD")
definition from_until_lfp_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("FROM (_)/ UNTIL (_)/ DO (_) OD")
where "FROM init UNTIL exit DO body OD = init ; (\<mu>\<^sub>a X \<bullet> IF \<not> exit THEN (body ; X) ELSE SKIP FI)"  

no_notation from_until_lfp_invr_prog("FROM (_)/ INVAR (_)/ UNTIL (_)/  DO (_) OD")   
definition from_until_lfp_invr_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("FROM (_)/ INVAR (_)/ UNTIL (_)/  DO (_) OD")  
where "FROM init INVAR invar UNTIL exit DO body OD = FROM init UNTIL exit DO body OD"

no_notation from_until_lfp_invr_vrt_prog("FROM (_)/ INVAR (_)/ VRT (_)/ UNTIL (_)/ DO (_) OD")  
definition from_until_lfp_invr_vrt_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> ('t,'s) uexpr \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("FROM (_)/ INVAR (_)/ VRT (_)/ UNTIL (_)/ DO (_) OD")
where "FROM init INVAR invar VRT vari UNTIL exit DO body OD = FROM init UNTIL exit DO body OD" 

no_notation while_lfp_prog("WHILE (_)/ DO (_) OD")  
definition while_lfp_a :: 
  "'s cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("WHILE (_)/ DO (_) OD")
where "WHILE b DO body OD = FROM SKIP UNTIL \<not>b DO body OD"    

no_notation while_lfp_invr_prog  ("INVAR (_)/ WHILE (_)/ DO (_) OD")
definition while_lfp_invr_a :: 
  "'s cond \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("INVAR (_)/ WHILE (_)/ DO (_) OD")
where "INVAR invar WHILE b DO body OD = WHILE b DO body OD" 

no_notation while_lfp_invr_vrt_prog ("INVAR (_)/ VRT (_)/ WHILE (_)/ DO (_) OD")  
definition while_lfp_invr_vrt_a :: 
  "'s cond \<Rightarrow> ('t,'s) uexpr \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("INVAR (_)/ VRT (_)/ WHILE (_)/ DO (_) OD")
where "INVAR invar VRT vari WHILE b DO body OD = WHILE b DO body OD"

no_notation do_while_lfp_prog ("DO (_)/  WHILE (_) OD")
definition do_while_lfp_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog" 
  ("DO (_)/  WHILE (_) OD")
where "DO body WHILE b OD = FROM body UNTIL \<not> b DO body OD"

no_notation do_while_lfp_invr_prog ("DO (_)/ INVAR (_)/ WHILE (_) OD")
definition do_while_lfp_invr_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog" 
  ("DO (_)/ INVAR (_)/ WHILE (_) OD")
where "DO body INVAR invar WHILE b OD = DO body WHILE b OD"

no_notation do_while_lfp_invr_vrt_prog  ("DO (_)/ INVAR (_)/ VRT (_)/ WHILE (_) OD") 
definition do_while_lfp_invr_vrt_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> ('t,'s) uexpr \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog" 
  ("DO (_)/ INVAR (_)/ VRT (_)/ WHILE (_) OD")
where "DO body INVAR invar VRT vari WHILE b OD = DO body WHILE b OD"  

no_notation for_lfp_prog  ("FOR ('(_,_,_'))/ DO (_) OD")   
definition for_lfp_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("FOR ('(_,_,_'))/ DO (_) OD")
where "FOR (init, b, incr) DO body OD = FROM init UNTIL \<not> b DO body ; incr OD"

no_notation for_lfp_invr_prog  ("FOR ('(_,_,_'))/ INVAR (_)/ DO (_) OD")   

definition for_lfp_invr_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("FOR ('(_,_,_'))/ INVAR (_)/ DO (_) OD")
where "FOR (init, b, incr) INVAR invar DO body OD = FOR (init, b, incr) DO body OD"  

no_notation for_lfp_invr_vrt_prog  ("FOR ('(_,_,_'))/ INVAR (_)/ VRT (_)/ DO (_) OD")   

definition for_lfp_invr_vrt_a :: 
  "('s, 'e) aprog \<Rightarrow> 's cond \<Rightarrow> ('s, 'e) aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'s) uexpr \<Rightarrow> ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" 
  ("FOR ('(_,_,_'))/ INVAR (_)/ VRT (_)/ DO (_) OD")
where "FOR (init, b, incr) INVAR invar VRT vari DO body OD = FOR (init, b, incr) DO body OD"  

subsection \<open>Frame and anti-frame\<close>

abbreviation 
  "frame_abr x P \<equiv>  abrh (pframe_prog (lens_comp x \<Sigma>\<^sub>A\<^sub>B\<^sub>R) P)"
abbreviation 
  "anti_frame_abr x P \<equiv>  abrh (pantiframe_prog (lens_comp x \<Sigma>\<^sub>A\<^sub>B\<^sub>R) P)" 
no_notation pframe_prog  ("RESTOR (_)/ DO (_) OD")  

lift_definition frame_a :: "('a \<Longrightarrow> 's) \<Rightarrow>  ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" ("RESTOR (_)/ DO (_) OD") is
  "frame_abr " by simp
no_notation pantiframe_prog  ("MODIFY (_)/ DO (_) OD")  

lift_definition antiframe_a :: "('a \<Longrightarrow> 's) \<Rightarrow>  ('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog" ("MODIFY (_)/ DO (_) OD") is
  "anti_frame_abr" by simp
    
translations
  "_frame x P" => "CONST frame_a x P"
  "_frame (_salphaset (_salphamk x)) P" <= "CONST frame_a x P"
  "_antiframe x P" => "CONST antiframe_a x P"
  "_antiframe (_salphaset (_salphamk x)) P" <= "CONST antiframe_a x P"
  
subsection \<open>monotonicity for PROG\<close>
lift_definition mono_a::"(('s, 'e) aprog \<Rightarrow> ('s, 'e) aprog) \<Rightarrow> bool" is
  "mono_abr_prog" 
  unfolding prog_order_def abrh_hcond_def
  by (simp add: isotone_def)
 (*TODO CONTINUE FROM HERE!*)

  
section \<open>Transfer rules\<close>  
declare lfp_a.rep_eq       [aprog_rep_eq]  
declare gfp_a.rep_eq       [aprog_rep_eq]  
declare meet_a.rep_eq      [aprog_rep_eq]  
declare join_a.rep_eq      [aprog_rep_eq]    
declare Lower_a.rep_eq     [aprog_rep_eq]  
declare Upper_a.rep_eq     [aprog_rep_eq] 
declare least_a.rep_eq     [aprog_rep_eq]  
declare greatest_a.rep_eq  [aprog_rep_eq]   
declare inf_a.rep_eq       [aprog_rep_eq]  
declare sup_a.rep_eq       [aprog_rep_eq]  
declare abort_a.rep_eq     [aprog_rep_eq]
declare magic_a.rep_eq     [aprog_rep_eq]
declare skip_a.rep_eq      [aprog_rep_eq]
declare seq_a.rep_eq       [aprog_rep_eq]
declare assigns_a.rep_eq   [aprog_rep_eq]
declare try_catch_a.rep_eq [aprog_rep_eq]
declare throw_a.rep_eq     [aprog_rep_eq]
declare cond_a.rep_eq      [aprog_rep_eq]
declare frame_a.rep_eq     [aprog_rep_eq]
declare antiframe_a.rep_eq [aprog_rep_eq]   
declare psubst.rep_eq      [aprog_rep_eq]
  
section \<open>Substitution Laws\<close>
(*sledgehammer_params[stop_on_first,parallel_subgoals, join_subgoals, timeout = 60]*)
  
lemma psubst_seq [usubst]:
  "\<sigma> \<dagger> (P ; Q) = (\<sigma> \<dagger> P) ; Q"
  apply transfer 
  apply (simp add: psubst_seq abrh_def)
  apply (metis abrh_def comp_prog_if_prog_left_distr if_prog_L6 pseq_abrh_closed)
  done
  
section \<open>Laws of Programming\<close>

theorem skip_a_left_unit [simp]:
  "SKIP ; P = P"
  by atransfer
 
theorem skip_a_right_unit [simp]:
  "P ; SKIP = P"
   by atransfer

theorem abort_a_left_zero [simp]:
  "ABORT ; P = ABORT"
  apply ( atransfer, ptransfer)
  apply (metis Rep_prog_abrh_closed abort_left_zero abrh_def comp_prog_if_prog_left_distr if_prog_L6 if_prog_assoc pseq.rep_eq skip_prog_left_unit) 
done
  
theorem magic_a_left_zero [simp]:
  "MAGIC ; P = MAGIC"
  apply atransfer
  apply ptransfer
  unfolding abrh_def
  apply (metis Rep_prog_abrh_closed abrh_def comp_prog_if_prog_left_distr if_prog_L6 if_prog_assoc  magic_left_zero pseq.rep_eq skip_prog_left_unit)
  done


lemma seq_a_throw_a_try_catch1:
  "TRY THROW(e) CATCH Q END = Q"
  apply transfer
  unfolding throw_abr_def abrh_def assigns_abr_def try_catch_abr_def 
  apply ptransfer
  apply rel_simp
  apply transfer
  apply auto
  unfolding Healthy_def H1_def H3_def
  apply (rel_simp, metis option.distinct(1))+
  done         
  
lemma throw_left_zero:
  "THROW(e); P = THROW(e)"
  apply transfer
  unfolding throw_abr_def abrh_def assigns_abr_def try_catch_abr_def    
  apply ptransfer    
  apply rel_simp
  apply transfer
  apply safe
    apply simp_all
   unfolding Healthy_def H1_def H3_def
   apply (rel_simp,  metis  option.distinct(1) option.inject)+
 done  

lemma try_catch_skip:
  "TRY SKIP CATCH Q END = SKIP"
  apply transfer                                               
  unfolding throw_abr_def abrh_def assigns_abr_def try_catch_abr_def    
  apply ptransfer    
  apply rel_simp
  apply transfer
  apply safe
  unfolding Healthy_def H1_def H3_def
  apply blast+
 done  

lemma try_catch_assigns_a:
  "TRY \<langle>\<sigma>\<rangle>\<^sub>a CATCH Q END = \<langle>\<sigma>\<rangle>\<^sub>a"
  apply transfer
  unfolding throw_abr_def abrh_def assigns_abr_def try_catch_abr_def    
  apply ptransfer    
  apply rel_simp
  apply transfer
  apply safe
    unfolding Healthy_def H1_def H3_def
  apply (rel_simp,  metis)+ 
  done  

lemma    
  "(TRY \<langle>\<sigma>\<rangle>\<^sub>a ; THROW(e) CATCH Q END) = (\<langle>\<sigma>\<rangle>\<^sub>a ; Q)"
  apply (simp add: aprog_rep_eq) 
  apply transfer
  unfolding throw_abr_def abrh_def assigns_abr_def try_catch_abr_def
  apply (simp add: prog_rep_eq) 
  apply rel_simp
  apply transfer
  unfolding Healthy_def H1_def H3_def
  apply rel_simp
  apply safe
   apply simp_all
  apply (metis option.distinct(1))+
  done

section \<open>DSL\<close>

term " SKIP "

term " THROW (e)"

term " (x :== e) :: ('s, 'e) aprog "

term " IF b THEN P\<^sub>1 ELSE P\<^sub>2 FI "

term " WHILE b DO body OD "

term " INVAR invar VRT vari WHILE b DO body OD "

term " FROM init UNTIL exit DO body OD "

term " FOR (init, b, incr) DO body OD "

term " DO body WHILE b OD "

term " MODIFY VARS DO body OD "

term " RESTOR VARS DO body OD "

term " TRY P\<^sub>1 CATCH P\<^sub>2 END "

term " TRY WHILE b DO body\<^sub>1 ; THROW(e) ; body\<^sub>2 OD CATCH SKIP END "

term " WHILE b DO TRY body\<^sub>1 ; THROW(e) ; body\<^sub>2 CATCH SKIP END OD "

term " MODIFY VARS DO TRY \<langle>init\<rangle>\<^sub>a; body CATCH  restore ; THROW(e) END OD; return"

section \<open>Hoare logic\<close>

term "((Abs_aprog o Abs_prog) ((P \<oplus>\<^sub>p \<Sigma>\<^sub>A\<^sub>B\<^sub>R) \<turnstile>\<^sub>n (\<lceil>(Q \<oplus>\<^sub>p \<Sigma>\<^sub>A\<^sub>B\<^sub>R) \<and> (A \<oplus>\<^sub>p abrupt)\<rceil>\<^sub>>))  \<sqsubseteq> PROG)"
definition hoare_aprog_t :: \<open>'s upred  \<Rightarrow> ('s) upred \<Rightarrow> 
                             'e option upred \<Rightarrow> ('s, 'e) aprog \<Rightarrow> bool\<close>
where "hoare_aprog_t P Q A PROG  = 
       ((Abs_aprog o Abs_prog) ((P \<oplus>\<^sub>p \<Sigma>\<^sub>A\<^sub>B\<^sub>R) \<turnstile>\<^sub>n (\<lceil>(Q \<oplus>\<^sub>p \<Sigma>\<^sub>A\<^sub>B\<^sub>R) \<and> (A \<oplus>\<^sub>p abrupt)\<rceil>\<^sub>>))  \<sqsubseteq> PROG)"
term \<open>hoare_aprog_t P (&x =\<^sub>u \<guillemotleft>1\<guillemotright>) (\<guillemotleft>e \<notin> REF\<guillemotright>) (THROW(e); x :== \<guillemotleft>1\<guillemotright>)\<close>   
end
