(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)
  
theory utp_prog_des_lift
  imports 
    "../../Isabelle-UTP/impl/utp_prog"
begin


section \<open>More Operators\<close>

subsection \<open>Conditional\<close>
 
lift_definition pcond_prog :: "'\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" ("IF (_)/ THEN (_) ELSE (_) FI") 
  is "IfD" 
  by (simp add: H1_H3_impl_H2 H3_unrest_out_alpha if_d_H1_H3_intro)
    
declare pcond_prog.rep_eq [prog_rep_eq]
  
subsection \<open>assert and assume\<close>

abbreviation passume_prog :: "'\<alpha> cond \<Rightarrow> '\<alpha> prog" ("_\<^sup>\<top>\<^sup>P")  
  where "passume_prog c \<equiv> (IF c THEN SKIP ELSE magic FI)"

abbreviation passert_prog :: "'\<alpha> upred \<Rightarrow> '\<alpha> prog" ("_\<^sub>\<bottom>\<^sub>P")
  where "passert_prog c \<equiv> (IF c THEN SKIP ELSE abort FI)"
    
subsection \<open>Recursion\<close>    

lift_definition Lower_prog:: "'\<alpha> prog set \<Rightarrow> '\<alpha> prog set"
  is "Lower (uthy_order NDES)" 
  by (metis Lower_closed is_Healthy_subset_member is_Ncarrier_is_ndesigns utp_order_carrier)
    
lift_definition Upper_prog:: "'\<alpha> prog set \<Rightarrow> '\<alpha> prog set"
  is "Upper (uthy_order NDES)" 
  by (metis Upper_closed is_Healthy_subset_member is_Ncarrier_is_ndesigns utp_order_carrier)
  
lift_definition least_prog:: " '\<alpha> prog \<Rightarrow> ('\<alpha> prog) set \<Rightarrow> bool"
  is "least (uthy_order NDES)".

lift_definition greatest_prog:: " '\<alpha> prog \<Rightarrow> ('\<alpha> prog) set \<Rightarrow> bool"
  is "greatest (uthy_order NDES)". 
    
lift_definition inf_prog :: "('\<alpha> prog) set \<Rightarrow> '\<alpha> prog" ("\<Sqinter>\<^sub>p_" [900] 900)
  is "inf (uthy_order NDES)"  
  by (simp add: subsetI)

lift_definition sup_prog  :: "('\<alpha> prog) set \<Rightarrow> '\<alpha> prog" ("\<Squnion>\<^sub>p_" [900] 900)
  is "sup (uthy_order NDES)"
  by (simp add: subsetI)
    
declare Lower_prog.rep_eq    [prog_rep_eq]  
declare Upper_prog.rep_eq    [prog_rep_eq] 
declare least_prog.rep_eq    [prog_rep_eq]  
declare greatest_prog.rep_eq [prog_rep_eq]   
declare inf_prog.rep_eq      [prog_rep_eq]  
declare sup_prog.rep_eq      [prog_rep_eq]
   
lift_definition meet_prog:: "'\<alpha> prog \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" (infixl "\<sqinter>\<^sub>p" 70)
  is "meet (uthy_order NDES)"
  by (simp add: closure)

lift_definition join_prog:: "'\<alpha> prog \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" (infixl "\<squnion>\<^sub>p" 65)
  is "join (uthy_order NDES)"
  by (simp add: closure)

declare meet_prog.rep_eq [prog_rep_eq]  
declare join_prog.rep_eq [prog_rep_eq]


lift_definition prec_lfp_prog :: "('\<alpha> prog \<Rightarrow> '\<alpha> prog) \<Rightarrow> '\<alpha> prog" 
  is "ndesign_lfp" 
  apply (simp)
  apply (subst (1) normal_design_theory_continuous.LFP_healthy_comp)
  apply (subst (2) normal_design_theory_continuous.LFP_healthy_comp) 
  apply (simp add: comp_def H1_H3_commute H1_idem H3_idem Healthy_def ndes_hcond_def)
  done

lift_definition prec_gfp_prog :: "('\<alpha> prog \<Rightarrow> '\<alpha> prog) \<Rightarrow> '\<alpha> prog" 
  is "ndesign_gfp" 
  apply (simp)
  apply (subst (1) normal_design_theory_continuous.GFP_healthy_comp)
  apply (subst (2) normal_design_theory_continuous.GFP_healthy_comp) 
  apply (simp add: comp_def H1_H3_commute H1_idem H3_idem Healthy_def ndes_hcond_def)
  done
    
declare prec_lfp_prog.rep_eq [prog_rep_eq]  
declare prec_gfp_prog.rep_eq [prog_rep_eq]  
  
syntax
  "_pmu" :: "pttrn \<Rightarrow> logic \<Rightarrow> logic" ("\<mu>\<^sub>p _ \<bullet> _" [0, 10] 10)
  "_pnu" :: "pttrn \<Rightarrow> logic \<Rightarrow> logic" ("\<nu>\<^sub>p _ \<bullet> _" [0, 10] 10)

notation prec_lfp_prog ("\<mu>\<^sub>p")
notation prec_gfp_prog ("\<nu>\<^sub>p")

translations
  "\<mu>\<^sub>p X \<bullet> P" == "CONST prec_lfp_prog (\<lambda> X. P)"
  "\<nu>\<^sub>p X \<bullet> P" == "CONST prec_gfp_prog (\<lambda> X. P)" 

subsection{*Iterations*}    

definition from_until_lfp_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("FROM (_)/ UNTIL (_)/ DO (_) OD")
where "FROM init UNTIL exit DO body OD = init ; (\<mu>\<^sub>p X \<bullet> IF \<not> exit THEN (body ; X) ELSE SKIP FI)"  

declare from_until_lfp_prog_def [prog_rep_eq]

definition from_until_lfp_invr_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("FROM (_)/ INVAR (_)/ UNTIL (_)/  DO (_) OD")  
where "FROM init INVAR invar UNTIL exit DO body OD = FROM init UNTIL exit DO body OD"

declare from_until_lfp_invr_prog_def [prog_rep_eq]

definition from_until_lfp_invr_vrt_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("FROM (_)/ INVAR (_)/ VRT (_)/ UNTIL (_)/ DO (_) OD")
  where "FROM init INVAR invar VRT vari UNTIL exit DO body OD = FROM init UNTIL exit DO body OD" 

declare from_until_lfp_invr_vrt_prog_def [prog_rep_eq]

definition while_lfp_prog :: 
  "'\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("WHILE (_)/ DO (_) OD")
  where "WHILE b DO body OD = FROM SKIP UNTIL \<not>b DO body OD"   

declare while_lfp_prog_def [prog_rep_eq]

definition while_lfp_invr_prog :: 
  "'\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("INVAR (_)/ WHILE (_)/ DO (_) OD")
where "INVAR invar WHILE b DO body OD = WHILE b DO body OD" 

declare while_lfp_invr_prog_def [prog_rep_eq]

definition while_lfp_invr_vrt_prog :: 
  "'\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("INVAR (_)/ VRT (_)/ WHILE (_)/ DO (_) OD")
where "INVAR invar VRT vari WHILE b DO body OD = WHILE b DO body OD"

declare while_lfp_invr_vrt_prog_def [prog_rep_eq]

definition do_while_lfp_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog" 
  ("DO (_)/  WHILE (_) OD")
where "DO body WHILE b OD = FROM body UNTIL \<not> b DO body OD"

declare do_while_lfp_prog_def [prog_rep_eq]

definition do_while_lfp_invr_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog" 
  ("DO (_)/ INVAR (_)/ WHILE (_) OD")
where "DO body INVAR invar WHILE b OD = DO body WHILE b OD"

declare do_while_lfp_invr_prog_def [prog_rep_eq]

definition do_while_lfp_invr_vrt_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog" 
  ("DO (_)/ INVAR (_)/ VRT (_)/ WHILE (_) OD")
where "DO body INVAR invar VRT vari WHILE b OD = DO body WHILE b OD"  

declare do_while_lfp_invr_vrt_prog_def [prog_rep_eq]

definition for_lfp_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("FOR ('(_,_,_'))/ DO (_) OD")
  where "FOR (init, b, incr) DO body OD = FROM init UNTIL \<not> b DO body ; incr OD"

declare for_lfp_prog_def [prog_rep_eq]

definition for_lfp_invr_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("FOR ('(_,_,_'))/ INVAR (_)/ DO (_) OD")
where "FOR (init, b, incr) INVAR invar DO body OD = FOR (init, b, incr) DO body OD"  

declare for_lfp_invr_prog_def [prog_rep_eq]

definition for_lfp_invr_vrt_prog :: 
  "'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("FOR ('(_,_,_'))/ INVAR (_)/ VRT (_)/ DO (_) OD")
  where "FOR (init, b, incr) INVAR invar VRT vari DO body OD = FOR (init, b, incr) DO body OD"  

declare for_lfp_invr_vrt_prog_def [prog_rep_eq]

subsection{*Frame and anti-frame*}

lift_definition pframe_prog :: "('a \<Longrightarrow> '\<alpha>) \<Rightarrow>  '\<alpha> prog \<Rightarrow> '\<alpha> prog " ("RESTOR (_)/ DO (_) OD") is
  "frame\<^sub>D" 
  by rel_blast 

lift_definition pantiframe_prog :: "('a \<Longrightarrow> '\<alpha>) \<Rightarrow>  '\<alpha> prog \<Rightarrow> '\<alpha> prog " ("MODIFY (_)/ DO (_) OD") is
  "antiframe\<^sub>D"  
  by rel_blast 

subsection{*Lift and transfer package setup*}  
declare pframe_prog.rep_eq [prog_rep_eq]
declare pantiframe_prog.rep_eq [prog_rep_eq] 
  
translations
  "_frame x P" => "CONST pframe_prog x P"
  "_frame (_salphaset (_salphamk x)) P" <= "CONST pframe_prog x P"
  "_antiframe x P" => "CONST pantiframe_prog x P"
  "_antiframe (_salphaset (_salphamk x)) P" <= "CONST pantiframe_prog x P"

subsection \<open>monotonicity for PROG\<close>

lift_definition mono_prog::  "('a prog \<Rightarrow> 'a prog) \<Rightarrow> bool" 
 is "Mono\<^bsub>uthy_order NDES\<^esub>"   
  unfolding utp_order_def ndes_hcond_def
  by (simp add: isotone_def) 
  
lift_definition idem_prog::  "('a prog \<Rightarrow> 'a prog) \<Rightarrow> bool" 
 is "Idem\<^bsub>uthy_order NDES\<^esub>"
   unfolding utp_order_def ndes_hcond_def idempotent_def Ball_def
   by (metis (no_types, lifting) mem_Collect_eq partial_object.select_convs(1))
  
declare mono_prog.rep_eq[prog_rep_eq]
declare idem_prog.rep_eq[prog_rep_eq]  
         
end
