(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)

chapter \<open>Abrupt termination and exceptions\<close>
  
theory utp_prog_abrupt
  imports 
    "../../HoareLogic/TotalCorrectness/utp_hoare_ndes_prog"
begin
  
section \<open>parallel sledgehammer configuration\<close>
  
section \<open>Alphabet\<close>

text 
\<open>In order to capture exceptions, we extend the alphabet of UTP by an additional global 
   state variables:
   \begin{itemize}   
      \<^item> abrupt: a boolean variable used to specify if the program is in an abrupt state or not.
   \end{itemize}\<close>

alphabet 'e abr_vars = 
  abrupt :: \<open>'e option\<close>

declare  abr_vars.splits [alpha_splits]
  
subsection \<open>Alphabet proofs\<close>
text \<open>
  The two locale interpretations below are a technicality to improve automatic
  proof support via the predicate and relational tactics. This is to enable the
  (re-)interpretation of state spaces to remove any occurrences of lens types
  after the proof tactics @{method pred_simp} and @{method rel_simp}, or any
  of their derivatives have been applied. Eventually, it would be desirable to
  automate both interpretations as part of a custom outer command for defining
  alphabets.\<close>

interpretation cp_abr:
  lens_interp "\<lambda> (ok, r) . (ok, abrupt\<^sub>v r, more r)"
apply (unfold_locales)
apply (rule injI)
apply (clarsimp)
done

interpretation cp_abr_rel: lens_interp "\<lambda>(ok, ok', r, r').
  (ok, ok', abrupt\<^sub>v r, abrupt\<^sub>v r', more r, more r')"
apply (unfold_locales)
apply (rule injI)
apply (clarsimp)
done

subsection \<open>Type lifting\<close>

type_synonym ('s, 'e) abr_prog  = "('e, 's) abr_vars_scheme prog"
type_synonym ('s, 'e) abr  = "('e , 's) abr_vars_scheme"

translations
  (type) "('s, 'e) abr" <= (type) "('e, 's) abr_vars_scheme"
  (type) "('s, 'e) abr" <= (type) "('e, 's) abr_vars_ext"
  (type) "('s, 'e) abr_prog" <= (type) "('e, 's) abr_vars_scheme prog"
  (type) "('s, 'e) abr_prog" <= (type) "('e, 's) abr_vars_ext prog"

  
subsection \<open>Syntactic type setup\<close>

notation abr_vars_child_lens ("\<Sigma>\<^sub>A\<^sub>B\<^sub>R")

syntax
  "_svid_st_alpha"  :: "svid" ("\<Sigma>\<^sub>A\<^sub>B\<^sub>R")
translations
  "_svid_st_alpha" => "CONST abr_vars_child_lens"

subsection \<open>Substitution lift and drop\<close>
  
abbreviation lift_rel_usubst_abr ("\<lceil>_\<rceil>\<^sub>S\<^sub>A\<^sub>B\<^sub>R")
where "\<lceil>\<sigma>\<rceil>\<^sub>S\<^sub>A\<^sub>B\<^sub>R \<equiv> \<sigma> \<oplus>\<^sub>s \<Sigma>\<^sub>A\<^sub>B\<^sub>R "

abbreviation lift_usubst_abr ("\<lceil>_\<rceil>\<^sub>s\<^sub>A\<^sub>B\<^sub>R")
  where "\<lceil>\<sigma>\<rceil>\<^sub>s\<^sub>A\<^sub>B\<^sub>R \<equiv> \<lceil>\<lceil>\<sigma>\<rceil>\<^sub>s\<rceil>\<^sub>S\<^sub>A\<^sub>B\<^sub>R"
    
abbreviation drop_usubst_abr ("\<lfloor>_\<rfloor>\<^sub>S\<^sub>A\<^sub>B\<^sub>R")
where "\<lfloor>\<sigma>\<rfloor>\<^sub>S\<^sub>A\<^sub>B\<^sub>R \<equiv> \<sigma> \<restriction>\<^sub>s \<Sigma>\<^sub>A\<^sub>B\<^sub>R "

abbreviation drop_usubst_abr_rel ("\<lfloor>_\<rfloor>\<^sub>s\<^sub>A\<^sub>B\<^sub>R")
where "\<lfloor>\<sigma>\<rfloor>\<^sub>s\<^sub>A\<^sub>B\<^sub>R \<equiv> \<lfloor>\<lfloor>\<sigma>\<rfloor>\<^sub>S\<^sub>A\<^sub>B\<^sub>R\<rfloor>\<^sub>s"

subsection \<open>Expressions lift and drop\<close>

abbreviation lift_uexpr_abr ("\<lceil>_\<rceil>\<^sub>A\<^sub>B\<^sub>R")
where "\<lceil>P\<rceil>\<^sub>A\<^sub>B\<^sub>R \<equiv> P \<oplus>\<^sub>p \<Sigma>\<^sub>A\<^sub>B\<^sub>R "

abbreviation lift_pre_uexpr_abr ("\<lceil>_\<rceil>\<^sub>A\<^sub>B\<^sub>R\<^sub><")
where "\<lceil>p\<rceil>\<^sub>A\<^sub>B\<^sub>R\<^sub>< \<equiv> \<lceil>\<lceil>p\<rceil>\<^sub><\<rceil>\<^sub>A\<^sub>B\<^sub>R"

abbreviation lift_post_uexpr_abr ("\<lceil>_\<rceil>\<^sub>A\<^sub>B\<^sub>R\<^sub>>")
where "\<lceil>p\<rceil>\<^sub>A\<^sub>B\<^sub>R\<^sub>> \<equiv> \<lceil>\<lceil>p\<rceil>\<^sub>>\<rceil>\<^sub>A\<^sub>B\<^sub>R"

abbreviation drop_uexpr_abr ("\<lfloor>_\<rfloor>\<^sub>A\<^sub>B\<^sub>R")
where "\<lfloor>P\<rfloor>\<^sub>A\<^sub>B\<^sub>R \<equiv> P \<restriction>\<^sub>p \<Sigma>\<^sub>A\<^sub>B\<^sub>R"

abbreviation drop_pre_uexpr_abr ("\<lfloor>_\<rfloor>\<^sub><\<^sub>A\<^sub>B\<^sub>R")
where "\<lfloor>P\<rfloor>\<^sub><\<^sub>A\<^sub>B\<^sub>R \<equiv> \<lfloor>\<lfloor>P\<rfloor>\<^sub>A\<^sub>B\<^sub>R\<rfloor>\<^sub><"

abbreviation drop_post_uexpr_abr ("\<lfloor>_\<rfloor>\<^sub>>\<^sub>A\<^sub>B\<^sub>R")
where "\<lfloor>P\<rfloor>\<^sub>>\<^sub>A\<^sub>B\<^sub>R \<equiv> \<lfloor>\<lfloor>P\<rfloor>\<^sub>A\<^sub>B\<^sub>R\<rfloor>\<^sub>>"


section \<open>Healthiness conditions\<close>
  
definition abrh_def [upred_defs]: 
  "abrh(P) = (IF &abrupt =\<^sub>u \<guillemotleft>None\<guillemotright> THEN P  ELSE SKIP FI)"

lemma abrh_idem[simp]:"(abrh (abrh X)) = (abrh X)"  
  unfolding abrh_def
  by (metis if_prog_L6 if_prog_idem)
    
lemma abrh_subset_member:
  "\<lbrakk>A \<subseteq> {P. abrh P = P}; P \<in> A \<rbrakk> \<Longrightarrow> abrh(P) = P"
  by (meson Ball_Collect )

lemma abrh_image_subset:
  "abrh ` A \<subseteq> {P. abrh P = P}"
  by (simp add: image_subsetI)

lemma abr_abrh_closed[closure]:
  "(abrh P) \<in> {P. abrh P = P}"
  by simp

lemma mem_collectP:
 "x \<in> {P. abrh P = P} = (abrh x = x)"
  by auto

section \<open>Control flow statements\<close>
  
subsection \<open>Top for exceptions theory\<close>

abbreviation
  "top_abr \<equiv> abrh MAGIC"

lemma abrh_below_top_abr_prog:
  "x \<in> {P. abrh P = P} \<Longrightarrow> x \<sqsubseteq> abrh MAGIC"
  apply (simp add: abrh_def  prog_rep_eq   )
  apply (metis H1_below_top Healthy_def Rep_prog_H1_H3_closed cond_mono order_refl)
  done

subsection \<open>Bot for exceptions theory\<close>

abbreviation
  "bot_abr \<equiv> abrh ABORT"

lemma bottom_abr_prog_abrh_lower_prog:
  "x \<in> {P. abrh P = P} \<Longrightarrow> abrh (ABORT) \<sqsubseteq> x"
  apply (simp add: abrh_def  prog_rep_eq   )
  apply (metis cond_mono order_refl utp_pred_laws.top_greatest)
  done

subsection \<open>SKIP\<close>
  
text \<open>For the below reason we do not re-define skip\<close>
  
subsubsection \<open>Algebraic laws\<close>

lemma abrh_skipe[simp]:
  "abrh(SKIP) = SKIP"
  unfolding abrh_def
  by (simp add: if_prog_idem)

subsection \<open>Assign\<close>
  
definition  
assigns_abr_def: "assigns_abr \<sigma> =  abrh(\<langle>\<sigma>  \<oplus>\<^sub>s \<Sigma>\<^sub>A\<^sub>B\<^sub>R\<rangle>\<^sub>p)"

adhoc_overloading
  uassigns assigns_abr

translations
  "_assignment xs vs" => "CONST assigns_abr (_mk_usubst (CONST id) xs vs)"
  "x :== v" <= "CONST assigns_abr (CONST subst_upd (CONST id) (CONST svar x) v)"

subsubsection \<open>Algebraic laws\<close>

subsection \<open>Throw\<close>

definition  throw_abr:: "'e \<Rightarrow> ('s, 'e) abr_prog" ("THROW '(_')")
where       "THROW(e) =  abrh(passigns (subst_upd id abrupt (\<guillemotleft>Some e\<guillemotright>)))"

subsubsection \<open>Algebraic laws\<close>
  
subsection \<open>Sequential composition\<close>
  
text  \<open>It does not need a definition. We get it by type inference.\<close>  
  
lemma pseq_abrh_closed [closure]: (*TODO @Yakoub: make this proof  nicer with clearer reasoning*)
  assumes "abrh P = P" "abrh Q = Q"
  shows "abrh (P ; Q) = (P ; Q)"  
  using assms
  unfolding abrh_def   
  apply (simp add: prog_rep_eq)    
  apply rel_simp 
  apply safe
           apply auto
      apply metis
  apply metis
    apply (subst (asm) not_None_eq[THEN sym])+
  apply (metis option.distinct(1))+
  done 
    
subsection \<open>Conditional\<close>

  
definition if_abr :: "'s upred \<Rightarrow> ('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog" 
  where  "if_abr b P Q = pcond_prog (b \<oplus>\<^sub>p \<Sigma>\<^sub>A\<^sub>B\<^sub>R) P Q"

subsubsection \<open>Algebraic laws\<close>
    
subsection \<open>Try catch\<close>
    
definition try_catch_abr:: "('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog"
where "try_catch_abr P Q =
       abrh (P ; pcond_prog (&abrupt =\<^sub>u \<guillemotleft>None\<guillemotright>) SKIP ((passigns (subst_upd id abrupt \<guillemotleft>None\<guillemotright>)) ; Q))"

term \<open>(passigns (subst_upd id abrupt \<guillemotleft>None\<guillemotright>)) ; Q\<close>
(*definition try_catch_abr:: "'a abr_prog \<Rightarrow> 'a abr_prog \<Rightarrow> 'a abr_prog"
where "try_catch_abr P Q =
       abrh (P ; pcond_prog (&abrupt) (passigns (subst_upd id abrupt (\<not> &abrupt)) ; Q) SKIP)"*)

subsubsection \<open>Algebraic laws\<close>

section \<open>Lattices for exceptions\<close>

definition                                                            
  "greatest_abr g A \<longleftrightarrow> A \<subseteq> {P. abrh P = P} \<and> greatest_prog g A"  
     
lemma greatest_abr_prog_alt_def: 
 "greatest_abr g A = (A \<subseteq> {P. abrh P = P} \<and> g \<in> A \<and> (ALL x : A. x \<sqsubseteq> g))"
  unfolding greatest_abr_def  prog_rep_eq greatest_def image_def Ball_def
  by (auto simp add: Rep_prog_inject Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns)
    
definition                                                            
  "least_abr g A \<longleftrightarrow> A \<subseteq> {P. abrh P = P} \<and> least_prog g A"
  
lemma least_abr_prog_alt_def: 
 "least_abr l A = (A \<subseteq> {P. abrh P = P} \<and> l \<in> A \<and> (ALL x : A. l \<sqsubseteq> x))"
  unfolding least_abr_def  prog_rep_eq least_def image_def Ball_def
  by (auto simp add: Rep_prog_inject Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns)
    
definition
  "Lower_abr A = ((Lower_prog (A\<inter> {P. abrh P = P})) \<inter> {P. abrh P = P})"
  
lemma lower_abr_prog_alt_def:
  "Lower_abr A = {l. (ALL x. x \<in> (A \<inter> {P. abrh P = P}) \<longrightarrow> l \<sqsubseteq> x)} \<inter> {P. abrh P = P}"
  unfolding Lower_abr_def  prog_rep_eq Lower_def image_def Ball_def Lower_prog_def
  apply (simp only: Int_Collect)
  apply auto
   apply (metis (mono_tags, lifting) Abs_prog_Rep_prog_Ncarrier Healthy_if Int_Collect Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns)
  unfolding Bex_def
  apply auto
  apply (metis Rep_prog_H1_H3_closed Rep_prog_inverse is_Ncarrier_is_ndesigns)
  done

lemma Lower_abr_prog_subset_Lower_prog:
  "Lower_abr (A \<inter> {P. abrh P = P}) \<subseteq> Lower_prog (A \<inter> {P. abrh P = P})"
  by (simp add: Lower_abr_def)
    
definition
  "Upper_abr A = ((Upper_prog (A \<inter> {P. abrh P = P})) \<inter> {P. abrh P = P})"

lemma upper_abr_prog_alt_def:
  "Upper_abr A = {u. (ALL x. x \<in> (A \<inter> {P. abrh P = P}) \<longrightarrow> x \<sqsubseteq> u)} \<inter> {P. abrh P = P}"  
  unfolding Upper_abr_def prog_rep_eq Upper_def image_def Ball_def Upper_prog_def
  apply (simp only: Int_Collect)
  apply auto
   apply (metis (mono_tags, lifting) Abs_prog_Rep_prog_Ncarrier Healthy_if Int_Collect 
                                     Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns)
  unfolding Bex_def
  apply (subst Int_Collect)
  apply (subst Int_Collect)
  apply simp
  apply (metis Rep_prog_H1_H3_closed Rep_prog_inverse is_Ncarrier_is_ndesigns)
  done
 
lemma upper_abr_prog_subset_upper_prog:
  "Upper_abr (A \<inter> {P. abrh P = P}) \<subseteq> Upper_prog (A \<inter> {P. abrh P = P})"
  by (simp add: Upper_abr_def)

subsection \<open>Inf for exceptions\<close>

definition inf_abr :: "('s, 'e) abr_prog set \<Rightarrow> ('s, 'e) abr_prog" ("\<^bold>\<Sqinter>\<^sub>a_" [900] 900)
where "inf_abr A = (SOME x. greatest_abr x (Lower_abr A))"

subsubsection \<open>Algebraic laws\<close>
    
lemma inf_prog_Int_abrh:
  "(\<And>x. x \<in> A \<Longrightarrow> abrh x = x) \<Longrightarrow> 
     \<Sqinter>\<^sub>p(A \<inter> {P. abrh P = P}) = \<Sqinter>\<^sub>p A"
  apply (simp add: prog_rep_eq)
  unfolding inf_def image_def Ball_def 
  apply (rule someI2_ex)
   apply (rule exI[where x="(\<^bold>\<Sqinter>\<^bsub>NDES\<^esub>{y. \<exists>x\<in>A. y = \<lbrakk>x\<rbrakk>\<^sub>p})"])
   apply (rule normal_design_theory_continuous.inf_glb)
  using Rep_prog apply auto[]
  apply (rule ndes_utp_theory.greatest_unique)
    defer
   apply auto[1]
      apply (rule someI2_ex)
   apply (rule exI[where x="(\<^bold>\<Sqinter>\<^bsub>NDES\<^esub>{y. \<exists>x\<in>A \<inter> {P. \<lbrakk>abrh P\<rbrakk>\<^sub>p = \<lbrakk>P\<rbrakk>\<^sub>p}. y = \<lbrakk>x\<rbrakk>\<^sub>p})"])
       apply (rule normal_design_theory_continuous.inf_glb)
  using Rep_prog apply auto[]
  unfolding Bex_def
  apply (subst (asm) Int_Collect)
  apply simp
  apply (subst Collect_cong)
   apply simp_all
    apply blast
  done

lemma inf_prog_abrh_greatest_prog:
  "(\<And>x. x \<in> A \<Longrightarrow> abrh z \<sqsubseteq> abrh x) \<Longrightarrow> abrh z \<sqsubseteq> \<Sqinter>\<^sub>p (abrh ` A)"
  by (metis (mono_tags, lifting) imageE inf_prog_greatest)
    
lemma inf_prog_abrh_image_lower_prog:
  "x \<in> A \<Longrightarrow> \<Sqinter>\<^sub>p (abrh ` A) \<sqsubseteq> abrh x"
  by (simp add: inf_prog_lower)   

lemma inf_prog_abrh_in_Lower_prog_Int_abrh:
  "abrh (\<Sqinter>\<^sub>pA) \<in> Lower_prog (A \<inter> {P. abrh P = P})" 
   unfolding inf_prog_alt_def 
    apply (rule someI2_ex)    
   using inf_prog_is_greatest_lower_prog 
    apply blast
    apply (simp only: greatest_prog_alt_def Lower_prog_alt_def)
    apply auto
    apply (metis (no_types, lifting) Rep_prog_refine abrh_def cond_mono order_refl pcond_prog.rep_eq)
   done

lemma inf_abr_prog_in_Lower_abr_prog:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> 
  (\<^bold>\<Sqinter>\<^sub>aA) \<in> Lower_abr (A \<inter> {P. abrh P = P})"
   unfolding greatest_abr_def inf_abr_def
   apply (subst Lower_abr_def)
      apply (subst Lower_abr_def)
  apply auto
  apply (rule someI2_ex)    
   apply (rule exI[where x = "abrh (\<Sqinter>\<^sub>pA)"]) 
   apply (simp add:prog_rep_eq image_def greatest_def Bex_def)
   apply auto
     apply (metis Healthy_def Rep_prog_H1_H3_closed ndes_hcond_def)
    apply (rule exI[where x = "abrh (\<Sqinter>\<^sub>pA)"])
    apply auto[]
    apply (metis (mono_tags, lifting) Collect_cong Rep_prog_eq inf_prog_abrh_in_Lower_prog_Int_abrh)
   apply (simp only:inf_prog_alt_def Bex_def)
   apply (rule someI2_ex)   
  using inf_prog_is_greatest_lower_prog apply blast
   apply (metis (no_types, lifting) Rep_prog_inject abrh_def cond_mono greatest_def 
                                    greatest_prog.rep_eq image_eqI order_refl 
                                    pcond_prog.rep_eq semilattice_inf_class.inf.orderE utp_order_le)
   apply (simp add:Lower_abr_def greatest_prog_alt_def)
  done

lemma inf_abr_prog_in_Lower_prog:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> 
  (\<^bold>\<Sqinter>\<^sub>aA) \<in> Lower_prog (A \<inter> {P. abrh P = P})"
  by (metis (mono_tags) Int_iff Lower_abr_def inf_abr_prog_in_Lower_abr_prog semilattice_inf_class.inf.right_idem)

lemma inf_abr_prog_greatest:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> abrh z = z \<Longrightarrow> (\<And>x. x \<in> A \<Longrightarrow> z \<sqsubseteq> x) \<Longrightarrow>  
   z \<sqsubseteq> (\<^bold>\<Sqinter>\<^sub>aA)"
  (*The KEY proof for the rest of the theory*)
  unfolding greatest_abr_def Lower_abr_def inf_abr_def
  apply auto
  apply (rule someI2_ex)    
   apply (rule exI[where x = "abrh (\<Sqinter>\<^sub>pA)"])    
   apply (simp add: prog_rep_eq image_def greatest_def Bex_def)
   apply auto
     apply (metis Healthy_def Rep_prog_H1_H3_closed ndes_hcond_def)
    apply (rule exI[where x = "abrh (\<Sqinter>\<^sub>pA)"])
    apply auto[]   
      apply (metis (mono_tags, lifting) Collect_cong Rep_prog_eq inf_prog_abrh_in_Lower_prog_Int_abrh)
  unfolding inf_prog_alt_def 
   defer
   apply (simp only: greatest_prog_alt_def Lower_prog_alt_def)
   apply auto
  apply (rule someI2_ex)
  using inf_prog_is_greatest_lower_prog 
   apply auto[1]
  apply (metis (no_types, lifting) Rep_prog_eq Rep_prog_refine abrh_def cond_mono 
                                   greatest_prog_alt_def order_refl pcond_prog.rep_eq 
                                   semilattice_inf_class.inf.orderE)
  done 
  
lemma inf_prog_abrh_is_greatest_lower:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> 
   greatest_prog (abrh (\<Sqinter>\<^sub>pA)) (Lower_prog (A \<inter> {P. abrh P = P}) \<inter> {P. abrh P = P})"
  unfolding inf_prog_alt_def 
  apply (rule someI2_ex)    
  using inf_prog_is_greatest_lower_prog 
   apply blast
  apply (simp only: greatest_prog_alt_def Lower_prog_alt_def)
  apply auto
   apply (metis (no_types, lifting) Rep_prog_refine abrh_def cond_mono order_refl pcond_prog.rep_eq)
  unfolding abrh_def
  apply (metis (mono_tags, lifting) if_prog_monoI mem_Collect_eq order_refl subsetCE)
  done 
     
lemma inf_abr_prog_lower:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow>  x \<in> A \<Longrightarrow>
    (\<^bold>\<Sqinter>\<^sub>aA) \<sqsubseteq> x" 
  unfolding greatest_abr_def Lower_abr_def inf_abr_def
  apply auto
  apply (rule someI2_ex)    
   apply (rule exI[where x = "abrh (\<Sqinter>\<^sub>pA)"])    
   apply (simp add: inf_prog_abrh_is_greatest_lower)
  apply (metis (no_types, lifting) Int_lower1 dual_order.trans greatest_prog_alt_def 
                                   inf_prog_is_greatest_lower_prog inf_prog_lower 
                                   semilattice_inf_class.inf.orderE subsetCE)
  done
 
lemma inf_abr_prog_lower2:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> u \<in> A \<Longrightarrow>  u \<sqsubseteq> v \<Longrightarrow> (\<^bold>\<Sqinter>\<^sub>aA) \<sqsubseteq> v"  
  using inf_abr_prog_lower [of A u ]
  by auto
     
lemma inf_abr_prog_is_abrh[closure]:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow>  
   abrh (\<^bold>\<Sqinter>\<^sub>aA) = (\<^bold>\<Sqinter>\<^sub>aA)"
  using Lower_abr_def inf_abr_prog_in_Lower_abr_prog
  by blast    
    
lemma inf_abr_prog_empty:
  "(\<^bold>\<Sqinter>\<^sub>a {}) = abrh (MAGIC)"
  unfolding inf_abr_def
  apply (rule someI2_ex)    
   apply (metis (no_types, lifting) Ball_Collect Lower_abr_def empty_iff greatest_abr_def inf_prog_abrh_is_greatest_lower semilattice_inf_class.inf_le2)
  apply (metis (no_types, lifting) Lower_abr_def greatest_abr_def greatest_prog_unique inf_prog_abrh_is_greatest_lower inf_prog_empty order_bot_class.bot.extremum)
  done 
    
lemma inf_abr_prog_is_greatest_abr_prog_Lower_abr_prog:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> greatest_abr (\<^bold>\<Sqinter>\<^sub>a A) (Lower_abr A)"
  unfolding greatest_abr_def 
  apply auto
   apply (simp add: Lower_abr_def)
  unfolding  greatest_prog_alt_def
  apply auto
   apply (metis (mono_tags, lifting) inf_abr_prog_in_Lower_abr_prog semilattice_inf_class.inf.orderE)
  apply (rule inf_abr_prog_greatest)
    apply auto   
  using Lower_abr_def apply auto[1]
  using lower_abr_prog_alt_def apply auto
  done
 
lemma inf_abr_prog_univ:
  "(\<^bold>\<Sqinter>\<^sub>a {P. abrh P = P}) = abrh (ABORT)"
  unfolding inf_abr_def
  apply (rule someI2_ex)    
   apply (rule exI[where x = "(\<^bold>\<Sqinter>\<^sub>a {P. abrh P = P})"])
   apply (rule inf_abr_prog_is_greatest_abr_prog_Lower_abr_prog)
   apply simp
  unfolding greatest_abr_prog_alt_def lower_abr_prog_alt_def
  apply auto
  apply (simp add: bottom_abr_prog_abrh_lower_prog dual_order.antisym)
  done
     
subsection \<open>Sup for exceptions\<close>
  
definition sup_abr :: "('s, 'e) abr_prog set \<Rightarrow> ('s, 'e) abr_prog" ("\<^bold>\<Squnion>\<^sub>a_" [900] 900)
where "sup_abr A = (SOME x. least_abr x (Upper_abr A))"  

subsubsection \<open>Algebraic laws\<close>
       
lemma sup_prog_abrh_in_upper_prog_Int_abrh:
  "abrh (\<Squnion>\<^sub>pA) \<in> Upper_prog (A \<inter> {P. abrh P = P})" 
  unfolding sup_prog_alt_def 
  apply (rule someI2_ex)
  using sup_prog_is_least_upper_prog
   apply auto[1]
  apply (simp only: least_prog_alt_def Upper_prog_alt_def)
  apply auto
  apply (metis (no_types, lifting) Rep_prog_refine abrh_def cond_mono order_refl pcond_prog.rep_eq)
  done
  
lemma sup_abr_prog_in_upper_abr_prog:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> 
  (\<^bold>\<Squnion>\<^sub>a A) \<in> Upper_abr (A \<inter> {P. abrh P = P})"
   unfolding least_abr_def sup_abr_def
   apply (subst Upper_abr_def)
      apply (subst Upper_abr_def)
  apply auto
  apply (rule someI2_ex)    
   apply (rule exI[where x = "abrh (\<Squnion>\<^sub>pA)"]) 
   apply (simp add:  prog_rep_eq image_def least_def Bex_def)
   apply auto
     apply (metis Healthy_def Rep_prog_H1_H3_closed ndes_hcond_def)
    apply (rule exI[where x = "abrh (\<Squnion>\<^sub>pA)"])
    apply auto[]
    apply (metis (mono_tags, lifting) Collect_cong Rep_prog_eq sup_prog_abrh_in_upper_prog_Int_abrh)
   apply (simp only: sup_prog_alt_def Bex_def)
   apply (rule someI2_ex)   
   using sup_prog_is_least_upper_prog
     apply blast
   apply (metis (no_types, lifting) Rep_prog_inject abrh_def cond_mono least_def 
                                    least_prog.rep_eq image_eqI order_refl 
                                    pcond_prog.rep_eq semilattice_inf_class.inf.orderE utp_order_le)
   apply (simp add: Upper_abr_def least_prog_alt_def)
  done

lemma sup_abr_prog_in_upper_prog:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> 
  \<^bold>\<Squnion>\<^sub>a A \<in> Upper_prog (A \<inter> {P. abrh P = P})"
  by (metis (mono_tags) Int_iff Upper_abr_def sup_abr_prog_in_upper_abr_prog semilattice_inf_class.inf.right_idem)

lemma sup_abr_prog_least:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> abrh z = z \<Longrightarrow> (\<And>x. x \<in> A \<Longrightarrow>  x \<sqsubseteq> z ) \<Longrightarrow>  
   \<^bold>\<Squnion>\<^sub>a A \<sqsubseteq> z"
  unfolding least_abr_def Upper_abr_def sup_abr_def
  apply auto
  apply (rule someI2_ex)    
   apply (rule exI[where x = "abrh (\<Squnion>\<^sub>pA)"])    
   apply (simp add:  prog_rep_eq image_def least_def Bex_def)
   apply auto
     apply (metis Healthy_def Rep_prog_H1_H3_closed ndes_hcond_def)
    apply (rule exI[where x = "abrh (\<Squnion>\<^sub>pA)"])
    apply auto[]   
      apply (metis (mono_tags, lifting) Collect_cong Rep_prog_eq sup_prog_abrh_in_upper_prog_Int_abrh)
  unfolding sup_prog_alt_def 
   defer
   apply (simp only: least_prog_alt_def Upper_prog_alt_def)
   apply auto
  apply (rule someI2_ex)
  using sup_prog_is_least_upper_prog
    apply auto[1]
  apply (metis (no_types, lifting) Rep_prog_eq Rep_prog_refine abrh_def cond_mono 
                                   least_prog_alt_def order_refl pcond_prog.rep_eq 
                                   semilattice_inf_class.inf.orderE)
  done 

lemma sup_prog_abrh_is_least_prog_upper_prog:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> 
   least_prog (abrh (\<Squnion>\<^sub>pA)) (Upper_prog (A \<inter> {P. abrh P = P}) \<inter> {P. abrh P = P})"
  unfolding sup_prog_alt_def 
  apply (rule someI2_ex)    
  using sup_prog_is_least_upper_prog
   apply blast
  apply (simp only: least_prog_alt_def Upper_prog_alt_def)
  apply auto
   apply (metis (no_types, lifting) Rep_prog_refine abrh_def cond_mono order_refl pcond_prog.rep_eq)
  unfolding abrh_def
  apply (metis (mono_tags, lifting) if_prog_monoI mem_Collect_eq order_refl subsetCE)
  done

lemma sup_abr_prog_upper:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> x \<in> A \<Longrightarrow>
    x \<sqsubseteq> \<^bold>\<Squnion>\<^sub>a A" 
  unfolding least_abr_def Upper_abr_def sup_abr_def
  apply auto
  apply (rule someI2_ex)    
   apply (rule exI[where x = "abrh (\<Squnion>\<^sub>pA)"])    
   apply (simp add: sup_prog_abrh_is_least_prog_upper_prog)
  apply (metis (no_types, lifting) Int_lower1 dual_order.trans least_prog_alt_def 
                                   sup_prog_is_least_upper_prog sup_prog_upper 
                                   semilattice_inf_class.inf.orderE subsetCE)
  done
 
lemma sup_abr_prog_upper2:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> l \<in> A \<Longrightarrow> v \<sqsubseteq> l \<Longrightarrow> v \<sqsubseteq> (\<^bold>\<Squnion>\<^sub>aA) "  
  using sup_abr_prog_upper [of A l] 
  by auto 


lemma sup_abr_prog_is_abrh[closure]:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow>  
   abrh (\<^bold>\<Squnion>\<^sub>a A) = \<^bold>\<Squnion>\<^sub>a A"
  using Upper_abr_def sup_abr_prog_in_upper_abr_prog
  by blast       
    
lemma sup_abr_prog_empty:
  "\<^bold>\<Squnion>\<^sub>a {} = abrh ABORT"
  unfolding sup_abr_def
  apply (rule someI2_ex)    
   apply (rule exI[where x = "abrh ABORT"])
   apply (metis (no_types, lifting) Upper_abr_def empty_subsetI least_abr_def semilattice_inf_class.inf_le2 sup_prog_abrh_is_least_prog_upper_prog sup_prog_empty)
  apply (metis (no_types, lifting) Upper_abr_def empty_subsetI least_abr_def least_prog_unique sup_prog_abrh_is_least_prog_upper_prog sup_prog_empty)
  done
    
lemma sup_abr_prog_univ:
  "\<^bold>\<Squnion>\<^sub>a {P. abrh P = P} = abrh MAGIC"
  unfolding sup_abr_def
  apply (rule someI2_ex)    
   apply (rule exI[where x = "abrh MAGIC"])
   apply(simp_all add: least_abr_prog_alt_def upper_abr_prog_alt_def)
   apply (metis UNIV_I abrh_def cond_mono dual_order.refl less_eq_prog.rep_eq  
                pcond_prog.rep_eq skip.rep_eq sup_prog_upper sup_prog_univ)
  unfolding least_abr_def Upper_abr_def
  apply auto
    apply (simp add: abrh_below_top_abr_prog dual_order.antisym)
  done

lemma sup_abr_prog_is_least_abr_prog_upper_abr:
  "A \<subseteq> {P. abrh P = P} \<Longrightarrow> least_abr (\<^bold>\<Squnion>\<^sub>a A) (Upper_abr A)"
  unfolding least_abr_def 
  apply auto
   apply (simp add: Upper_abr_def)
  unfolding least_prog_alt_def
  apply auto
    apply (metis (mono_tags) semilattice_inf_class.inf.orderE sup_abr_prog_in_upper_abr_prog)
    apply (rule sup_abr_prog_least)
    apply (auto simp: upper_abr_prog_alt_def)  
  done
    
subsection\<open>meet for exceptions theory\<close>

definition meet_abr_prog :: "('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog" (infixl "\<^bold>\<sqinter>\<^sub>a" 70)
  where  "meet_abr_prog a b = \<^bold>\<Sqinter>\<^sub>a{a, b}"

subsubsection\<open>algebraic laws\<close>

lemma le_meet_prog_abrI1:
  "abrh a = a \<Longrightarrow> abrh b = b \<Longrightarrow> a \<sqsubseteq> x \<Longrightarrow> a \<^bold>\<sqinter>\<^sub>a b \<sqsubseteq> x"
  apply (rule order_trans)
   apply auto
  unfolding meet_abr_prog_def
  apply (rule inf_abr_prog_lower)
   apply auto
  done

lemma le_meet_prog_abrI2:
  "abrh a = a \<Longrightarrow> abrh b = b \<Longrightarrow> b \<sqsubseteq> x \<Longrightarrow> a \<^bold>\<sqinter>\<^sub>a b \<sqsubseteq> x"
  apply (rule order_trans)
   apply auto
  unfolding meet_abr_prog_def
  apply (rule inf_abr_prog_lower)
   apply auto
  done

lemma le_meet_prog_abrI: 
  "abrh a = a \<Longrightarrow> abrh b = b \<Longrightarrow> a \<sqsubseteq> x \<Longrightarrow> b \<sqsubseteq> x \<Longrightarrow> a \<^bold>\<sqinter>\<^sub>a b \<sqsubseteq> x"
  apply (rule order_trans)
   apply auto
  unfolding meet_abr_prog_def
  apply (rule inf_abr_prog_lower)
   apply auto
  done 

lemma le_meet_prog_abr_is_abrh[closure]:
  "abrh a = a \<Longrightarrow> abrh b = b \<Longrightarrow> abrh (a \<^bold>\<sqinter>\<^sub>a b) = a \<^bold>\<sqinter>\<^sub>a b"
  unfolding meet_abr_prog_def
  by (simp add: inf_abr_prog_is_abrh)

subsection\<open>join for exception theory\<close>
  
definition join_abr_prog :: "('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog" (infixl "\<^bold>\<squnion>\<^sub>a" 70)
where  "join_abr_prog a b = \<^bold>\<Squnion>\<^sub>a{a, b}"

subsubsection\<open>algebraic laws\<close>
  
lemma le_join_prog_abrI1: 
  "abrh x = x \<Longrightarrow> abrh y = y \<Longrightarrow> s \<sqsubseteq> x \<Longrightarrow> s \<sqsubseteq> x \<^bold>\<squnion>\<^sub>a y"
  apply (rule order_trans)
   apply auto
  unfolding join_abr_prog_def
  apply (rule sup_abr_prog_upper)
   apply auto
  done

lemma le_join_prog_abrI2: 
 "abrh x = x \<Longrightarrow> abrh y = y \<Longrightarrow> s \<sqsubseteq> y \<Longrightarrow> s \<sqsubseteq> x \<^bold>\<squnion>\<^sub>a y"
  apply (rule order_trans)
   apply auto
  unfolding join_abr_prog_def
  apply (rule sup_abr_prog_upper)
   apply auto
  done

lemma le_join_prog_abrI: 
  "abrh x = x \<Longrightarrow> abrh y = y \<Longrightarrow> s \<sqsubseteq> x \<Longrightarrow> s \<sqsubseteq> y \<Longrightarrow>  s \<sqsubseteq> x \<^bold>\<squnion>\<^sub>a y"
 apply (rule order_trans)
   apply auto
  unfolding join_abr_prog_def
  apply (rule sup_abr_prog_upper)
   apply auto
  done
    
lemma le_join_prog_abr_is_abrh[closure]:
  "abrh x = x \<Longrightarrow> abrh y = y \<Longrightarrow> abrh (x \<^bold>\<squnion>\<^sub>a y) =  x \<^bold>\<squnion>\<^sub>a y"
  unfolding join_abr_prog_def
  by (simp add: sup_abr_prog_is_abrh)
     

section \<open>Ordering for programs\<close>

type_synonym '\<alpha> prog_health = "'\<alpha> prog \<Rightarrow> '\<alpha> prog"

definition prog_order :: "'\<alpha> prog_health \<Rightarrow> '\<alpha> prog gorder" where
"prog_order H = \<lparr> carrier = {P. P = H P}, eq = (op =), le = op \<sqsubseteq> \<rparr>"

subsection \<open>setup for generalisation purpose\<close>

typedecl ABRH

abbreviation "ABRH \<equiv> UTHY(ABRH,  ('s, 'e) abr)"

definition abrh_hcond :: "(ABRH, ('s, 'e) abr) uthy \<Rightarrow> (('s, 'e) abr) prog_health" where
  "abrh_hcond t = abrh"
  
definition abrh_unit :: "(ABRH, ('s, 'e) abr) uthy \<Rightarrow> ('s, 'e) abr_prog" where
  [upred_defs]: "abrh_unit t = abrh SKIP"  
  
subsubsection \<open>algebraic laws\<close>
  
  lemma prog_order_carrier [simp]:
  "carrier (prog_order H) = {P. H P =P}"
    by (auto simp add: prog_order_def)
      
lemma prog_order_eq [simp]:
  "eq (prog_order T) = op ="
  by (simp add: prog_order_def)

lemma prog_order_le [simp]:
  "le (prog_order T) = op \<sqsubseteq>"
  by (simp add: prog_order_def)

lemma prog_partial_order: 
  "partial_order (prog_order T)"
  by (unfold_locales, simp_all)

lemma prog_weak_partial_order: "weak_partial_order (prog_order T)"
  by (unfold_locales, simp_all)   
    
section \<open>Monotonicity for exception\<close>

abbreviation
  "mono_abr_prog F \<equiv> Mono\<^bsub>(prog_order (abrh_hcond ABRH))\<^esub> F"  

lemma mono_abr_prog_alt_def: 
  "Mono\<^bsub>(prog_order (abrh_hcond ABRH))\<^esub> F = 
   mono_prog (F o abrh) \<and> weak_partial_order (prog_order (abrh_hcond ABRH))"
  apply (simp only : isotone_def prog_rep_eq prog_order_def weak_partial_order_def abrh_hcond_def)
  apply auto
  using ndes_utp_theory.equivalence_axioms apply auto[1]
  using ndes_utp_theory.weak_partial_order_axioms weak_partial_order_def apply auto[1]
           apply (simp add: ndes_utp_theory.equivalence_axioms)
  using ndes_utp_theory.weak_partial_order_axioms weak_partial_order_def apply blast
         apply (metis (no_types, lifting) Abs_prog_Rep_prog_Ncarrier Healthy_if abrh_def abrh_idem cond_mono eq_iff pcond_prog.rep_eq)
        apply (rule equivalence.intro)
          prefer 3
          apply (subst (asm) equivalence.intro)
            prefer 5
            apply (subst (asm) equivalence.intro)
               prefer 7
               apply (subst (asm) weak_partial_order_axioms.intro)
                   prefer 9   
                   apply auto
      apply (subst weak_partial_order_axioms.intro)
          apply auto
      apply (subst (asm) equivalence.intro)
         apply auto
      apply (subst (asm) weak_partial_order_axioms.intro)
          apply auto
      defer
      apply (subst equivalence.intro)
         apply auto
     apply (subst (asm) equivalence.intro)
        apply auto
     apply (subst (asm) weak_partial_order_axioms.intro)
         apply auto
     defer 
     apply (subst equivalence.intro)
        apply auto
    apply (subst weak_partial_order_axioms.intro)
        apply auto
    apply (simp add: Rep_prog_eq)
   apply (simp add: Rep_prog_inject)
  apply (metis Rep_prog_H1_H3_closed Rep_prog_inverse is_Ncarrier_is_ndesigns)
  done

subsubsection \<open>algebraic laws\<close>
  
lemma mono_Monotone_prog_order:
  "mono F \<Longrightarrow> Mono\<^bsub>(prog_order (abrh_hcond ABRH))\<^esub> F" 
  apply (auto simp add: isotone_def)
   apply (simp add: mono_abr_prog_alt_def)
  apply (simp add: mono_def prog_order_def)
  done   
    

lemma weak_order_prog_monoI:
  shows "(\<And>x y. H x = x  \<Longrightarrow>  H y = y  \<Longrightarrow> x \<sqsubseteq> y \<Longrightarrow> F x \<sqsubseteq> F y) \<Longrightarrow> Mono\<^bsub>prog_order H\<^esub> F"    
  unfolding isotone_def prog_order_def using prog_weak_partial_order 
  by (auto intro!: equivalence.intro weak_partial_order_axioms.intro weak_partial_order.intro)

lemma weak_order_prog_monoD:
  assumes "Mono\<^bsub>prog_order H\<^esub> F"
    and    "H x = x"
    and    "H y = y"
    and    "y \<sqsubseteq> x"    
  shows   "(F y) \<sqsubseteq> (F x)"
  using assms prog_weak_partial_order unfolding isotone_def prog_order_def 
  by (auto intro!: equivalence.intro weak_partial_order_axioms.intro weak_partial_order.intro)

lemma weak_order_prog_monoE:
  "Mono\<^bsub>prog_order H\<^esub> F \<Longrightarrow> H x = x \<Longrightarrow> H y = y\<Longrightarrow> x \<sqsubseteq> y \<Longrightarrow> 
   (H x = x \<Longrightarrow> H y = y \<Longrightarrow> F x \<sqsubseteq> F y \<Longrightarrow> thesis) \<Longrightarrow> thesis"
  by (simp add: weak_order_prog_monoD)

section  \<open>Fixed points for exceptions\<close>
  
subsection \<open>Least fixed point\<close>

definition LFP_abr :: " (('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog) \<Rightarrow> ('s, 'e) abr_prog" 
  where "LFP_abr F = \<^bold>\<Sqinter>\<^sub>a {x \<in> {P. abrh P = P} . F x \<sqsubseteq> x}" 
syntax
  "_amu" :: "pttrn \<Rightarrow> logic \<Rightarrow> logic" ("\<^bold>\<mu>\<^sub>a _ \<bullet> _" [0, 10] 10)
  
translations
  "\<^bold>\<mu>\<^sub>a X \<bullet> P" == "CONST LFP_abr (\<lambda> X. P)" 
  
notation LFP_abr ("\<^bold>\<mu>\<^sub>a")

subsubsection \<open>Algebraic laws\<close>
  
lemma lfp_abr_healthy_comp:
  "\<^bold>\<mu>\<^sub>a F = \<^bold>\<mu>\<^sub>a (F \<circ> abrh)" 
proof -
  have "{P. (abrh P = P) \<and> F P \<sqsubseteq> P} = {P. (abrh P = P) \<and> F (abrh P) \<sqsubseteq> P}"
    by auto
  thus ?thesis
    unfolding LFP_abr_def  
    by simp
qed
  
lemma lfp_abr_prog_lowerbound:
  "abrh x = x \<Longrightarrow> F x \<sqsubseteq> x \<Longrightarrow> \<^bold>\<mu>\<^sub>a F \<sqsubseteq> x"
  by (simp add: Collect_mono_iff LFP_abr_def inf_abr_prog_lower)  
    
lemma lfp_abr_prog_greatest:
  "abrh x = x \<Longrightarrow> (\<And>z. F z \<sqsubseteq> z \<Longrightarrow> x \<sqsubseteq> z) \<Longrightarrow> x \<sqsubseteq> \<^bold>\<mu>\<^sub>a F"   
  by (metis (mono_tags, lifting) Collect_mono LFP_abr_def inf_abr_prog_greatest mem_Collect_eq)

lemma lfp_abr_fixed_point:
  assumes M:"mono_abr_prog F"
  assumes H: "F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P}"
  shows " F (\<^bold>\<mu>\<^sub>a F) = \<^bold>\<mu>\<^sub>a F"    
proof (rule antisym)
  have ne: "{P. (abrh P = P) \<and> F P \<sqsubseteq> P} \<noteq> {}"
  proof -
    have "F (abrh MAGIC) \<sqsubseteq> (abrh MAGIC)"
      using H abrh_below_top_abr_prog by force
    thus ?thesis       
      by (auto, rule_tac x="(abrh MAGIC)" in exI, auto)
  qed
  show "\<^bold>\<mu>\<^sub>a F \<sqsubseteq> F (\<^bold>\<mu>\<^sub>a F)"
  proof -
    have "\<^bold>\<Sqinter>\<^sub>a{P. (abrh P = P) \<and> F(P) \<sqsubseteq> P} \<sqsubseteq> \<^bold>\<Sqinter>\<^sub>a{P. (abrh P = P) \<and> F (abrh(P)) \<sqsubseteq> P}"
    proof -
      have 1: "\<And> P.  F (abrh(P)) = abrh(F (abrh(P)))" 
      proof -
        fix P :: "('b, 'a) abr_prog"
        have "abrh P \<in> {p. abrh p = p}"
          by (metis abr_abrh_closed)
        then have "F (abrh P) \<in> {p. abrh p = p}"
          using H by blast
        then show "F (abrh P) = abrh (F (abrh P))"
          by auto
      qed
      show ?thesis
        apply (rule inf_abr_prog_greatest)
          apply auto[]
         defer
         apply (rule inf_abr_prog_lower2[of _ "F (abrh _)"])
           apply simp_all
           apply auto 
          apply (metis "1")
         apply (metis "1" abrh_hcond_def assms(1) weak_order_prog_monoD)
        apply  (simp add: Collect_mono inf_abr_prog_is_abrh)
        done
    qed
    have 1: "\<And> P.  F (abrh(P)) = abrh(F (abrh(P)))" 
    proof -
      fix P :: "('b, 'a) abr_prog"
      have "abrh P \<in> {p. abrh p = p}"
        by (metis abr_abrh_closed)
      then have "F (abrh P) \<in> {p. abrh p = p}"
        using H by blast
      then show "F (abrh P) = abrh (F (abrh P))"
        by auto
    qed
    with ne show ?thesis
      apply (simp add: LFP_abr_def)
      apply (rule inf_abr_prog_lower)
       apply auto[]
      apply (rule CollectI)
      apply auto        
       apply (metis (no_types, lifting) Collect_mono_iff inf_abr_prog_is_abrh)        
      apply (rule weak_order_prog_monoD[where F= F and 
            y ="(F (\<^bold>\<Sqinter>\<^sub>a{x. abrh x = x \<and> F x \<sqsubseteq> x}))" and
            x = "(\<^bold>\<Sqinter>\<^sub>a{x. abrh x = x \<and> F x \<sqsubseteq> x})"])
      using M
         apply simp        
        apply (simp add: Collect_mono abrh_hcond_def inf_abr_prog_is_abrh)
       apply (metis (no_types, lifting) Collect_mono abrh_hcond_def inf_abr_prog_is_abrh)
      apply (rule inf_abr_prog_greatest)
        apply auto     
       apply (metis (no_types, lifting) Collect_mono_iff inf_abr_prog_is_abrh)
      apply (simp add: inf_abr_def)
      apply (rule someI2_ex)            
       apply (metis (no_types, lifting) Collect_mono_iff inf_abr_prog_is_greatest_abr_prog_Lower_abr_prog)
      unfolding greatest_abr_prog_alt_def lower_abr_prog_alt_def Ball_def abrh_def
      apply auto
      apply (rule order_trans[OF weak_order_prog_monoD])
          apply auto
       apply (simp add: weak_order_prog_monoI)
      apply (metis abrh_def abrh_hcond_def assms(1) weak_order_prog_monoE)
      done 
  qed 
  from ne show " F (\<^bold>\<mu>\<^sub>a F) \<sqsubseteq> \<^bold>\<mu>\<^sub>a F"
    apply (insert H M)
    apply (simp add: LFP_abr_def)
    apply (rule inf_abr_prog_greatest)
      apply auto   
     apply (erule PiE)
      apply auto[]
     apply (simp add: Collect_mono inf_abr_prog_is_abrh)
    apply (simp add: inf_abr_def)
    apply (rule someI2_ex)
     apply (metis (no_types, lifting) Collect_mono_iff inf_abr_prog_is_greatest_abr_prog_Lower_abr_prog)
    unfolding greatest_abr_prog_alt_def lower_abr_prog_alt_def Ball_def abrh_def
    apply auto
    apply (rule order_trans[OF weak_order_prog_monoD])
        apply auto
     apply (simp add: weak_order_prog_monoI)        
    apply (simp add: abrh_def abrh_hcond_def weak_order_prog_monoD)
    done 
qed
  
lemma lfp_abr_prog_unfold: 
  assumes M:"mono_abr_prog F"
  assumes H: "F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P}"  
  shows "\<^bold>\<mu>\<^sub>a F = F (\<^bold>\<mu>\<^sub>a F)"  
  by (simp add: H M lfp_abr_fixed_point) 
    
lemma lfp_abr_prog_const:
  "abrh t = t \<Longrightarrow> \<^bold>\<mu>\<^sub>a (\<lambda> x . t) = t"
  apply (rule lfp_abr_prog_unfold)
   apply (rule weak_order_prog_monoI)
   apply simp_all
  done

lemma lfp_abr_prog_is_abrh[closure]:
   "abrh (\<^bold>\<mu>\<^sub>a F) = \<^bold>\<mu>\<^sub>a F"
  by (metis (mono_tags, lifting) CollectD Collect_mono LFP_abr_def inf_abr_prog_is_abrh)

lemma lfp_abr_prog_eqI:
  "mono_abr_prog F \<Longrightarrow> F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P} \<Longrightarrow> abrh x = x \<Longrightarrow> F x = x \<Longrightarrow>
   (\<And>z. F z = z \<Longrightarrow> x \<sqsubseteq> z ) \<Longrightarrow> \<^bold>\<mu>\<^sub>a F = x"
  by (rule antisym) (simp_all add: lfp_abr_prog_lowerbound lfp_abr_prog_unfold[symmetric])

lemma lfp_abr_prog_ordinal_induct [case_names M H step union]:
  assumes M:"mono_abr_prog F"
  assumes H: "F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P}"
  assumes P_f:   "\<And>S. P S \<Longrightarrow> S \<sqsubseteq> \<^bold>\<mu>\<^sub>a F \<Longrightarrow> abrh S =S \<Longrightarrow> P (F S)"
  assumes P_Union: "\<And>M. M \<subseteq> {P. abrh P = P} \<Longrightarrow> (\<And>S. S \<in> M \<Longrightarrow>  P S) \<Longrightarrow> P (sup_abr M)"
  shows "P (\<^bold>\<mu>\<^sub>a F)"
proof -
  let ?M = "{S. S \<sqsubseteq> \<^bold>\<mu>\<^sub>a F \<and> P S \<and> abrh S = S}"
  from P_Union have "P (sup_abr ?M)" by (smt Collect_mono_iff mem_Collect_eq)
  also have "sup_abr ?M = \<^bold>\<mu>\<^sub>a F"
  proof (rule antisym)
    show "sup_abr ?M \<sqsubseteq> \<^bold>\<mu>\<^sub>a F"
      apply (subst sup_abr_prog_least)
         apply (auto simp add: Collect_mono lfp_abr_prog_is_abrh)
      done
    then have "F(sup_abr ?M)  \<sqsubseteq> F (\<^bold>\<mu>\<^sub>a F)"
      by (metis (no_types, lifting) Collect_mono M abrh_hcond_def lfp_abr_prog_is_abrh sup_abr_prog_is_abrh weak_order_prog_monoE)
    then have "F (sup_abr ?M) \<sqsubseteq> \<^bold>\<mu>\<^sub>a F"
      by (simp add: H M lfp_abr_fixed_point)
    then have "F (sup_abr ?M) \<in> ?M"
      using P_Union
      apply simp
      apply (subst P_f)
         apply simp_all
         apply (simp add: calculation)
         apply (simp add: \<open>\<^bold>\<Squnion>\<^sub>a{S. S \<sqsubseteq> \<^bold>\<mu>\<^sub>a F \<and> P S \<and> abrh S = S} \<sqsubseteq> \<^bold>\<mu>\<^sub>a F\<close>)
       apply (simp add: Collect_mono_iff sup_abr_prog_is_abrh)
      using H
      apply (elim PiE)
       apply simp_all
      apply (simp add: Collect_mono sup_abr_prog_is_abrh)
      done
    then have "F (sup_abr ?M)  \<sqsubseteq> sup_abr ?M"
      by (simp add: Collect_mono sup_abr_prog_upper)
    then show "\<^bold>\<mu>\<^sub>a F  \<sqsubseteq> sup_abr ?M"
      by (simp add: Collect_mono lfp_abr_prog_lowerbound sup_abr_prog_is_abrh)
  qed
  finally show ?thesis .
qed

theorem lfp_abr_prog_induct:
  assumes M:"mono_abr_prog F"
  assumes H: "F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P}"
  assumes ind: "F (join_abr_prog (\<^bold>\<mu>\<^sub>a F) P) \<sqsubseteq> P"
  assumes HP: "abrh P = P"
  shows "\<^bold>\<mu>\<^sub>a F \<sqsubseteq> P"
proof (induct rule: lfp_abr_prog_ordinal_induct)
  case M
  then show ?case by fact
next
  case H
  then show ?case by fact
next
  case (step S)
  then show ?case
    apply (intro order_trans[OF ind _] weak_order_prog_monoD[OF M])
      apply (auto simp: abrh_hcond_def)
       apply (simp add: HP lfp_abr_prog_is_abrh sup_abr_prog_is_abrh join_abr_prog_def)
    apply (simp add: HP lfp_abr_prog_is_abrh le_join_prog_abrI)
    done
next
  case (union M)
  then show ?case  by (auto intro!: sup_abr_prog_least HP)
qed

subsection \<open>greatest fixed point\<close>

definition GFP_abr :: " (('s, 'e) abr_prog \<Rightarrow> ('s, 'e) abr_prog) \<Rightarrow> ('s, 'e) abr_prog"
  where "GFP_abr F = \<^bold>\<Squnion>\<^sub>a {x \<in> {P. abrh P = P} . x \<sqsubseteq> F x}"

syntax
  "_anu" :: "pttrn \<Rightarrow> logic \<Rightarrow> logic" ("\<^bold>\<nu>\<^sub>a _ \<bullet> _" [0, 10] 10)

notation GFP_abr ("\<^bold>\<nu>\<^sub>a")
  
translations
  "\<^bold>\<nu>\<^sub>a X \<bullet> P" == "CONST GFP_abr (\<lambda> X. P)"
  
lemma gfp_abr_healthy_comp:
  "\<^bold>\<nu>\<^sub>a F = \<^bold>\<nu>\<^sub>a (F \<circ> abrh)"
proof -
  have "{P. (abrh P = P) \<and> P \<sqsubseteq> F P} = {P. (abrh P = P) \<and> P \<sqsubseteq> F (abrh P)}"
    by auto
  thus ?thesis
    unfolding GFP_abr_def  
    by simp
qed
     
lemma gfp_abr_prog_upperbound:
  "abrh x = x \<Longrightarrow> x \<sqsubseteq> F x  \<Longrightarrow> x \<sqsubseteq> \<^bold>\<nu>\<^sub>a F "
  by (simp add: Collect_mono_iff GFP_abr_def sup_abr_prog_upper)
    
lemma gfp_abr_prog_least:
  "abrh x = x \<Longrightarrow> (\<And>z. z \<sqsubseteq> F z \<Longrightarrow> z \<sqsubseteq>  x) \<Longrightarrow> \<^bold>\<nu>\<^sub>a F  \<sqsubseteq> x"   
  by (metis (mono_tags, lifting) Collect_mono GFP_abr_def sup_abr_prog_least mem_Collect_eq)

lemma gfp_abr_fixed_point:
  assumes M:"mono_abr_prog F"
  assumes H: "F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P}"
  shows " F (\<^bold>\<nu>\<^sub>a F) = \<^bold>\<nu>\<^sub>a F"
proof (rule antisym)
  have ne: "{P. (abrh P = P) \<and> P \<sqsubseteq> F P} \<noteq> {}"
  proof -
    have "(abrh ABORT) \<sqsubseteq>  F (abrh ABORT) "
      using H bottom_abr_prog_abrh_lower_prog by force
    thus ?thesis
      by (auto, rule_tac x="(abrh ABORT)" in exI, auto)
  qed
  show "F (\<^bold>\<nu>\<^sub>a F)  \<sqsubseteq> \<^bold>\<nu>\<^sub>a F"
  proof -
    have "\<^bold>\<Squnion>\<^sub>a{P. (abrh P = P) \<and> P \<sqsubseteq> F(P)} \<sqsubseteq> \<^bold>\<Squnion>\<^sub>a{P. (abrh P = P) \<and> P  \<sqsubseteq> F (abrh(P)) }"
    proof -
      have 1: "\<And> P.  F (abrh(P)) = abrh(F (abrh(P)))"
      proof -
        fix P :: "('b,'a) abr_prog"
        have "abrh P \<in> {p. abrh p = p}"
          by (metis abr_abrh_closed)
        then have "F (abrh P) \<in> {p. abrh p = p}"
          using H by blast
        then show "F (abrh P) = abrh (F (abrh P))"
          by auto
      qed
      show ?thesis
        apply (rule sup_abr_prog_least)
          apply auto[]
         defer
         apply (rule sup_abr_prog_upper2)
           apply simp_all
           apply auto
            apply (simp add: Collect_mono sup_abr_prog_is_abrh)
        done
    qed
    have 1: "\<And> P.  F (abrh(P)) = abrh(F (abrh(P)))"
    proof -
      fix P :: "('b,'a) abr_prog"
      have "abrh P \<in> {p. abrh p = p}"
        by (metis abr_abrh_closed)
      then have "F (abrh P) \<in> {p. abrh p = p}"
        using H by blast
      then show "F (abrh P) = abrh (F (abrh P))"
        by auto
    qed
    with ne show ?thesis
      apply (simp add: GFP_abr_def)
      apply (rule sup_abr_prog_upper)
       apply auto[]
      apply (rule CollectI)
      apply auto
       apply (metis (no_types, lifting) Collect_mono_iff sup_abr_prog_is_abrh)
      apply (rule weak_order_prog_monoD[where F= F and
            x ="(F (\<^bold>\<Squnion>\<^sub>a{x. abrh x = x \<and> x \<sqsubseteq> F x}))" and
            y = "(\<^bold>\<Squnion>\<^sub>a{x. abrh x = x \<and> x  \<sqsubseteq> F x})"])
      using M
         apply simp
        apply (simp add: Collect_mono abrh_hcond_def inf_abr_prog_is_abrh)
        apply (metis (no_types, lifting) Collect_mono sup_abr_prog_is_abrh)
        apply (simp add: Collect_mono abrh_hcond_def sup_abr_prog_is_abrh)
      apply (rule sup_abr_prog_least)
        apply auto
       apply (metis (no_types, lifting) Collect_mono_iff sup_abr_prog_is_abrh)
      apply (simp add: sup_abr_def)
      apply (rule someI2_ex)
       apply (metis (no_types, lifting) Collect_mono_iff sup_abr_prog_is_least_abr_prog_upper_abr)
      unfolding least_abr_prog_alt_def upper_abr_prog_alt_def Ball_def abrh_def
      apply (rule order_trans[OF weak_order_prog_monoD])
          apply auto
       apply (simp add: weak_order_prog_monoI)
      apply (metis abrh_def abrh_hcond_def assms(1) weak_order_prog_monoE)
      done
  qed
  from ne show " \<^bold>\<nu>\<^sub>a F \<sqsubseteq> F (\<^bold>\<nu>\<^sub>a F)"
    apply (insert H M)
    apply (simp add: GFP_abr_def)
    apply (rule sup_abr_prog_least)
      apply auto
     apply (erule PiE)
      apply auto[]
          apply (simp add: Collect_mono sup_abr_prog_is_abrh)
    apply (simp add: sup_abr_def)
    apply (rule someI2_ex)   
      apply (metis (no_types, lifting) Collect_mono_iff sup_abr_prog_is_least_abr_prog_upper_abr)
    unfolding least_abr_prog_alt_def upper_abr_prog_alt_def Ball_def abrh_def 
    apply (rule order_trans[OF weak_order_prog_monoD])
        apply auto
     apply (simp add: weak_order_prog_monoI )  
    apply (simp add: abrh_def abrh_hcond_def weak_order_prog_monoD)      
    done 
qed

lemma gfp_abr_prog_unfold:
  assumes M:"mono_abr_prog F"
  assumes H: "F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P}"
  shows "\<^bold>\<nu>\<^sub>a F = F (\<^bold>\<nu>\<^sub>a F)"
  by (simp add: H M gfp_abr_fixed_point)

lemma gfp_abr_prog_const:
  "abrh t = t \<Longrightarrow> \<^bold>\<nu>\<^sub>a (\<lambda> x . t) = t"
  apply (rule gfp_abr_prog_unfold)
   apply (rule weak_order_prog_monoI)
   apply simp_all
  done

lemma gfp_abr_prog_eqI:
  "mono_abr_prog F \<Longrightarrow> F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P} \<Longrightarrow> abrh x = x \<Longrightarrow> F x = x \<Longrightarrow>
   (\<And>z. F z = z \<Longrightarrow> z \<sqsubseteq> x) \<Longrightarrow> \<^bold>\<nu>\<^sub>a F = x"  
  by (rule antisym) (simp_all add: gfp_abr_prog_upperbound gfp_abr_prog_unfold[symmetric])
    
lemma gfp_abr_prog_is_abrh[closure]:
   "abrh (\<^bold>\<nu>\<^sub>a F) = \<^bold>\<nu>\<^sub>a F"
   by (metis (mono_tags, lifting) CollectD Collect_mono GFP_abr_def sup_abr_prog_is_abrh)

lemma gfp_abr_prog_ordinal_induct [case_names M H step union]:
  assumes M:"mono_abr_prog F"
  assumes H: "F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P}"
  assumes P_f:   "\<And>S. P S \<Longrightarrow> \<^bold>\<nu>\<^sub>a F \<sqsubseteq> S \<Longrightarrow> abrh S =S \<Longrightarrow> P (F S)"
  assumes P_Union: "\<And>M. M \<subseteq> {P. abrh P = P} \<Longrightarrow> (\<And>S. S \<in> M \<Longrightarrow>  P S) \<Longrightarrow> P (inf_abr M)"
  shows "P (\<^bold>\<nu>\<^sub>a F)"
proof -
  let ?M = "{S. \<^bold>\<nu>\<^sub>a F \<sqsubseteq> S \<and> P S \<and> abrh S = S}"
  from P_Union have "P (inf_abr ?M)" by (smt Collect_mono_iff mem_Collect_eq)
  also have "inf_abr ?M = \<^bold>\<nu>\<^sub>a F"
  proof (rule antisym)
    show "\<^bold>\<nu>\<^sub>a F \<sqsubseteq> inf_abr ?M"
      apply (subst inf_abr_prog_greatest)
         apply (auto simp add: Collect_mono gfp_abr_prog_is_abrh)
      done
    then have "F (\<^bold>\<nu>\<^sub>a F) \<sqsubseteq> F(inf_abr ?M)"
      by (metis (no_types, lifting) Collect_mono M abrh_hcond_def gfp_abr_prog_is_abrh inf_abr_prog_is_abrh weak_order_prog_monoE)
    then have "\<^bold>\<nu>\<^sub>a F \<sqsubseteq> F (inf_abr ?M)"
      by (simp add: H M gfp_abr_fixed_point)
    then have "F (inf_abr ?M) \<in> ?M"
      using P_Union
      apply simp
      apply (subst P_f)
         apply simp_all
         apply (simp add: calculation)
        apply (simp add: \<open>\<^bold>\<nu>\<^sub>a F \<sqsubseteq> \<^bold>\<Sqinter>\<^sub>a{S. \<^bold>\<nu>\<^sub>a F \<sqsubseteq> S \<and> P S \<and> abrh S = S}\<close>)
       apply (simp add: Collect_mono_iff inf_abr_prog_is_abrh)
      using H
      apply (elim PiE)
       apply simp_all
      apply (simp add: Collect_mono inf_abr_prog_is_abrh)
      done
    then have "inf_abr ?M \<sqsubseteq> F (inf_abr ?M)"
      by (simp add: Collect_mono inf_abr_prog_lower)
    then show "inf_abr ?M \<sqsubseteq> \<^bold>\<nu>\<^sub>a F"
      by (simp add: Collect_mono gfp_abr_prog_upperbound inf_abr_prog_is_abrh)
  qed
  finally show ?thesis .
qed

theorem gfp_abr_prog_induct:
  assumes M:"mono_abr_prog F"
  assumes H: "F \<in> {P. abrh P = P} \<rightarrow> {P. abrh P = P}"
  assumes ind: "P \<sqsubseteq> F (meet_abr_prog (\<^bold>\<nu>\<^sub>a F) P)"
  assumes HP: "abrh P = P"
  shows "P \<sqsubseteq> \<^bold>\<nu>\<^sub>a F"
proof (induct rule: gfp_abr_prog_ordinal_induct)
  case M
  then show ?case by fact
next
  case H
  then show ?case by fact
next
  case (step S)
  then show ?case
    apply (intro order_trans[OF _ ind] weak_order_prog_monoD[OF M])
      apply (auto simp: abrh_hcond_def)
       apply (simp add: HP gfp_abr_prog_is_abrh inf_abr_prog_is_abrh meet_abr_prog_def)
    apply (simp add: HP gfp_abr_prog_is_abrh le_meet_prog_abrI)
    done
next
  case (union M)
  then show ?case  by (auto intro!: inf_abr_prog_greatest HP)
qed

section \<open>generalisation to any healthiness condition\<close>  

definition Idempotent :: "'\<alpha> prog_health \<Rightarrow> bool" where
  "Idempotent(H) \<longleftrightarrow> (\<forall> P. H(H(P)) = H(P))"

abbreviation Monotonic :: "'\<alpha> prog_health \<Rightarrow> bool" where
  "Monotonic(H) \<equiv> mono H"
  
consts
  prog_hcond :: "('\<T>, '\<alpha>) uthy \<Rightarrow> '\<alpha> prog_health" ("\<H>\<P>\<index>") 
  
locale prog_theory =
  fixes \<T> :: "('\<T>, '\<alpha>) uthy" (structure)
  assumes PHCond_Idem: " \<H>\<P>( \<H>\<P>(P)) =  \<H>\<P>(P)"
begin

  lemma uthy_simp:
    "uthy = \<T>"
    by blast

  text {* A UTP theory fixes @{term "\<T>"}, the structural element denoting the UTP theory. All
    constants associated with UTP theories can then be resolved by the type system. *}

  lemma HCond_Idempotent [closure,intro]: "Idempotent  \<H>\<P>"
    by (simp add: Idempotent_def PHCond_Idem)
  sublocale partial_order "prog_order (prog_hcond \<T>)"
    by (unfold_locales, simp_all add: prog_order_def)
end

(*interpretation ndes_utp_theory: prog_theory ABRH
  apply (simp add:   abrh_hcond_def utp_theory.intro)


interpretation ndes_unital: utp_theory_unital NDES
  apply (unfold_locales, simp_all add: ndes_hcond_def ndes_unit_def)
  using seq_r_H1_H3_closed apply blast
  apply (metis H1_rdesign H3_def Healthy_def' design_skip_idem skip_d_def)
  apply (metis H1_idem H1_left_unit Healthy_def')
  apply (metis H1_H3_commute H3_def H3_idem Healthy_def')
done



interpretation normal_design_theory_continuous: utp_theory_continuous NDES
  rewrites "\<And> P. P \<in> carrier (uthy_order NDES) \<longleftrightarrow> P is \<^bold>N"
  and "carrier (uthy_order NDES) \<rightarrow> carrier (uthy_order NDES) \<equiv> \<lbrakk>\<^bold>N\<rbrakk>\<^sub>H \<rightarrow> \<lbrakk>\<^bold>N\<rbrakk>\<^sub>H"
  and "le (uthy_order NDES) = op \<sqsubseteq>"
  and "A \<subseteq> carrier (uthy_order NDES) \<longleftrightarrow> A \<subseteq> \<lbrakk>\<^bold>N\<rbrakk>\<^sub>H"  
  and "eq (uthy_order NDES) = op ="  
  by (unfold_locales, simp_all add: ndes_hcond_def H1_H3_Continuous utp_order_def)*)

  
     
(*section {*algebraic laws*} 

lemma [algebraic_laws_prog]:"IF b THEN P ELSE P FI = P"
  by (simp add: prog_rep_eq Algebraic_Laws.cond_idem)

lemma [algebraic_laws_prog]:"SKIP ;  P =  P" "P ; SKIP = P"
  unfolding  abrh_def
  by (simp_all)  

lemma [algebraic_laws_prog]:"throw_abr ;  (abrh P) = throw_abr"     
  unfolding throw_abr_def abrh_def assigns_abr_def 
  apply (simp add: prog_rep_eq) 
   apply rel_auto
  done

lemma "abrh(\<langle>a\<rangle>\<^sub>p) ; abrh (throw_abr ;  IF &abrupt THEN (passigns (subst_upd id abrupt (\<not> &abrupt)) ; abrh(\<langle>b\<rangle>\<^sub>p)) ELSE SKIP FI) = 
         (abrh(\<langle>a\<rangle>\<^sub>p); abrh(\<langle>b\<rangle>\<^sub>p))"
  unfolding throw_abr_def abrh_def assigns_abr_def 
    apply (simp add: prog_rep_eq) 
  apply rel_auto
  apply (smt abr_vars.ext_inject abr_vars_ext_def)+
  done  
 
lemma "abrh(P) ; abrh (throw_abr ;  if_abr  (&abrupt) (passigns (subst_upd id abrupt (\<not> &abrupt)) ; abrh(Q))  (SKIP) = 
         (abrh(P); abrh(Q))"
  unfolding throw_abr_def abrh_def assigns_abr_def 
    apply (simp add: prog_rep_eq) 
  apply rel_auto
  apply transfer
  unfolding Healthy_def H1_def H3_def
    apply rel_simp
         apply blast
      apply transfer
    unfolding Healthy_def H1_def H3_def
    apply rel_simp
          apply blast
      apply transfer
    unfolding Healthy_def H1_def H3_def
    apply rel_simp
         apply blast
      apply transfer
    unfolding Healthy_def H1_def H3_def
    apply rel_simp
    apply metis
   apply transfer
    unfolding Healthy_def H1_def H3_def
    apply rel_simp
       apply metis
      apply transfer
    unfolding Healthy_def H1_def H3_def
    apply rel_simp
      apply metis
      apply transfer
    unfolding Healthy_def H1_def H3_def
    apply rel_simp
     apply metis
    apply (smt abr_vars.select_convs(2))
done *)

end
