(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)
(******************************************************************************
* Authors: Nyle Rogers, Yakoub Nemouchi
 ******************************************************************************)
theory utp_heap
imports     
          "../../tools/autocorres-1.3/autocorres/AutoCorres" (*Needs the import autocoress heap model*)
          "../../impl/utp_prog_abrupt_lift"
begin
recall_syntax \<comment> \<open>Fixes notation issue with inclusion of HOL libraries on a top of UTP.\<close>


section \<open>UTP heap alphabet\<close>
text 
\<open>In order to capture arrays, we extend the alphabet of UTP by an additional
  global state variables:
   \begin{itemize}   
      \<^item> heap_raw: A pair of two functions. The first element of the pair 
        is a total mapping from addresses to bytes. The second element of the pair
        is \emph{heap type}. Namely, it is a pair of where the first element of the pair
        is a predicate on addresses. The second element of the pair is ...  
   \end{itemize}\<close>

alphabet hp_vars = 
  heap_raw:: "32 word \<Rightarrow> 32 word ptr" (*TODO: continue from here!*)

declare hp_vars.splits [alpha_splits]

subsection \<open>Alphabet proofs\<close>
text \<open>
  The two locale interpretations below are a technicality to improve automatic
  proof support via the predicate and relational tactics. This is to enable the
  (re-)interpretation of state spaces to remove any occurrences of lens types
  after the proof tactics @{method pred_simp} and @{method rel_simp}, or any
  of their derivatives have been applied. Eventually, it would be desirable to
  automate both interpretations as part of a custom outer command for defining
  alphabets.\<close>

interpretation cp_hp:
  lens_interp "\<lambda> (ok, r) . (ok, heap_raw\<^sub>v r, more r)"
apply (unfold_locales)
apply (rule injI)
apply (clarsimp)
  done
    
interpretation cp_hp_rel: lens_interp "\<lambda>(ok, ok', r, r').
  (ok, ok', heap_raw\<^sub>v r, heap_raw\<^sub>v r', more r, more r')"
apply (unfold_locales)
apply (rule injI)
apply (clarsimp)
done

subsection \<open>Type lifting\<close>

type_synonym '\<alpha> hp_aprog  = "'\<alpha> hp_vars_scheme  aprog"
type_synonym '\<alpha> hp  = "'\<alpha> hp_vars_scheme"
  
translations
  (type) "'\<alpha> hp" <= (type) "'\<alpha> hp_vars_scheme"
  (type) "'\<alpha> hp" <= (type) "'\<alpha> hp_vars_ext"
  (type) "'\<alpha> hp_aprog" <= (type) "'\<alpha> hp_vars_scheme  aprog"
  (type) "'\<alpha> hp_aprog" <= (type) "'\<alpha> hp_vars_ext prog"

subsection \<open>Syntactic type setup\<close>

notation hp_vars_child_lens ("\<Sigma>\<^sub>H\<^sub>P")

syntax
  "_svid_st_alpha"  :: "svid" ("\<Sigma>\<^sub>H\<^sub>P")
  
translations
  "_svid_st_alpha" => "CONST hp_vars_child_lens"

subsection \<open>Substitution lift and drop\<close>
  
abbreviation lift_usubst_hpa_rel ("\<lceil>_\<rceil>\<^sub>S\<^sub>H\<^sub>P\<^sub>A")
where "\<lceil>\<sigma>\<rceil>\<^sub>S\<^sub>H\<^sub>P\<^sub>A \<equiv> \<sigma> \<oplus>\<^sub>s \<Sigma>\<^sub>H\<^sub>P"

abbreviation lift_usubst_hpa ("\<lceil>_\<rceil>\<^sub>s\<^sub>H\<^sub>P\<^sub>A")
  where "\<lceil>\<sigma>\<rceil>\<^sub>s\<^sub>H\<^sub>P\<^sub>A \<equiv> \<lceil>\<lceil>\<sigma>\<rceil>\<^sub>s\<rceil>\<^sub>S\<^sub>H\<^sub>P\<^sub>A"
    
abbreviation drop_usubst_hpa ("\<lfloor>_\<rfloor>\<^sub>S\<^sub>H\<^sub>P\<^sub>A")
where "\<lfloor>\<sigma>\<rfloor>\<^sub>S\<^sub>H\<^sub>P\<^sub>A \<equiv> \<sigma> \<restriction>\<^sub>s \<Sigma>\<^sub>H\<^sub>P"

abbreviation drop_usubst_hpa_rel ("\<lfloor>_\<rfloor>\<^sub>s\<^sub>H\<^sub>P\<^sub>A")
where "\<lfloor>\<sigma>\<rfloor>\<^sub>s\<^sub>H\<^sub>P\<^sub>A \<equiv> \<lfloor>\<lfloor>\<sigma>\<rfloor>\<^sub>S\<^sub>H\<^sub>P\<^sub>A\<rfloor>\<^sub>s"

subsection \<open>Expressions lift and drop\<close>

abbreviation lift_uexpr_hpa ("\<lceil>_\<rceil>\<^sub>H\<^sub>P\<^sub>A")
where "\<lceil>P\<rceil>\<^sub>H\<^sub>P\<^sub>A \<equiv> P \<oplus>\<^sub>p \<Sigma>\<^sub>H\<^sub>P "

abbreviation lift_pre_uexpr_hpa ("\<lceil>_\<rceil>\<^sub>H\<^sub>P\<^sub>A\<^sub><")
where "\<lceil>p\<rceil>\<^sub>H\<^sub>P\<^sub>A\<^sub>< \<equiv> \<lceil>\<lceil>p\<rceil>\<^sub><\<rceil>\<^sub>H\<^sub>P\<^sub>A"

abbreviation lift_post_uexpr_hpa ("\<lceil>_\<rceil>\<^sub>H\<^sub>P\<^sub>A\<^sub>>")
where "\<lceil>p\<rceil>\<^sub>H\<^sub>P\<^sub>A\<^sub>> \<equiv> \<lceil>\<lceil>p\<rceil>\<^sub>>\<rceil>\<^sub>H\<^sub>P\<^sub>A"

abbreviation drop_uexpr_hpa ("\<lfloor>_\<rfloor>\<^sub>H\<^sub>P\<^sub>A")
where "\<lfloor>P\<rfloor>\<^sub>H\<^sub>P\<^sub>A \<equiv> P \<restriction>\<^sub>p \<Sigma>\<^sub>H\<^sub>P"

abbreviation drop_pre_uexpr_hpa ("\<lfloor>_\<rfloor>\<^sub><\<^sub>H\<^sub>P\<^sub>A")
where "\<lfloor>P\<rfloor>\<^sub><\<^sub>H\<^sub>P\<^sub>A \<equiv> \<lfloor>\<lfloor>P\<rfloor>\<^sub>H\<^sub>P\<^sub>A\<rfloor>\<^sub><"

abbreviation droppost_uexpr_hpa ("\<lfloor>_\<rfloor>\<^sub>>\<^sub>H\<^sub>P\<^sub>A")
where "\<lfloor>P\<rfloor>\<^sub>>\<^sub>H\<^sub>P\<^sub>A \<equiv> \<lfloor>\<lfloor>P\<rfloor>\<^sub>H\<^sub>P\<^sub>A\<rfloor>\<^sub>>"

section \<open>Healthiness conditions\<close>
text \<open> For this heap model, we use the identity function as healthiness condition \<close>  

section \<open>Control flow statements\<close>
  
subsection \<open>Top for heap theory\<close>

term "MAGIC"
  
subsection \<open>Bot for heap theory\<close>

term "ABORT"

subsection \<open>Skip for heap theory\<close>

term "SKIP" 
  
subsection \<open>Throw for heap theory\<close>  
  
term "THROW"
  
subsection \<open>Assign for heap theory\<close>
 
abbreviation assigns_hpa :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a hp_aprog" ("\<langle>_\<rangle>\<^sub>h\<^sub>p\<^sub>a")
  where  "assigns_hpa \<sigma> \<equiv> \<langle>\<sigma> \<oplus>\<^sub>s \<Sigma>\<^sub>H\<^sub>P\<rangle>\<^sub>a"
    
adhoc_overloading
  uassigns assigns_hps

translations
  "_assignment xs vs" => "CONST assigns_hpa (_mk_usubst (CONST id) xs vs)"
  "x :== v" <= "CONST assigns_hpa (CONST subst_upd (CONST id) (CONST svar x) v)"

subsection \<open>Substitution for heap theory\<close> 

abbreviation subst_hpa :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a hp_aprog \<Rightarrow> 'a hp_aprog"
  where "subst_hpa \<sigma> P \<equiv> subst_a (\<sigma> \<oplus>\<^sub>s \<Sigma>\<^sub>H\<^sub>P) P"
    
adhoc_overloading
   usubst subst_hpa  
   
subsection \<open>Conditional for heap theory\<close>
   
no_notation cond_a ("IF (_)/ THEN (_) ELSE (_) FI") 

abbreviation cond_hpa :: "'a cond \<Rightarrow> 'a hp_aprog \<Rightarrow> 'a hp_aprog \<Rightarrow> 'a hp_aprog" ("IF (_)/ THEN (_) ELSE (_) FI")
where "cond_hpa b P Q \<equiv> (cond_a (b\<oplus>\<^sub>p \<Sigma>\<^sub>H\<^sub>P) P Q)"

subsection \<open>assert and assume for heap theory\<close>

no_notation assume_a ("_\<^sup>\<top>\<^sup>P")
  
abbreviation assume_hpa :: "'\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog" ("_\<^sup>\<top>\<^sup>P")  
  where "assume_hpa c \<equiv> (IF c THEN SKIP ELSE MAGIC FI)"

no_notation assert_a ("_\<^sub>\<bottom>\<^sub>P")  
  
abbreviation assert_hpa :: "'\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog" ("_\<^sub>\<bottom>\<^sub>P")
  where "assert_hpa c \<equiv> (IF c THEN SKIP ELSE ABORT FI)"

subsection \<open>sequential composition for heap theory\<close> 

term "x :== v ; P"
  
subsection \<open>try catch for heap theory\<close> 
  
term "TRY x :== v; SKIP CATCH THROW; P END" 
  
subsection \<open>recursion for heap theory\<close>
term "\<mu>\<^sub>a X \<bullet> IF b THEN P ELSE Q FI"
term  "\<nu>\<^sub>a X \<bullet> IF b THEN P ELSE Q FI"
  
subsection \<open>Iterations for heap theory\<close>
term "(IF \<not> exit THEN P ELSE SKIP FI)"
  
no_notation from_until_lfp_a  ("FROM (_)/ UNTIL (_)/ DO (_) OD")

definition from_until_lfp_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("FROM (_)/ UNTIL (_)/ DO (_) OD")
where "FROM init UNTIL exit DO body OD = init ; (\<mu>\<^sub>a X \<bullet> IF \<not> exit THEN (body ; X) ELSE SKIP FI)"  

no_notation from_until_lfp_invr_a("FROM (_)/ INVAR (_)/ UNTIL (_)/  DO (_) OD")   
definition from_until_lfp_invr_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("FROM (_)/ INVAR (_)/ UNTIL (_)/  DO (_) OD")  
where "FROM init INVAR invar UNTIL exit DO body OD = FROM init UNTIL exit DO body OD"

no_notation from_until_lfp_invr_vrt_a("FROM (_)/ INVAR (_)/ VRT (_)/ UNTIL (_)/ DO (_) OD")  
definition from_until_lfp_invr_vrt_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("FROM (_)/ INVAR (_)/ VRT (_)/ UNTIL (_)/ DO (_) OD")
where "FROM init INVAR invar VRT vari UNTIL exit DO body OD = FROM init UNTIL exit DO body OD" 

no_notation while_lfp_a("WHILE (_)/ DO (_) OD")  
definition while_lfp_hpa :: 
  "'\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("WHILE (_)/ DO (_) OD")
where "WHILE b DO body OD = FROM SKIP UNTIL \<not>b DO body OD"    

no_notation while_lfp_invr_a  ("INVAR (_)/ WHILE (_)/ DO (_) OD")
definition while_lfp_invr_hpa :: 
  "'\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("INVAR (_)/ WHILE (_)/ DO (_) OD")
where "INVAR invar WHILE b DO body OD = WHILE b DO body OD" 

no_notation while_lfp_invr_vrt_a ("INVAR (_)/ VRT (_)/ WHILE (_)/ DO (_) OD")  
definition while_lfp_invr_vrt_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("INVAR (_)/ VRT (_)/ WHILE (_)/ DO (_) OD")
where "INVAR invar VRT vari WHILE b DO body OD = WHILE b DO body OD"

no_notation do_while_lfp_a ("DO (_)/  WHILE (_) OD")
definition do_while_lfp_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog" 
  ("DO (_)/  WHILE (_) OD")
where "DO body WHILE b OD = FROM body UNTIL \<not> b DO body OD"

no_notation do_while_lfp_invr_a ("DO (_)/ INVAR (_)/ WHILE (_) OD")
definition do_while_lfp_invr_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog" 
  ("DO (_)/ INVAR (_)/ WHILE (_) OD")
where "DO body INVAR invar WHILE b OD = DO body WHILE b OD"

no_notation do_while_lfp_invr_vrt_a  ("DO (_)/ INVAR (_)/ VRT (_)/ WHILE (_) OD") 
definition do_while_lfp_invr_vrt_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog" 
  ("DO (_)/ INVAR (_)/ VRT (_)/ WHILE (_) OD")
where "DO body INVAR invar VRT vari WHILE b OD = DO body WHILE b OD"  

no_notation for_lfp_a  ("FOR ('(_,_,_'))/ DO (_) OD")   
definition for_lfp_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("FOR ('(_,_,_'))/ DO (_) OD")
where "FOR (init, b, incr) DO body OD = FROM init UNTIL \<not> b DO body ; incr OD"

no_notation for_lfp_invr_a  ("FOR ('(_,_,_'))/ INVAR (_)/ DO (_) OD")     
definition for_lfp_invr_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("FOR ('(_,_,_'))/ INVAR (_)/ DO (_) OD")
where "FOR (init, b, incr) INVAR invar DO body OD = FOR (init, b, incr) DO body OD"  

no_notation for_lfp_invr_vrt_a  ("FOR ('(_,_,_'))/ INVAR (_)/ VRT (_)/ DO (_) OD")     
definition for_lfp_invr_vrt_hpa :: 
  "'\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> hp_aprog \<Rightarrow> '\<alpha> hp_aprog" 
  ("FOR ('(_,_,_'))/ INVAR (_)/ VRT (_)/ DO (_) OD")
where "FOR (init, b, incr) INVAR invar VRT vari DO body OD = FOR (init, b, incr) DO body OD"  

subsection \<open>Frame and anti-frame\<close>

abbreviation 
  "frame_hpa x P \<equiv> (frame_a (lens_comp x \<Sigma>\<^sub>H\<^sub>P) P)"
  
abbreviation 
  "anti_frame_hpa x P \<equiv> (antiframe_a (lens_comp x \<Sigma>\<^sub>H\<^sub>P) P)" 


translations
  "_frame x P" => "CONST frame_hpa x P"
  "_frame (_salphaset (_salphamk x)) P" <= "CONST frame_hpa x P"
  "_antiframe x P" => "CONST anti_frame_hpa x P"
  "_antiframe (_salphaset (_salphamk x)) P" <= "CONST anti_frame_hpa x P"

subsection \<open>Monotonicity for heap theory\<close> 
  
section \<open>Arrays\<close>
term "mwb_lens"  
term "heap_w32"  
term "heap_w32_update"  
term "\<forall>x. P x"
term "lens_get X (lens_put X \<sigma> ( (p +\<^sub>p i)))"
term "lattice"  
term "partial_order"  
(*locale arrays' = mwb_lens +
  
  fixes is_valid_w32 :: "'b \<Rightarrow> 32 word ptr \<Rightarrow> bool"
  assumes put_valid: "is_valid_w32 (put \<sigma> v) p = is_valid_w32 \<sigma> p" 
begin
subsection \<open>Laws on pointers\<close> 
 lemma ptr_add_assoc:
  "p +\<^sub>p (i + j) = p +\<^sub>p i +\<^sub>p j"
  by (simp add: CTypesDefs.ptr_add_def distrib_right)

lemma ptr_add_commute:
  "p +\<^sub>p i +\<^sub>p j = p +\<^sub>p j +\<^sub>p i"
  by (metis ptr_add_assoc add.commute)
    
lemma ptr_add_sub: "p +\<^sub>p n = (p +\<^sub>p a) +\<^sub>p (n - a)"
proof -
  have "(p +\<^sub>p a) +\<^sub>p (n - a) = p +\<^sub>p (a + (n - a))" by (metis ptr_add_assoc)
  thus ?thesis by auto
qed
  
lemma ptr_suc_neq[simp]: "(p::32 word ptr) +\<^sub>p 1 \<noteq> p"
proof -
  have "ptr_val (p +\<^sub>p 1) \<noteq> ptr_val p" by simp
  thus "p +\<^sub>p 1 \<noteq> p" by (simp only: ptr_val_inj) simp
qed
 
definition
  not_at_end :: "word32 ptr \<Rightarrow> bool"
where
  "not_at_end p \<equiv> p < p +\<^sub>p 1"
  
lemma not_at_end_lt: 
  "(\<forall>i<(n::nat). not_at_end (p +\<^sub>p i)) \<Longrightarrow> n > 0 \<Longrightarrow> p < p +\<^sub>p n"
proof  (induction n)
  case 0
  then show ?case by auto
next
  case (Suc n)
  then show ?case 
  proof (auto simp: not_at_end_def ptr_add_assoc ptr_add_commute, goal_cases)
    case 1
    then show ?case 
      apply (cases n) 
       apply auto 
      done
  qed
qed

subsection \<open>Array validity\<close>
term "put \<sigma> v"  
term "(ptr_val (p +\<^sub>p i))"  
term "get (put \<sigma> (ptr_val (p +\<^sub>p i)))"  
definition
  array :: "'b \<Rightarrow> word32 ptr \<Rightarrow> word32 list \<Rightarrow> bool"
where
  "array s p data = (\<forall>i < (length data). 
                      not_at_end (p +\<^sub>p i) \<and> 
                      is_valid_w32 s (p +\<^sub>p i) \<and>  
                      heap_w32 s (p +\<^sub>p i) = data ! i)"
  
lemma arr_ind_def:
  "array s p [] = True \<and>
   array s p (x#xs) = (not_at_end p \<and>
                        ((is_valid_w32 s p) \<and> (heap_w32 s p = x) \<and> array s (p +\<^sub>p 1) xs))"
  apply (auto simp: array_def)
       apply (auto simp: ptr_add_assoc less_Suc_eq_0_disj)
  done
  
lemma empty_array[simp]: "length data = 0 \<Longrightarrow> array s p data"
  by (simp add: array_def)
    
lemma arr_access_is_list_access[simp]:
  "\<lbrakk>array s p data; i < length data\<rbrakk> \<Longrightarrow> 
  not_at_end (p +\<^sub>p i) \<and> is_valid_w32 s (p +\<^sub>p i) \<and> heap_w32 s (p +\<^sub>p i) = data ! i"
  by (auto simp: array_def)
    
end  *)
locale arrays = 
  fixes is_valid_w32 :: "'s \<Rightarrow> 32 word ptr \<Rightarrow> bool"
  fixes heap_w32 :: "'s \<Rightarrow> 32 word ptr \<Rightarrow> 32 word"
  fixes heap_w32_update :: "((32 word ptr \<Rightarrow> 32 word) \<Rightarrow> 32 word ptr \<Rightarrow> 32 word) \<Rightarrow> 's \<Rightarrow> 's"
  assumes update_update: "heap_w32_update f\<^sub>1 (heap_w32_update f\<^sub>2 s) = heap_w32_update (f\<^sub>1 o f\<^sub>2) s"
  assumes get_update: "heap_w32 (heap_w32_update f s) = f (heap_w32 s)"
  assumes update_id: "heap_w32_update id s = s"
  assumes update_valid: "is_valid_w32 (heap_w32_update f s) p = is_valid_w32 s p"
begin

(* Some pointer arithmetic facts *)
lemma ptr_add_assoc:
  "p +\<^sub>p (i + j) = p +\<^sub>p i +\<^sub>p j"
  by (simp add: CTypesDefs.ptr_add_def distrib_right)

lemma ptr_add_commute:
  "p +\<^sub>p i +\<^sub>p j = p +\<^sub>p j +\<^sub>p i"
  by (metis ptr_add_assoc add.commute)
    
lemma ptr_add_sub: "p +\<^sub>p n = (p +\<^sub>p a) +\<^sub>p (n - a)"
proof -
  have "(p +\<^sub>p a) +\<^sub>p (n - a) = p +\<^sub>p (a + (n - a))" by (metis ptr_add_assoc)
  thus ?thesis by auto
qed
  
lemma ptr_suc_neq[simp]: "(p::32 word ptr) +\<^sub>p 1 \<noteq> p"
proof -
  have "ptr_val (p +\<^sub>p 1) \<noteq> ptr_val p" by simp
  thus "p +\<^sub>p 1 \<noteq> p" by (simp only: ptr_val_inj) simp
qed

definition
  not_at_end :: "word32 ptr \<Rightarrow> bool"
where
  "not_at_end p \<equiv> p < p +\<^sub>p 1"

lemma not_at_end_lt: "\<lbrakk>(\<forall>i<(n::nat). not_at_end (p +\<^sub>p i)); n > 0\<rbrakk> \<Longrightarrow> p < p +\<^sub>p n"
  apply (induction n)
   apply (auto simp: not_at_end_def ptr_add_assoc ptr_add_commute)
  apply (case_tac "n=0")
  by auto
    
definition
  array :: "'s \<Rightarrow> word32 ptr \<Rightarrow> word32 list \<Rightarrow> bool"
where
  "array s p data = (\<forall>i < (length data). 
                      not_at_end (p +\<^sub>p i) \<and> 
                      is_valid_w32 s (p +\<^sub>p i) \<and>  
                      heap_w32 s (p +\<^sub>p i) = data ! i)"
  
lemma arr_ind_def:
  "array s p [] = True \<and>
   array s p (x#xs) = (not_at_end p \<and>
                        ((is_valid_w32 s p) \<and> (heap_w32 s p = x) \<and> array s (p +\<^sub>p 1) xs))"
  apply (auto simp: array_def)
       apply (auto simp: ptr_add_assoc less_Suc_eq_0_disj)
  done
  
lemma empty_array[simp]: "length data = 0 \<Longrightarrow> array s p data"
  by (simp add: array_def)
    
lemma arr_access_is_list_access[simp]:
  "\<lbrakk>array s p data; i < length data\<rbrakk> \<Longrightarrow> 
  not_at_end (p +\<^sub>p i) \<and> is_valid_w32 s (p +\<^sub>p i) \<and> heap_w32 s (p +\<^sub>p i) = data ! i"
  by (auto simp: array_def)
    
definition read_array :: "'s \<Rightarrow> 32 word ptr \<Rightarrow> nat \<Rightarrow> 32 word list" 
  where "read_array s p n = map (heap_w32 s) (map (\<lambda>i. (p +\<^sub>p i)) [0 ..< n])"

(* If you read an array it's the data in the array *)
lemma read_array_works[simp]: 
  "array s p data \<Longrightarrow> read_array s p (length data) = data"
  by (subst List.list_eq_iff_nth_eq) (simp add: read_array_def)
    
lemma read_array_len[simp]:
  "length (read_array s p n) = n"
  by (auto simp: read_array_def)
    
lemma read_array_zero[simp]:
  "read_array s p 0 = []"
  by (auto simp: read_array_def)

lemma subarray_aux: "\<lbrakk>(i::nat) < length data; j \<le> i; ia < i - j\<rbrakk> \<Longrightarrow> 
                     ia + j < length data \<and> p +\<^sub>p int ia +\<^sub>p int j = p +\<^sub>p (ia + j)"
  by (auto simp: ptr_add_assoc)
      
lemma subarray_is_array: "\<lbrakk>array s p data; i < length data; j \<le> i\<rbrakk> \<Longrightarrow> 
                          array s (p +\<^sub>p j) (read_array s (p +\<^sub>p j) (i-j))"
  apply (simp only: array_def)
  apply (auto simp: read_array_def ptr_add_commute ptr_add_assoc)
   apply (simp only: subarray_aux)+
  done
  
lemma arr_access_lt_weak: 
  assumes "array s p data" and "i < length data" and 0: "j < i"
  shows "p +\<^sub>p j < p +\<^sub>p i"
proof -
  have "array s (p +\<^sub>p j) (read_array s (p +\<^sub>p j) (i - j))" using assms subarray_is_array by auto
  hence "p +\<^sub>p j < p +\<^sub>p (j + (i - j))" by (simp add: 0 not_at_end_lt ptr_add_assoc) 
  thus "p +\<^sub>p j < p +\<^sub>p i" using 0 by auto
qed
  
lemma arr_access_cmp[simp]: 
      "\<lbrakk>array s p data; i < length data; j < length data\<rbrakk> \<Longrightarrow> 
        (p +\<^sub>p i = p +\<^sub>p j) = (i = j) \<and> (p +\<^sub>p i \<noteq> p +\<^sub>p j) = (i \<noteq> j) \<and>
        (p +\<^sub>p i < p +\<^sub>p j) = (i < j) \<and> (p +\<^sub>p i \<le> p +\<^sub>p j) = (i \<le> j)"
  apply (cases "\<not>i < j")
   apply (cases "j < i")
    apply (auto)
  using arr_access_lt_weak apply fastforce+
    done

lemma updated_array_exists[simp]: "array s p (read_array s p n) \<Longrightarrow> 
       array (heap_w32_update f s) p (read_array (heap_w32_update f s) p n)"
  apply (auto simp: read_array_def)
  apply (simp only: array_def)
  by (auto simp: update_valid)

lemma read_array_upd_prop: "\<lbrakk>array s p (read_array s p n); i < n\<rbrakk> \<Longrightarrow> 
        ((read_array (heap_w32_update ((\<lambda>a. a(p +\<^sub>p i := x)) \<circ> f) s) p n) = 
         (read_array (heap_w32_update f s) p n)[i := x])"
  by (intro nth_equalityI) (auto simp: read_array_def get_update fun_upd_apply)
    
lemma of_id: "f = f \<circ> id"
  by simp
    
lemma read_array_upd_prop': "\<lbrakk>array s p (read_array s p n); i < n\<rbrakk> \<Longrightarrow> 
        ((read_array (heap_w32_update (\<lambda>a. a(p +\<^sub>p i := x)) s) p n) = 
         (read_array s p n)[i := x])"
  apply (subst (9) of_id)
  by (simp only: read_array_upd_prop update_id)
    
lemma array_write: "\<lbrakk>array s p data; i < length data\<rbrakk> \<Longrightarrow> array (heap_w32_update (\<lambda>a. a(p +\<^sub>p i := x)) s) p (data[i := x])"
  apply (simp only: array_def)
  apply (auto simp: update_valid get_update array_def[THEN sym] fun_upd_apply)
  done

(* Rewriting things *)
lemma exp_arr_unat_add: 
  "\<lbrakk>array s p (read_array s p (unat (n::addr))); unat (i::addr) < unat n\<rbrakk> \<Longrightarrow> unat (i + 1) = unat i + 1"
  by unat_arith
    
lemma exp_unat_sub:
  "\<lbrakk>unat (i::addr) \<le> unat (n::addr)\<rbrakk> \<Longrightarrow> unat (n - i) = unat n - unat i"
  by unat_arith
    
lemma uint_add_to_unat_add: "p +\<^sub>p uint i = p +\<^sub>p unat i"
proof -
  have "uint i = unat i" using uint_nat by auto
  thus "p +\<^sub>p uint i = p +\<^sub>p unat i" by simp
qed
  
lemma read_array_upd_prop_alt: "\<lbrakk>array s p (read_array s p n); (1+b) < n\<rbrakk> \<Longrightarrow> 
        ((read_array (heap_w32_update ((\<lambda>a. a(p +\<^sub>p 1 +\<^sub>p b := x)) \<circ> f) s) p n) = 
         (read_array (heap_w32_update f s) p n)[(1+b) := x])"
  apply (simp only: ptr_add_assoc[THEN sym])
  apply (metis Suc_eq_plus1_left of_nat_Suc read_array_upd_prop)
  done
    
lemma arr_access_is_list_access_alt:
  "\<lbrakk>array s p data; 1 + i < length data\<rbrakk> \<Longrightarrow> 
  not_at_end (p +\<^sub>p 1 +\<^sub>p i) \<and> is_valid_w32 s (p +\<^sub>p 1 +\<^sub>p i) \<and> heap_w32 s (p +\<^sub>p 1 +\<^sub>p i) = data ! (1 + i)"
  apply (simp only: ptr_add_assoc[THEN sym])
    using arr_access_is_list_access  apply fastforce
    done

lemma array_append:
  "array s p (a @ b) = (array s p a \<and> array s (p +\<^sub>p int (length a)) b)"
  apply (induction a arbitrary: p) 
  unfolding arr_ind_def ptr_add_def
    apply (simp_all add: field_simps ptr_add_assoc arr_ind_def ptr_add_def)
  done
    
lemma read_array_append:
  "(read_array s p n) @ (read_array s (p +\<^sub>p n) m) = (read_array s p (n + m))"
  apply (rule nth_equalityI)
   apply (auto simp: read_array_def nth_append ptr_add_assoc[THEN sym])
    done
    
lemma array_read_array_append:
  "(array s p (read_array s p n) \<and> array s (p +\<^sub>p n) (read_array s (p +\<^sub>p n) m)) = 
   array s p (read_array s p (n + m))"
  by (simp only: read_array_len read_array_append[THEN sym] array_append)

method array_rewrite = 
    (simp only: 
      uint_add_to_unat_add exp_arr_unat_add exp_unat_sub ptr_add_assoc ptr_add_0_id
      arr_access_is_list_access_alt
      read_array_upd_prop_alt read_array_upd_prop read_array_upd_prop'
      read_array_len (*read_array_works*) arr_access_cmp (*updated_array_exists*) arr_access_is_list_access (*ptr_suc_neq*)
      read_array_append[THEN sym] array_append
      )

end  
term "is_valid_w32" 
term "lens_get(fst\<^sub>L ;\<^sub>L heap_raw)" (* A kind of This is the type of lens that I need*)

term "weak_lens.update (fst\<^sub>L ;\<^sub>L heap_raw)"  (*A kind of This is the type of lens that I need*)  
  
sublocale mwb_lens \<subseteq> arrays undefined 
  apply unfold_locales
    oops
  
  
lemma [simp]: 
  "(h_val (hrs_mem (hrs_mem_update (heap_update (p::'b::mem_type ptr) v) (h,tt))) p) = v" 
  unfolding hrs_mem_def hrs_mem_update_def
  by (simp add: h_val_heap_update)
    


end

