(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)
(******************************************************************************
 * Authors: Josh Bockenek, Yakoub Nemouchi.
 ******************************************************************************)

section \<open>Syntax extensions for UTP\<close>

theory utp_extensions
imports
  "../../Isabelle-UTP/utp/utp"
  "~~/src/HOL/Library/Multiset"
begin
recall_syntax \<comment> \<open>Fixes notation issue with inclusion of HOL libraries.\<close>

subsection \<open>Notation\<close>

text \<open>We need multisets for concise list invariants for sorting. Also, int/nat conversion is
sometimes needed as some loop methods mix array indices and loop variables (which sometimes rely on
going below 0 for termination). Bitwise operations and record access/update are included for
completeness.\<close>

text \<open>A helper function for record updating.\<close>
lift_definition rec_update_wrapper :: \<open>('a, '\<alpha>) uexpr \<Rightarrow> ('a \<Rightarrow> 'a, '\<alpha>) uexpr\<close> is
  \<open>\<lambda>v s _. v s\<close> .

syntax
  "_umset" :: \<open>('a list, '\<alpha>) uexpr \<Rightarrow> ('a multiset, '\<alpha>) uexpr\<close> ("mset\<^sub>u'(_')")
  "_unat" :: \<open>(nat, '\<alpha>) uexpr \<Rightarrow> (int, '\<alpha>) uexpr\<close> ("int\<^sub>u'(_')")
  "_uint" :: \<open>(int, '\<alpha>) uexpr \<Rightarrow> (nat, '\<alpha>) uexpr\<close> ("nat\<^sub>u'(_')")
  "_uapply_rec" :: \<open>('a, '\<alpha>) uexpr \<Rightarrow> utuple_args \<Rightarrow> ('b, '\<alpha>) uexpr\<close> ("_\<lparr>_\<rparr>\<^sub>r" [999,0] 999)
  "_uupd_rec"   :: \<open>('a, '\<alpha>) uexpr \<Rightarrow> (('b \<Rightarrow> 'b) \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('b, '\<alpha>) uexpr \<Rightarrow> ('a, '\<alpha>) uexpr\<close> ("_/'(_ /\<mapsto>/ _')\<^sub>r" [900,0,0] 900)

translations
  "_umset" \<rightleftharpoons> "CONST uop CONST mset"
  "_uint" \<rightleftharpoons> "CONST uop CONST int"
  "_unat" \<rightleftharpoons> "CONST uop CONST nat"
  "f\<lparr>kf\<rparr>\<^sub>r" \<rightharpoonup> "CONST uop kf f"
  "f(k \<mapsto> v)\<^sub>r" \<rightharpoonup> "CONST bop k (CONST rec_update_wrapper v) f"

subsection \<open>Extra stuff to work more-arg functions into UTP\<close>

lift_definition qiop ::
  \<open>('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e \<Rightarrow> 'f) \<Rightarrow>
   ('a, '\<alpha>) uexpr \<Rightarrow> ('b, '\<alpha>) uexpr \<Rightarrow> ('c, '\<alpha>) uexpr \<Rightarrow> ('d, '\<alpha>) uexpr \<Rightarrow> ('e, '\<alpha>) uexpr \<Rightarrow>
   ('f, '\<alpha>) uexpr\<close>
  is \<open>\<lambda>f u v w x y b. f (u b) (v b) (w b) (x b) (y b)\<close> .
lift_definition sxop ::
  \<open>('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e \<Rightarrow> 'f \<Rightarrow> 'g) \<Rightarrow>
   ('a, '\<alpha>) uexpr \<Rightarrow> ('b, '\<alpha>) uexpr \<Rightarrow> ('c, '\<alpha>) uexpr \<Rightarrow> ('d, '\<alpha>) uexpr \<Rightarrow> ('e, '\<alpha>) uexpr \<Rightarrow>
   ('f, '\<alpha>) uexpr \<Rightarrow> ('g, '\<alpha>) uexpr\<close>
  is \<open>\<lambda>f u v w x y z b. f (u b) (v b) (w b) (x b) (y b) (z b)\<close> .
lift_definition sepop ::
  \<open>('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e \<Rightarrow> 'f \<Rightarrow> 'g \<Rightarrow> 'h) \<Rightarrow>
   ('a, '\<alpha>) uexpr \<Rightarrow> ('b, '\<alpha>) uexpr \<Rightarrow> ('c, '\<alpha>) uexpr \<Rightarrow> ('d, '\<alpha>) uexpr \<Rightarrow> ('e, '\<alpha>) uexpr \<Rightarrow>
   ('f, '\<alpha>) uexpr \<Rightarrow> ('g, '\<alpha>) uexpr \<Rightarrow> ('h, '\<alpha>) uexpr\<close>
  is \<open>\<lambda>f u v w x y z a b. f (u b) (v b) (w b) (x b) (y b) (z b) (a b)\<close> .
update_uexpr_rep_eq_thms \<comment> \<open>Necessary to get the above utilized by \{pred,rel\}\_\{auto,simp\}\<close>

text \<open>The below lemmas do not seem useful in general but are included for completeness.\<close>
lemma qiop_ueval [ueval]: \<open>\<lbrakk>qiop f v x y z w\<rbrakk>\<^sub>e b = f (\<lbrakk>v\<rbrakk>\<^sub>e b) (\<lbrakk>x\<rbrakk>\<^sub>e b) (\<lbrakk>y\<rbrakk>\<^sub>e b) (\<lbrakk>z\<rbrakk>\<^sub>e b) (\<lbrakk>w\<rbrakk>\<^sub>e b)\<close>
  by transfer simp

lemma subst_qiop [usubst]: \<open>\<sigma> \<dagger> qiop f t u v w x = qiop f (\<sigma> \<dagger> t) (\<sigma> \<dagger> u) (\<sigma> \<dagger> v) (\<sigma> \<dagger> w) (\<sigma> \<dagger> x)\<close>
  by transfer simp

lemma unrest_qiop [unrest]: \<open>\<lbrakk>x \<sharp> t; x \<sharp> u; x \<sharp> v; x \<sharp> w; x \<sharp> y\<rbrakk> \<Longrightarrow> x \<sharp> qiop f t u v w y\<close>
  by transfer simp

lemma aext_qiop [alpha]:
  \<open>qiop f t u v w x \<oplus>\<^sub>p a = qiop f (t \<oplus>\<^sub>p a) (u \<oplus>\<^sub>p a) (v \<oplus>\<^sub>p a) (w \<oplus>\<^sub>p a) (x \<oplus>\<^sub>p a)\<close>
  by pred_auto

lemma lit_qiop_simp [lit_simps]:
  \<open>\<guillemotleft>i x y z u t\<guillemotright> = qiop i \<guillemotleft>x\<guillemotright> \<guillemotleft>y\<guillemotright> \<guillemotleft>z\<guillemotright> \<guillemotleft>u\<guillemotright> \<guillemotleft>t\<guillemotright>\<close>
  by transfer simp

lemma sxop_ueval [ueval]: \<open>\<lbrakk>sxop f v x y z w t\<rbrakk>\<^sub>e b = f (\<lbrakk>v\<rbrakk>\<^sub>eb) (\<lbrakk>x\<rbrakk>\<^sub>eb) (\<lbrakk>y\<rbrakk>\<^sub>eb) (\<lbrakk>z\<rbrakk>\<^sub>eb) (\<lbrakk>w\<rbrakk>\<^sub>eb) (\<lbrakk>t\<rbrakk>\<^sub>eb)\<close>
  by transfer simp

lemma subst_sxop [usubst]:
  \<open>\<sigma> \<dagger> sxop f t u v w x y = sxop f (\<sigma> \<dagger> t) (\<sigma> \<dagger> u) (\<sigma> \<dagger> v) (\<sigma> \<dagger> w) (\<sigma> \<dagger> x) (\<sigma> \<dagger> y)\<close>
  by transfer simp

lemma unrest_sxop [unrest]: \<open>\<lbrakk>x \<sharp> t; x \<sharp> u; x \<sharp> v; x \<sharp> w; x \<sharp> y; x \<sharp> z\<rbrakk> \<Longrightarrow> x \<sharp> sxop f t u v w y z\<close>
  by transfer simp

lemma aext_sxop [alpha]:
  \<open>sxop f t u v w x y \<oplus>\<^sub>p a = sxop f (t \<oplus>\<^sub>p a) (u \<oplus>\<^sub>p a) (v \<oplus>\<^sub>p a) (w \<oplus>\<^sub>p a) (x \<oplus>\<^sub>p a) (y \<oplus>\<^sub>p a)\<close>
  by pred_auto

lemma lit_sxop_simp [lit_simps]:
  \<open>\<guillemotleft>i x y z u t v\<guillemotright> = sxop i \<guillemotleft>x\<guillemotright> \<guillemotleft>y\<guillemotright> \<guillemotleft>z\<guillemotright> \<guillemotleft>u\<guillemotright> \<guillemotleft>t\<guillemotright> \<guillemotleft>v\<guillemotright>\<close>
  by transfer simp

end